#include "dsp.h"

float g_fft_sin[ 22 ] = 
{
    (float)-5.89793E-10,
    (float)-1,
    (float)-0.707106781,
    (float)-0.382683432,
    (float)-0.195090322,
    (float)-0.09801714,
    (float)-0.049067674,
    (float)-0.024541229,
    (float)-0.012271538,
    (float)-0.006135885,
    (float)-0.003067957,
    (float)-0.00153398,
    (float)-0.00076699,
    (float)-0.000383495,
    (float)-0.000191748,
    (float)-9.58738E-05,
    (float)-4.79369E-05,
    (float)-2.39684E-05,
    (float)-1.19842E-05,
    (float)-5.99211E-06,
    (float)-2.99606E-06,
    (float)-1.49803E-06
};

float g_fft_sin2[ 22 ] = 
{
    (float)-1,
    (float)-0.707106781,
    (float)-0.382683432,
    (float)-0.195090322,
    (float)-0.09801714,
    (float)-0.049067674,
    (float)-0.024541229,
    (float)-0.012271538,
    (float)-0.006135885,
    (float)-0.003067957,
    (float)-0.00153398,
    (float)-0.00076699,
    (float)-0.000383495,
    (float)-0.000191748,
    (float)-9.58738E-05,
    (float)-4.79369E-05,
    (float)-2.39684E-05,
    (float)-1.19842E-05,
    (float)-5.99211E-06,
    (float)-2.99606E-06,
    (float)-1.49803E-06,
    (float)-7.49014E-07
};

void fft_float32( bool inverse, float* fi, float* fr, int fft_size )
{
    int i;
    if( inverse )
    {
	for( i = 0; i < fft_size; i++ ) fi[ i ] = -fi[ i ];
    }
    int nn = fft_size;
    int nd2 = nn / 2;
    int m = 0;
    while( 1 )
    {
	if( 1 << m == fft_size ) break;
	m++;
    }
    int j = nd2;
    int k;
    for( i = 1; i <= nn - 2; i++ )
    {
	if( i < j )
	{
	    float tr = fr[ j ];
	    float ti = fi[ j ];
	    fr[ j ] = fr[ i ];
	    fi[ j ] = fi[ i ];
	    fr[ i ] = tr;
	    fi[ i ] = ti;
	}
	k = nd2;
	nit:
	if( k <= j )
	{
	    j = j - k;
	    k = k / 2;
	    goto nit;
	}
	j = j + k;
    }
    
    int mmax = 1;
    int n = fft_size;
    int istep;
    int cnt = 0;
    while( n > mmax )
    {
	istep = mmax << 1; //istep = 2; mmax = 1; ... istep = fft_size; mmax = fft_size >> 1;
	float wtemp = g_fft_sin2[ cnt ];
	float wpr = -2.0F * wtemp * wtemp;
	float wpi = g_fft_sin[ cnt ];
	float wr = 1.0;
	float wi = 0.0;
	for( m = 0; m < mmax; m ++ ) //m = 0..1; 0..2; ... 0..fft_size >> 1;
	{
	    for( i = m; i < n; i += istep ) 
	    {
		j = i + mmax;
		float tempr = wr * fr[ j ] - wi * fi[ j ];
		float tempi = wr * fi[ j ] + wi * fr[ j ];
		fr[ j ] = fr[ i ] - tempr;
		fi[ j ] = fi[ i ] - tempi;
		fr[ i ] += tempr;
		fi[ i ] += tempi;
	    }
	    wr = ( wtemp = wr ) * wpr - wi * wpi + wr;
	    wi = wi * wpr + wtemp * wpi + wi;
	}
	mmax = istep;
	cnt++;
    }
    if( inverse )
    {
	for( i = 0; i < fft_size; i++ ) 
	{
	    fr[ i ] = fr[ i ] / fft_size;
	    fi[ i ] = -fi[ i ] / fft_size;
	}
    }
}

void fft_float64( bool inverse, double* fi, double* fr, int fft_size )
{
    int i;
    if( inverse )
    {
	for( i = 0; i < fft_size; i++ ) fi[ i ] = -fi[ i ];
    }
    int nn = fft_size;
    int nd2 = nn / 2;
    int m = 0;
    while( 1 )
    {
	if( 1 << m == fft_size ) break;
	m++;
    }
    int j = nd2;
    int k;
    for( i = 1; i <= nn - 2; i++ )
    {
	if( i < j )
	{
	    double tr = fr[ j ];
	    double ti = fi[ j ];
	    fr[ j ] = fr[ i ];
	    fi[ j ] = fi[ i ];
	    fr[ i ] = tr;
	    fi[ i ] = ti;
	}
	k = nd2;
	nit:
	if( k <= j )
	{
	    j = j - k;
	    k = k / 2;
	    goto nit;
	}
	j = j + k;
    }
    
    int mmax = 1;
    int n = fft_size;
    int istep;
    int cnt = 0;
    while( n > mmax )
    {
	istep = mmax << 1; //istep = 2; mmax = 1; ... istep = fft_size; mmax = fft_size >> 1;
	double wtemp = g_fft_sin2[ cnt ];
	double wpr = -2.0F * wtemp * wtemp;
	double wpi = g_fft_sin[ cnt ];
	double wr = 1.0;
	double wi = 0.0;
	for( m = 0; m < mmax; m ++ ) //m = 0..1; 0..2; ... 0..fft_size >> 1;
	{
	    for( i = m; i < n; i += istep ) 
	    {
		j = i + mmax;
		double tempr = wr * fr[ j ] - wi * fi[ j ];
		double tempi = wr * fi[ j ] + wi * fr[ j ];
		fr[ j ] = fr[ i ] - tempr;
		fi[ j ] = fi[ i ] - tempi;
		fr[ i ] += tempr;
		fi[ i ] += tempi;
	    }
	    wr = ( wtemp = wr ) * wpr - wi * wpi + wr;
	    wi = wi * wpr + wtemp * wpi + wi;
	}
	mmax = istep;
	cnt++;
    }
    if( inverse )
    {
	for( i = 0; i < fft_size; i++ ) 
	{
	    fr[ i ] = fr[ i ] / fft_size;
	    fi[ i ] = -fi[ i ] / fft_size;
	}
    }
}
