//################################
//## IOS (IPHONE, IPAD)         ##
//################################

#include <AudioToolbox/AudioToolbox.h>
#include <AudioUnit/AudioUnit.h>

enum
{
    SDRIVER_AUDIOUNIT,
#ifdef AUDIOBUS_URL
    SDRIVER_AUDIOBUS,
#endif
    NUMBER_OF_SDRIVERS
};

const UTF8_CHAR* g_sdriver_infos[] =
{
    "Audio Unit",
#ifdef AUDIOBUS_URL
    "Audiobus"
#endif
};

const UTF8_CHAR* g_sdriver_names[] =
{
    "audiounit",
#ifdef AUDIOBUS_URL
    "audiobus"
#endif
};

int g_sound_driver = 0;

#define kOutputBus 0
#define kInputBus 1

#define INPUT_BUF_FRAMES ( 64 * 1024 )

AudioComponentInstance g_unit = 0;
char g_input_bufs[ 512 ];
void* g_input_buf_lowlevel;
uint* g_input_buf = 0;
void* g_input_buf2 = 0;
volatile uint g_input_buf_wp = 0;
uint g_input_buf_rp = 0;

bool g_au_input = 0;

volatile int g_au_pause_request = 0;

#ifdef AUDIOBUS_URL
extern "C" void iphone_sundog_audiobus_init( void );
extern "C" void iphone_sundog_audiobus_deinit( void );
#endif

static OSStatus playback_callback(
				  void* inRefCon,
				  AudioUnitRenderActionFlags* ioActionFlags,
				  const AudioTimeStamp* inTimeStamp,
				  UInt32 inBusNumber,
				  UInt32 inNumberFrames,
				  AudioBufferList* ioData )
{
    double t = (double)inTimeStamp->mHostTime * g_timebase_nanosec / 20000;
    unsigned long long output_time = (unsigned long long)t;
    for( int b = 0; b < ioData->mNumberBuffers; b++ )
    {
	AudioBuffer* buf = &ioData->mBuffers[ b ];
	g_snd.out_buffer = buf->mData;
	g_snd.out_frames = buf->mDataByteSize / ( 2 * g_snd.channels );
	g_snd.out_time = (ticks_hr_t)output_time;
	g_snd.latency = g_snd.out_frames;
	g_snd.in_buffer = g_input_buf2;
	if( g_input_buf2 && ( ( g_input_buf_wp - g_input_buf_rp ) & ( INPUT_BUF_FRAMES - 1 ) ) >= g_snd.out_frames )
	{
	    for( int i = 0; i < g_snd.out_frames; i++ )
	    {
		((uint*)g_input_buf2)[ i ] = g_input_buf[ g_input_buf_rp ];
		g_input_buf_rp++;
		g_input_buf_rp &= ( INPUT_BUF_FRAMES - 1 );
	    }
	}
	
	if( g_au_pause_request > 0 )
	{
	    g_au_pause_request = 2;
	    mem_set( g_snd.out_buffer, buf->mDataByteSize, 0 );
	}
	else
	{
	    if( main_sound_output_callback( &g_snd, 0 ) )
		*ioActionFlags |= kAudioUnitRenderAction_OutputIsSilence;
	}
    }
    
    return noErr;
}

static OSStatus recording_callback(
				   void* inRefCon, 
				   AudioUnitRenderActionFlags* ioActionFlags, 
				   const AudioTimeStamp* inTimeStamp, 
				   UInt32 inBusNumber, 
				   UInt32 inNumberFrames, 
				   AudioBufferList* ioData ) 
{
    if( inNumberFrames > 8192 ) return noErr;
    
    AudioBufferList* bufs = (AudioBufferList*)g_input_bufs;
    bufs->mNumberBuffers = 1;
    AudioBuffer* buf = &bufs->mBuffers[ 0 ];
    buf->mNumberChannels = g_snd.channels;
    buf->mDataByteSize = inNumberFrames * ( 2 * g_snd.channels );
    buf->mData = g_input_buf_lowlevel;
    
    //Obtain recorded samples:
    OSStatus status = AudioUnitRender(
				      g_unit, 
				      ioActionFlags, 
				      inTimeStamp, 
				      inBusNumber, 
				      inNumberFrames, 
				      bufs );
        
    volatile uint wp = g_input_buf_wp;
    for( int i = 0; i < inNumberFrames; i++ )
    {
	g_input_buf[ wp ] = ((uint*)buf->mData)[ i ];
	wp++;
	wp &= ( INPUT_BUF_FRAMES - 1 );
    }
    g_input_buf_wp = wp;
    
    return noErr;
}

int device_sound_stream_init( void )
{
    OSStatus status;
    
    //
    // Audio session
    //
    
    static bool audio_session_initialized = 0;
    if( audio_session_initialized == 0 )
    {
	status = AudioSessionInitialize( 0, 0, 0, 0 );
	if( status != noErr )
	{
	    dprint( "AUDIO SESSION ERROR: AudioSessionInitialize failed with code %d\n", status );
	}
	audio_session_initialized = 1;
    }
    status = AudioSessionSetActive( true );
    if( status != noErr )
    {
	dprint( "AUDIO SESSION ERROR: AudioSessionSetActive( true ) failed with code %d\n", status );
    }
    UInt32 inputAvailable = 0;
    UInt32 vsize = sizeof( inputAvailable );
    AudioSessionGetProperty( kAudioSessionProperty_AudioInputAvailable, &vsize, &inputAvailable );
    UInt32 v;
    if( inputAvailable && g_au_input )
    {
	v = kAudioSessionCategory_PlayAndRecord;
    }
    else
    {
	v = kAudioSessionCategory_MediaPlayback;
    }
    status = AudioSessionSetProperty( kAudioSessionProperty_AudioCategory, sizeof( v ), &v );
    if( status != noErr )
    {
	dprint( "AUDIO SESSION ERROR: AudioSessionSetProperty (AudioCategory) failed with code %d\n", status );
        if( v == kAudioSessionCategory_PlayAndRecord )
        {
            v = kAudioSessionCategory_MediaPlayback;
            inputAvailable = 0;
            status = AudioSessionSetProperty( kAudioSessionProperty_AudioCategory, sizeof( v ), &v );
            if( status != noErr )
            {
                dprint( "AUDIO SESSION ERROR: AudioSessionSetProperty (AudioCategory for output) failed with code %d\n", status );
            }
        }
    }
    if( v == kAudioSessionCategory_PlayAndRecord )
    {
        UInt32 inputChannels = 0;
        vsize = sizeof( inputChannels );
        status = AudioSessionGetProperty( kAudioSessionProperty_CurrentHardwareInputNumberChannels, &vsize, &inputChannels );
        if( status == kAudioSessionIncompatibleCategory )
        {
            AudioSessionSetActive( false );
            AudioSessionSetActive( true );
            status = AudioSessionGetProperty( kAudioSessionProperty_CurrentHardwareInputNumberChannels, &vsize, &inputChannels );
            if( status == kAudioSessionIncompatibleCategory )
            {
                dprint( "AUDIO SESSION ERROR: kAudioSessionIncompatibleCategory. NO INPUT.\n" );
                inputAvailable = 0;
            }
        }
    }
    v = 1;
    status = AudioSessionSetProperty( kAudioSessionProperty_OverrideCategoryMixWithOthers, sizeof( v ), &v );
    if( status != noErr )
    {
	dprint( "AUDIO SESSION ERROR: AudioSessionSetProperty [2] failed with code %d\n", status );
    }
    
    HANDLE_THREAD_EVENTS;
    
    iphone_sundog_audiobus_init();
    
    //
    // Audio unit
    //
    
    g_au_pause_request = 0;
    
    g_snd.in_type = sound_buffer_int16;
    g_snd.in_channels = g_snd.channels;
    
    g_input_buf_lowlevel = MEM_NEW2( 8192 * 4 );
    g_input_buf = (uint*)MEM_NEW2( INPUT_BUF_FRAMES * 4 );
    g_input_buf2 = MEM_NEW2( 8192 * 4 );
    mem_zero( g_input_buf_lowlevel );
    mem_zero( g_input_buf );
    mem_zero( g_input_buf2 );
    g_input_buf_wp = 0;
    g_input_buf_rp = 0;
    
    if( g_unit == 0 )
    {
	//Describe audio component:
	AudioComponentDescription desc;
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_RemoteIO;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	
	//Get component:
	AudioComponent comp = AudioComponentFindNext( NULL, &desc );
	if( comp == 0 )
	{
	    dprint( "AUDIO UNIT ERROR: AudioComponent not found\n" );
	    return -1;
	}
	
	//Get audio units:
	status = AudioComponentInstanceNew( comp, &g_unit );
	if( status != noErr )
	{
	    dprint( "AUDIO UNIT ERROR: AudioComponentInstanceNew failed with code %d\n", status );
	    g_unit = 0;
	    return -1;
	}
    }
        
    //Enable IO for playback:
    UInt32 flag = 1;
    status = AudioUnitSetProperty(
				  g_unit,
				  kAudioOutputUnitProperty_EnableIO,
				  kAudioUnitScope_Output,
				  kOutputBus,
				  &flag,
				  sizeof( flag ) );
    if( status != noErr )
    {
	dprint( "AUDIO UNIT ERROR: AudioUnitSetProperty (output) failed with code %d\n", status );
	return -2;
    }
    
    bool input = 0;
    
    //Enable IO for recording:
    if( inputAvailable && g_au_input )
    {
	flag = 1;
	input = 1;
    }
    else
    {
	flag = 0;
	input = 0;
    }
    status = AudioUnitSetProperty(
				  g_unit, 
				  kAudioOutputUnitProperty_EnableIO, 
				  kAudioUnitScope_Input, 
				  kInputBus, 
				  &flag, 
				  sizeof( flag ) );
    if( status != noErr )
    {
	dprint( "AUDIO UNIT ERROR: AudioUnitSetProperty (input) failed with code %d\n", status );
	input = 0;
    }
        
    //Describe format:
    AudioStreamBasicDescription af;
    af.mSampleRate = g_snd.freq;
    af.mFormatID = kAudioFormatLinearPCM;
    af.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked;
    af.mFramesPerPacket = 1;
    af.mChannelsPerFrame = g_snd.channels;
    af.mBitsPerChannel = 16;
    af.mBytesPerPacket = 2 * g_snd.channels;
    af.mBytesPerFrame = 2 * g_snd.channels;
    
    //Apply format:
    status = AudioUnitSetProperty(
	g_unit,
	kAudioUnitProperty_StreamFormat,
	kAudioUnitScope_Input,
	kOutputBus,
	&af,
	sizeof( af ) );
    if( status != noErr )
    {
	dprint( "AUDIO UNIT ERROR: AudioUnitSetProperty (output,streamformat) failed with code %d\n", status );
	return -3;
    }
    if( input )
    {
	AudioStreamBasicDescription input_af;
	UInt32 input_af_size = sizeof( input_af );
	status = AudioUnitGetProperty(
				      g_unit,
				      kAudioUnitProperty_StreamFormat,
				      kAudioUnitScope_Output,
				      kInputBus,
				      &input_af,
				      &input_af_size );
	if( status != noErr )
	{
	    dprint( "AUDIO UNIT ERROR: AudioUnitGetProperty (input,streamformat) failed with code %d\n", status );
	}
	else 
	{
	    dprint( "AUDIO INPUT Default Sample Rate: %d\n", (int)input_af.mSampleRate );
	    dprint( "AUDIO INPUT Channels: %d\n", (int)input_af.mChannelsPerFrame );
	    dprint( "AUDIO INPUT Bits Per Channel: %d\n", (int)input_af.mBitsPerChannel );
	}
	status = AudioUnitSetProperty(
				      g_unit,
				      kAudioUnitProperty_StreamFormat,
				      kAudioUnitScope_Output,
				      kInputBus,
				      &af,
				      sizeof( af ) );
	if( status != noErr )
	{
	    dprint( "AUDIO UNIT ERROR: AudioUnitSetProperty (input,streamformat) failed with code %d\n", status );
	    input = 0;
	}
    }
    
    //Set output callback:
    AURenderCallbackStruct cs;
    cs.inputProc = playback_callback;
    cs.inputProcRefCon = g_unit;
    status = AudioUnitSetProperty(
				  g_unit,
				  kAudioUnitProperty_SetRenderCallback,
				  kAudioUnitScope_Global,
				  kOutputBus,
				  &cs,
				  sizeof( cs ) );
    if( status != noErr )
    {
	dprint( "AUDIO UNIT ERROR: AudioUnitSetProperty (output,setrendercallback) failed with code %d\n", status );
	return -4;
    }
    
    //Set input callback:
    if( input )
    {
	AURenderCallbackStruct cs;
	cs.inputProc = recording_callback;
	cs.inputProcRefCon = g_unit;
	status = AudioUnitSetProperty(
				      g_unit, 
				      kAudioOutputUnitProperty_SetInputCallback, 
				      kAudioUnitScope_Global, 
				      kInputBus, 
				      &cs, 
				      sizeof( cs ) );
	if( status != noErr )
	{
	    dprint( "AUDIO UNIT ERROR: AudioUnitSetProperty (input,setinputcallback) failed with code %d\n", status );
	    input = 0;
	}
    }
        
    //Initialize:
    status = AudioUnitInitialize( g_unit );
    if( status != noErr )
    {
	dprint( "AUDIO UNIT ERROR: AudioUnitInitialize failed with code %d\n", status );
	return -5;
    }
    
    status = AudioOutputUnitStart( g_unit );
    if( status != noErr )
    {
	dprint( "AUDIO UNIT ERROR: AudioOutputUnitStart failed with code %d\n", status );
	return -6;
    }
    
    //
    // Disable recording for first time:
    //
    
    HANDLE_THREAD_EVENTS;
    
    return 0;
}

void device_sound_stream_input( bool enable )
{
    g_au_pause_request = 1;
    ticks_t t = time_ticks();
    while( g_au_pause_request == 1 ) 
    { 
	if( ( time_ticks() - t ) > time_ticks_per_second() / 2 ) break;
	time_sleep( 1 );
	HANDLE_THREAD_EVENTS; 
    }

    device_sound_stream_close();

    if( enable )
    {
	g_input_buf_wp = 0;
	g_input_buf_rp = 0;
	g_au_input = 1;
    }
    else 
    {
	g_input_buf_wp = 0;
	g_input_buf_rp = 0;
	g_au_input = 0;
    }
    
    device_sound_stream_init();
}

const UTF8_CHAR* device_sound_stream_get_driver_info( void )
{
    if( (unsigned)g_sound_driver < NUMBER_OF_SDRIVERS )
	return g_sdriver_infos[ g_sound_driver ];
    return 0;
}

const UTF8_CHAR* device_sound_stream_get_driver_name( void )
{
    if( (unsigned)g_sound_driver < NUMBER_OF_SDRIVERS )
	return g_sdriver_names[ g_sound_driver ];
    return 0;
}

int device_sound_stream_get_drivers( UTF8_CHAR*** names, UTF8_CHAR*** infos )
{
    UTF8_CHAR** n = (UTF8_CHAR**)MEM_NEW2( NUMBER_OF_SDRIVERS * sizeof( void* ) );
    UTF8_CHAR** i = (UTF8_CHAR**)MEM_NEW2( NUMBER_OF_SDRIVERS * sizeof( void* ) );
    for( int d = 0; d < NUMBER_OF_SDRIVERS; d++ )
    {
	n[ d ] = (UTF8_CHAR*)MEM_NEW2( mem_strlen( g_sdriver_names[ d ] ) + 1 ); n[ d ][ 0 ] = 0;
	mem_strcat( n[ d ], g_sdriver_names[ d ] );
	i[ d ] = (UTF8_CHAR*)MEM_NEW2( mem_strlen( g_sdriver_infos[ d ] ) + 1 ); i[ d ][ 0 ] = 0;
	mem_strcat( i[ d ], g_sdriver_infos[ d ] );
    }
    *names = n;
    *infos = i;
    return NUMBER_OF_SDRIVERS;
}

int device_sound_stream_get_devices( const UTF8_CHAR* driver, UTF8_CHAR*** names, UTF8_CHAR*** infos, bool input )
{
    return 0;
}

void device_sound_stream_close( void )
{
    if( g_unit )
    {
	AudioOutputUnitStop( g_unit );
	HANDLE_THREAD_EVENTS;
	AudioUnitUninitialize( g_unit );
	HANDLE_THREAD_EVENTS;
        AudioComponentInstanceDispose( g_unit );
	HANDLE_THREAD_EVENTS;
        g_unit = 0;
    }
    mem_free( g_input_buf_lowlevel );
    mem_free( g_input_buf );
    mem_free( g_input_buf2 );
    g_input_buf_lowlevel = 0;
    g_input_buf = 0;
    g_input_buf2 = 0;
    
    OSStatus status = AudioSessionSetActive( false );
    if( status != noErr )
    {
	dprint( "AUDIO SESSION ERROR: AudioSessionSetActive( false ) failed with code %d\n", status );
    }
    
    iphone_sundog_audiobus_deinit();

    HANDLE_THREAD_EVENTS;
}
