/*
    sound_player.cpp. Main sound playing function
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#ifndef NOSOUND

#include "../sound.h"
#include "../../memory/memory.h"

#ifndef PALMOS
    #include <stdio.h>
#endif

int main_sound_output_callback( sound_struct* ss, int flags )
{
    ss->main_sound_callback_working = 1;
    
    int empty = 1;
    
    if( sundog_mutex_lock( &ss->mutex ) ) goto mutex_end;

    for( int slot_num = 0; slot_num < SOUND_SLOTS; slot_num++ )
    {
        sound_slot* slot = &ss->slots[ slot_num ];
        if( slot->event_answer != slot->event )
        {
    	    slot->event_answer = slot->event;
	    switch( slot->event_answer >> 16 )
	    {
	        case 1: //Play:
	    	    slot->status = 1;
	    	    break;
		case 2: //Stop:
		    slot->status = 0;
		    break;
	    }
	}
	if( slot->callback && slot->status )
	{
	    int (*callback)( sound_struct*, int ) = ( int (*)( sound_struct*, int ) )slot->callback;
	    if( empty )
	    {
		if( ss->in_enabled )
		    slot->in_buffer = ss->in_buffer;
		else
		    slot->in_buffer = 0;
		slot->buffer = ss->out_buffer;
		slot->frames = ss->out_frames;
		slot->time = ss->out_time;
		int r = callback( ss, slot_num );
		if( r )
		{
		    empty = 0;
		}
		if( slot->scope_buffer && !slot->scope_buffer_locked )
		{
		    int vb_frames = slot->scope_buffer_frames;
		    if( vb_frames > slot->frames )
			vb_frames = slot->frames;
		    int channels = ss->channels;
		    if( r )
		    {
			for( int c = 0; c < slot->scope_buffer_channels; c++ )
			{
			    SOUND_SCOPE_TYPE* vb = slot->scope_buffer[ c ];
			    if( ss->mode & SOUND_MODE_INT16 )
			    {
				signed short* src = (signed short*)slot->buffer;
    				for( int i = 0, i2 = c; i < vb_frames; i++, i2 += channels )
				{
				    int v = src[ i2 ];
#if CPUMARK >= 10
#else
				    v >>= 8;
#endif
				    vb[ i ] = (SOUND_SCOPE_TYPE)v;
				}
			    }
			    else
			    if( ss->mode & SOUND_MODE_FLOAT32 )
			    {
				float* src = (float*)slot->buffer;
    				for( int i = 0, i2 = c; i < vb_frames; i++, i2 += channels )
				{
				    float v = src[ i2 ];
#if CPUMARK >= 10
				    v *= 32767;
#else
				    v *= 127;
#endif
				    vb[ i ] = (SOUND_SCOPE_TYPE)v;
				}
			    }
			}
		    }
		    else
		    {
			//Empty:
			for( int c = 0; c < slot->scope_buffer_channels; c++ )
			{
			    SOUND_SCOPE_TYPE* vb = slot->scope_buffer[ c ];
#ifdef PALMOS
    			    for( int i = 0; i < vb_frames; i++ ) vb[ i ] = 0;
#else
			    mem_set( vb, vb_frames * sizeof( SOUND_SCOPE_TYPE ), 0 );
#endif
			}
		    }
		    if( slot->scope_buffer_flags & SOUND_SCOPE_FLAG_WRITE_AND_LOCK )
		    {
			slot->scope_buffer_locked = 1;
		    }
		}
	    }
	    else
	    {
		int ptr = 0;
		while( 1 )
		{
		    int size = ss->out_frames - ptr;
		    if( size > ss->slot_buffer_size )
			size = ss->slot_buffer_size;
		    if( ss->in_enabled )
		    {
			if( ss->in_type == sound_buffer_int16 )
			    slot->in_buffer = (char*)ss->in_buffer + ptr * ss->in_channels * 2;
			if( ss->in_type == sound_buffer_float32 )
			    slot->in_buffer = (char*)ss->in_buffer + ptr * ss->in_channels * 4;
		    }
		    else 
		    {
			slot->in_buffer = 0;
		    }
		    slot->buffer = ss->slot_buffer;
		    slot->frames = size;
		    slot->time = ss->out_time;
		    int r = callback( ss, slot_num );
		    if( r )
		    {
			//Add result to the main buffer:
			if( ss->mode & SOUND_MODE_INT16 )
			{
			    signed short* dest = (signed short*)ss->out_buffer;
			    signed short* src = (signed short*)ss->slot_buffer;
			    int add = ptr * ss->channels;
			    for( int i = 0; i < size * ss->channels; i++ )
			    {
				int v = (int)dest[ i + add ] + src[ i ];
				if( v > 32767 ) v = 32767;
				if( v < -32767 ) v = -32767;
				dest[ i + add ] = (signed short)v;
			    }
			}
			else
			if( ss->mode & SOUND_MODE_FLOAT32 )
			{
			    float* dest = (float*)ss->out_buffer;
			    float* src = (float*)ss->slot_buffer;
			    int add = ptr * ss->channels;
			    for( int i = 0; i < size * ss->channels; i++ )
			    {
				dest[ i + add ] += src[ i ];
			    }
			}
		    }
		    ptr += size;
		    if( ptr >= ss->out_frames )
			break;
		}
	    }	    
	}
    }

    sundog_mutex_unlock( &ss->mutex );
    
mutex_end:
    
    if( empty )
    {
	if( ss->mode & SOUND_MODE_INT16 )
	{
	    signed short* ptr = (signed short*)ss->out_buffer;
#ifdef PALMOS
	    for( int i = 0; i < ss->out_frames * ss->channels; i++ ) ptr[ i ] = 0;
#else
	    mem_set( ptr, ss->out_frames * ss->channels * sizeof( short ), 0 );
#endif
	}
	else
	if( ss->mode & SOUND_MODE_FLOAT32 )
	{
	    float* ptr = (float*)ss->out_buffer;
#ifdef PALMOS
	    for( int i = 0; i < ss->out_frames * ss->channels; i++ ) ptr[ i ] = 0;
#else
	    mem_set( ptr, ss->out_frames * ss->channels * sizeof( float ), 0 );
#endif
	}
    }
    
    ss->main_sound_callback_working = 0;
    
    return empty;
}

#endif
