#include <stdio.h>
#include <time.h>
#include <string.h>
#include <pthread.h>

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <ifaddrs.h>
#include <errno.h>

#import <QuartzCore/QuartzCore.h>
#import <OpenGLES/EAGLDrawable.h>

#import <UIKit/UIPasteboard.h>
#import <MobileCoreServices/UTCoreTypes.h>

#include "core/core.h"
#include "sundog_brigde.h"

//Defined in main.cpp:
extern int sundog_main( int argc, char* argv[] );
//Defined in wm_iphone.h:
extern void iphone_sundog_exit_request( void );
extern void iphone_sundog_touches( int x, int y, bool dont_draw, int mode, int touch_id );

int g_iphone_sundog_screen_xsize = 320;
int g_iphone_sundog_screen_ysize = 480;
int g_iphone_sundog_screen_angle = -1;
void* g_iphone_sundog_framebuffer = 0;

extern char* g_osx_docs_path;
extern char* g_osx_caches_path;
extern char* g_osx_tmp_path;
char* g_osx_resources_path = 0;
char* g_iphone_host_addr = 0;
char* g_iphone_host_name = 0;

MyView* g_sundog_view;
CGDataProviderRef g_fb_provider;
CGColorSpaceRef g_fb_colorspace;
CGImageRef g_fb_image;
CGRect g_fb_rect;
EAGLContext* g_gl_context;
GLuint g_view_renderbuffer;
GLuint g_view_framebuffer;

#ifdef USECAMERA
unsigned char g_camera_buffer[ 256 * 256 ];
#endif

typedef struct sundog_info_struct
{
    int initialized;
    pthread_t pth;
    int pth_finished;
} sundog_info_struct;
sundog_info_struct g_sd;

void* sundog_thread( void* arg )
{
    sundog_info_struct* sd = (sundog_info_struct*)arg;

    [ EAGLContext setCurrentContext: g_gl_context ];
    
    sundog_main( 0, 0 );

    sd->pth_finished = 1;
    pthread_exit( NULL );
}

void sundog_init( MyView* view )
{
    int err;

    printf( "sundog_init() ...\n" );

    //Init the view:
    int view_ok = [ view viewInit ];
    if( view_ok ) return;
    
    memset( &g_sd, 0, sizeof( sundog_info_struct ) );
 
    //Save link to current view:
    g_sundog_view = view;
    
    //Get system paths:
    NSArray* paths = NSSearchPathForDirectoriesInDomains( NSDocumentDirectory, NSUserDomainMask, YES );
    const char* docs_path_utf8 = [ [ paths objectAtIndex: 0 ] UTF8String ];
    int len = strlen( docs_path_utf8 );
    g_osx_docs_path = (char*)malloc( len + 2 );
    memcpy( g_osx_docs_path, docs_path_utf8, len + 1 );
    strcat( g_osx_docs_path, "/" );     
    //
    paths = NSSearchPathForDirectoriesInDomains( NSCachesDirectory, NSUserDomainMask, YES );
    const char* caches_path_utf8 = [ [ paths objectAtIndex: 0 ] UTF8String ];
    len = strlen( caches_path_utf8 );
    g_osx_caches_path = (char*)malloc( len + 2 );
    memcpy( g_osx_caches_path, caches_path_utf8, len + 1 );
    strcat( g_osx_caches_path, "/" );
    //
    NSString* tmp = NSTemporaryDirectory();
    const char* tmp_path_utf8 = [ tmp UTF8String ];
    len = strlen( tmp_path_utf8 );
    g_osx_tmp_path = (char*)malloc( len + 2 );
    memcpy( g_osx_tmp_path, tmp_path_utf8, len + 1 );
    strcat( g_osx_tmp_path, "/" );
    //
    const char* res_path_utf8 = [ [ [ NSBundle mainBundle ] resourcePath ] UTF8String ];
    len = strlen( res_path_utf8 );
    g_osx_resources_path = (char*)malloc( len + 2 );
    memcpy( g_osx_resources_path, res_path_utf8, len + 1 );
    strcat( g_osx_resources_path, "/" );
    //
    printf( "%s\n", g_osx_docs_path );
    printf( "%s\n", g_osx_caches_path );
    printf( "%s\n", g_osx_tmp_path );
    printf( "%s\n", g_osx_resources_path );
        
    //Create framebuffer:
#ifdef FRAMEBUFFER
    //CGRect screen_bounds = [ [ UIScreen mainScreen ] applicationFrame ];
    //g_iphone_sundog_screen_xsize = screen_bounds.size.width;
    //g_iphone_sundog_screen_ysize = screen_bounds.size.height;
    g_iphone_sundog_framebuffer = malloc( g_iphone_sundog_screen_xsize * g_iphone_sundog_screen_ysize * 2 );
    if( g_iphone_sundog_framebuffer == 0 )
    {
	printf( "sundog_init(): can't create the framebuffer\n" );
	return;
    }
    g_fb_provider = CGDataProviderCreateWithData( 
	NULL, 
	g_iphone_sundog_framebuffer, 
	g_iphone_sundog_screen_xsize * g_iphone_sundog_screen_ysize * 2,
	NULL );
    g_fb_colorspace = CGColorSpaceCreateDeviceRGB();
    g_fb_rect = CGRectMake( 0, 0, g_iphone_sundog_screen_xsize, g_iphone_sundog_screen_ysize );
#endif

    printf( "Default screen width: %d\n", g_iphone_sundog_screen_xsize );
    printf( "Default screen height: %d\n", g_iphone_sundog_screen_ysize );

    //Create main thread:    
    err = pthread_create( &g_sd.pth, 0, &sundog_thread, (void*)&g_sd );
    if( err == 0 )
    {
	//The pthread_detach() function marks the thread identified by thread as
	//detached. When a detached thread terminates, its resources are 
	//automatically released back to the system.
	err = pthread_detach( g_sd.pth );
	if( err != 0 )
	{
	    printf( "sundog_init(): pthread_detach error %d\n", err );
	    return;
	}
    }
    else
    {
	printf( "sundog_init(): pthread_create error %d\n", err );
	return;
    }
    
    printf( "sundog_init(): done\n" );
    g_sd.initialized = 1;
}

void sundog_deinit( void )
{
    int err;
    
    printf( "sundog_deinit() ...\n" );
    
    if( g_sd.initialized == 0 ) return;

    //Stop the thread:
    iphone_sundog_exit_request();
    {
	g_eventloop_stop_request = 0;
	while( g_eventloop_stop_answer == 1 ) { }
    }
    int timeout_counter = 1000 * 6; //Timeout in milliseconds
    while( timeout_counter > 0 )
    {
	struct timespec delay;
	delay.tv_sec = 0;
	delay.tv_nsec = 1000000 * 20; //20 milliseconds
	if( g_sd.pth_finished ) break;
	nanosleep( &delay, NULL ); //Sleep for delay time
	timeout_counter -= 20;
    }
    if( timeout_counter <= 0 )
    {
	printf( "sundog_deinit(): thread timeout\n" );
	err = pthread_cancel( g_sd.pth );
	if( err != 0 )
	{
	    printf( "sundog_deinit(): pthread_cancel error %d\n", err );
	}
    }
    
    //Free the framebuffer:
#ifdef FRAMEBUFFER
    if( g_iphone_sundog_framebuffer )
	free( g_iphone_sundog_framebuffer );
    CGDataProviderRelease( g_fb_provider );
    CGColorSpaceRelease( g_fb_colorspace );
#endif
    
    //Free the paths and global names:
    if( g_osx_docs_path ) free( g_osx_docs_path );
    if( g_osx_caches_path ) free( g_osx_caches_path );
    if( g_osx_tmp_path ) free( g_osx_tmp_path );
    if( g_osx_resources_path ) free( g_osx_resources_path );
    if( g_iphone_host_addr ) free( g_iphone_host_addr );
    if( g_iphone_host_name ) free( g_iphone_host_name );

    printf( "sundog_deinit(): done\n" );
}

volatile bool g_eventloop_stop_request = 0;
volatile bool g_eventloop_stop_answer = 0;

void iphone_sundog_event_handler( void )
{
    HANDLE_THREAD_EVENTS;
    if( g_eventloop_stop_request )
    {
	g_eventloop_stop_answer = 1;
	while( g_eventloop_stop_request )
	{
	    usleep( 1000 * 100 );
	}
	g_eventloop_stop_answer = 0;
    }
}

void iphone_sundog_screen_angle_handler( int angle )
{
    UIInterfaceOrientation orient;
    switch( angle )
    {
	case 0: orient = UIInterfaceOrientationPortrait; break;
	case 1: orient = UIInterfaceOrientationLandscapeLeft; break;
	case 2: orient = UIInterfaceOrientationPortraitUpsideDown; break;
	case 3: orient = UIInterfaceOrientationLandscapeRight; break;
	default: return;
    }
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    [ [ UIApplication sharedApplication ] setStatusBarOrientation:orient animated:NO ];
    [ pool release ];
}

void iphone_sundog_get_host_info( void )
{
    if( g_iphone_host_addr ) free( g_iphone_host_addr );
    if( g_iphone_host_name ) free( g_iphone_host_name );
    g_iphone_host_addr = 0;
    g_iphone_host_name = 0;
    
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    
    /*NSHost* host = [ NSHost currentHost ];
    if( host )
    {
	NSString* str = [ host address ];
	if( str )
	{
	    const char* addr = [ str UTF8String ];
	    int len = strlen( addr );
	    g_iphone_host_addr = (char*)malloc( len + 1 );
	    memcpy( g_iphone_host_addr, addr, len + 1 );
	}
	str = [ host name ];
	if( str )
	{
	    const char* name = [ str UTF8String ];
	    int len = strlen( name );
	    g_iphone_host_name = (char*)malloc( len + 1 );
	    memcpy( g_iphone_host_name, name, len + 1 );
	}
    }*/
    
    struct ifaddrs* myaddrs;
    struct ifaddrs* ifa;
    struct sockaddr_in* s4;
    struct sockaddr_in6* s6;
    int status;
    char buf[ 128 ];
    buf[ 0 ] = 0;
    
    status = getifaddrs( &myaddrs );
    if( status != 0 )
    {
	perror( "getifaddrs" );
    }
    
    for( ifa = myaddrs; ifa != NULL; ifa = ifa->ifa_next )
    {
	if( ifa->ifa_addr == NULL ) continue;
	if( ( ifa->ifa_flags & IFF_UP ) == 0 ) continue;
	
	if( ifa->ifa_addr->sa_family == AF_INET )
	{
	    s4 = (struct sockaddr_in *)(ifa->ifa_addr);
	    if( inet_ntop( ifa->ifa_addr->sa_family, (void *)&(s4->sin_addr), buf, sizeof( buf ) ) == NULL )
	    {
		printf( "%s: inet_ntop failed!\n", ifa->ifa_name );
	    }
	    else
	    {
		printf( "%s: %s\n", ifa->ifa_name, buf );
		unsigned char* a = (unsigned char*)&(s4->sin_addr);
		if( a[ 0 ] == 192 ) break;
	    }
	}
	else if( ifa->ifa_addr->sa_family == AF_INET6 )
	{
	    s6 = (struct sockaddr_in6 *)(ifa->ifa_addr);
	    if( inet_ntop( ifa->ifa_addr->sa_family, (void *)&(s6->sin6_addr), buf, sizeof( buf ) ) == NULL )
	    {
		printf( "%s: inet_ntop failed!\n", ifa->ifa_name );
	    }
	    else
	    {
		printf( "%s: %s\n", ifa->ifa_name, buf );
	    }
	}
    }
    
    freeifaddrs( myaddrs );
    
    int len = strlen( buf );
    if( len )
    {
	g_iphone_host_addr = (char*)malloc( len + 1 );
	memcpy( g_iphone_host_addr, buf, len + 1 );
    }
    
    if( g_iphone_host_addr )
	printf( "Host addr: %s\n", g_iphone_host_addr );
    if( g_iphone_host_name )
	printf( "Host name: %s\n", g_iphone_host_name );
    
    [ pool release ];
}

// 5MB max per item in iPhone OS clipboard
#define BM_CLIPBOARD_CHUNK_SIZE ( 5 * 1024 * 1024 )

int iphone_sundog_copy( char* filename )
{
    int rv = 0;
    
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    
    while( 1 )
    {
	UIPasteboard* board = [ UIPasteboard generalPasteboard ];
	
	//Open file as binary data
	NSData* dataFile = [ NSData dataWithContentsOfMappedFile:[ NSString stringWithUTF8String:filename ] ];
	if( !dataFile ) 
	{
	    NSLog( @"doCopyButton: Can't open file" );
	    rv = 1;
	    break;
	}
	
	//Create chunked data and append to clipboard
	NSUInteger sz = [ dataFile length ];
	NSUInteger chunkNumbers = ( sz / BM_CLIPBOARD_CHUNK_SIZE ) + 1;
	NSMutableArray* items = [ NSMutableArray arrayWithCapacity:chunkNumbers ];
	NSRange curRange;
	
	for( NSUInteger i = 0; i < chunkNumbers; i++ ) 
	{
	    curRange.location = i * BM_CLIPBOARD_CHUNK_SIZE;
	    curRange.length = MIN( BM_CLIPBOARD_CHUNK_SIZE, sz - curRange.location );
	    NSData* subData = [ dataFile subdataWithRange:curRange ];
	    NSDictionary* dict = [ NSDictionary dictionaryWithObject:subData forKey:(NSString *)kUTTypeAudio ];
	    [ items addObject:dict ];
	}
	
	board.items = items;
	
	break;
    }
    
    [ pool release ];
    
    return rv;
}

char* iphone_sundog_paste( void )
{
    char* rv = 0;
    
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    
    while( 1 )
    {
	UIPasteboard* board = [ UIPasteboard generalPasteboard ];
	
	NSArray* typeArray = [ NSArray arrayWithObject:(NSString *) kUTTypeAudio ];
	NSIndexSet* set = [ board itemSetWithPasteboardTypes:typeArray ];
	if( !set ) 
	{
	    NSLog( @"doPasteButton: Can't get item set" );
	    break;
	}
	
	//Get the subset of kUTTypeAudio elements, and write each chunk to a temporary file:
	NSArray* items = [ board dataForPasteboardType:(NSString *) kUTTypeAudio inItemSet:set ];
	if( items ) 
	{
	    UInt32 cnt = [ items count ];
	    if( !cnt ) 
	    {
		NSLog( @"doPasteButton: Nothing to paste" );
		break;
	    }
	    
	    //Create a file and write each chunks to it:
	    NSString* path = [ NSTemporaryDirectory() stringByAppendingPathComponent:[ NSString stringWithString:@"temp-pasteboard" ] ];
	    if( ! [ [ NSFileManager defaultManager ] createFileAtPath:path contents:nil attributes:nil ] ) 
	    {
		NSLog( @"doPasteButton: Can't create file" );
	    }
	    
	    NSFileHandle* handle = [ NSFileHandle fileHandleForWritingAtPath:path ];
	    if( !handle ) 
	    {
		NSLog( @"doPasteButton: Can't open file for writing" );
		break;
	    }
	    
	    //Write each chunk to file:
	    for( UInt32 i = 0; i < cnt; i++ ) 
	    {
		[ handle writeData:[ items objectAtIndex:i ] ];
	    }
	    [ handle closeFile ];

	    rv = (char*)malloc( strlen( [ path UTF8String ] ) + 1 );
	    rv[ 0 ] = 0;
	    strcat( rv, [ path UTF8String ] );
	}
	
	break;
    }

    [ pool release ];
    
    return rv;
}

void iphone_sundog_open_url( const char* url_text )
{
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    
    NSURL* url = [ NSURL URLWithString:[ NSString stringWithUTF8String:url_text ] ];
    [ [ UIApplication sharedApplication ] openURL:url ];

    [ pool release ];
}

void iphone_sundog_send_mail( const char* email_text, const char* subj_text, const char* body_text )
{
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    
    NSString* mailLinkTemplate = @"mailto:%@?subject=%@&body=%@"; // to subject body
    NSString* body = [ NSString stringWithUTF8String:body_text ];
    
    NSString* bodyEncoded = [ body stringByAddingPercentEscapesUsingEncoding: NSUTF8StringEncoding ];
    NSString* themeEncoded = [ [ NSString stringWithUTF8String:subj_text ] stringByAddingPercentEscapesUsingEncoding: NSUTF8StringEncoding ];
    NSString* emailEncoded = [ [ NSString stringWithUTF8String:email_text ] stringByAddingPercentEscapesUsingEncoding: NSUTF8StringEncoding ];
    
    NSString* eMailURL = [ NSString stringWithFormat: mailLinkTemplate, emailEncoded, themeEncoded, bodyEncoded ];
    [ [ UIApplication sharedApplication ] openURL:[ NSURL URLWithString:eMailURL ] ];
    
    [ pool release ];
}

void iphone_sundog_audiobus_init( void )
{
#ifdef AUDIOBUS_URL
    [ g_sundog_view performSelectorOnMainThread: @selector(audiobusInit) withObject:nil waitUntilDone: YES ];
#endif
}

void iphone_sundog_audiobus_deinit( void )
{
#ifdef AUDIOBUS_URL
    [ g_sundog_view performSelectorOnMainThread: @selector(audiobusDeinit) withObject:nil waitUntilDone: YES ];
#endif
}

void iphone_sundog_screen_redraw( void )
{
    [ g_sundog_view redrawAll ];
}

/*#include <time.h>
int g_sec = -1;
int g_fps = 0;*/

#define DIV 1

@implementation MyView

+ (Class)layerClass
{
    return [ CAEAGLLayer class ];
}

- (BOOL)createFramebuffer
{
    glGenFramebuffersOES( 1, &viewFramebuffer );
    glGenRenderbuffersOES( 1, &viewRenderbuffer );
    
    g_view_renderbuffer = viewFramebuffer;
    g_view_framebuffer = viewRenderbuffer;
    
    glBindFramebufferOES( GL_FRAMEBUFFER_OES, viewFramebuffer );
    glBindRenderbufferOES( GL_RENDERBUFFER_OES, viewRenderbuffer );
    [ context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(id<EAGLDrawable>)self.layer ];
    glFramebufferRenderbufferOES( GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, viewRenderbuffer );
    
    glGetRenderbufferParameterivOES( GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth );
    glGetRenderbufferParameterivOES( GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight );
    g_iphone_sundog_screen_xsize = backingWidth / DIV;
    g_iphone_sundog_screen_ysize = backingHeight / DIV;
    
#ifdef GLZBUF
    glGenRenderbuffersOES( 1, &depthRenderbuffer );
    glBindRenderbufferOES( GL_RENDERBUFFER_OES, depthRenderbuffer );
    glRenderbufferStorageOES( GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, backingWidth, backingHeight );
    glFramebufferRenderbufferOES( GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer );
#endif
    
    if( glCheckFramebufferStatusOES( GL_FRAMEBUFFER_OES ) != GL_FRAMEBUFFER_COMPLETE_OES ) 
    {
	NSLog( @"failed to make complete framebuffer object %x", glCheckFramebufferStatusOES( GL_FRAMEBUFFER_OES ) );
	return NO;
    }
    
    return YES;
}

- (void)orientationChanged: (NSNotification*)aNotification
{
    int angle = 0;
    char* orient = "?";
    switch( [ [ UIDevice currentDevice ] orientation ] )
    {
	case UIDeviceOrientationUnknown: angle = -1; orient = "unknown"; break;
	case UIDeviceOrientationPortrait: angle = 0; orient = "portrait"; break;
	case UIDeviceOrientationLandscapeRight: angle = 1; orient = "landscape right"; break;
	case UIDeviceOrientationPortraitUpsideDown: angle = 2; orient = "portrait upside down"; break;
	case UIDeviceOrientationLandscapeLeft: angle = 3; orient = "landscape left"; break;
	case UIDeviceOrientationFaceUp: angle = -1; orient = "face up"; break;
	case UIDeviceOrientationFaceDown: angle = -1; orient = "face down"; break;
    }
    if( angle >= 0 )
    {
	g_iphone_sundog_screen_angle = angle;
    }
    printf( "ROTATE %d (%s)\n", angle, orient );
}

- (int)viewInit
{
    //Enable multitouch:
    self.multipleTouchEnabled = YES;
    
    //Disable high-resolution:
    SEL sel = NSSelectorFromString( @"contentScaleFactor" );
    if( sel )
    {
	if( [ self respondsToSelector:sel ] ) 
	{
	    self.contentScaleFactor = 1.0;
	}
	else 
	{
	    // Does not implement selector
	}
    }
    
    //Get the layer
    CAEAGLLayer* eaglLayer = (CAEAGLLayer*) self.layer;
        
    eaglLayer.opaque = YES;
    eaglLayer.drawableProperties = [ 
	NSDictionary dictionaryWithObjectsAndKeys:
#ifdef GLNORETAIN
	[ NSNumber numberWithBool:FALSE ], 
#else
	[ NSNumber numberWithBool:TRUE ], 
#endif
	kEAGLDrawablePropertyRetainedBacking, 
	kEAGLColorFormatRGB565, 
	kEAGLDrawablePropertyColorFormat, nil
    ];
    
    context = [ [ EAGLContext alloc ] initWithAPI:kEAGLRenderingAPIOpenGLES1 ];
    g_gl_context = context;
    
    if( !context || ![EAGLContext setCurrentContext:context] || ![self createFramebuffer] ) 
    {
	[ self release ];
	return -1;
    }
    
    [ [ UIDevice currentDevice ] beginGeneratingDeviceOrientationNotifications ];
    NSNotificationCenter* nc;
    nc = [ NSNotificationCenter defaultCenter ];
    [ nc addObserver:self selector:@selector( orientationChanged: ) name:UIDeviceOrientationDidChangeNotification object:nil ];    
    
    //Clean touches
    int i;
    for( i = 0; i < MAX_TOUCHES; i++ )
    {
	m_touches[ i ] = 0;
    } 
    touches_num = 0;
        
    return 0;
}

- (void)drawRect:(CGRect)rect
{    
    /*CGContextRef context = UIGraphicsGetCurrentContext();
    g_fb_image = CGImageCreate(
	g_iphone_sundog_screen_xsize,
	g_iphone_sundog_screen_ysize,
	5,
	16,
	2 * g_iphone_sundog_screen_xsize,
	g_fb_colorspace,
	kCGBitmapByteOrder16Little,
	g_fb_provider,
	NULL,
	NO,
	kCGRenderingIntentDefault
    );
    CGContextDrawImage( context, g_fb_rect, g_fb_image );
    CGImageRelease( g_fb_image );*/
    
    /*g_fps++;
    time_t t;
    time( &t );
    int sec = localtime( &t )->tm_sec;
    if( sec != g_sec )
    {
	printf( "FPS: %d\n", g_fps );
	g_fps = 0;
	g_sec = sec;
    }*/
}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
    NSEnumerator* e = [ touches objectEnumerator ];
    UITouch* touch;
    int i, i2;
    while( ( touch = [ e nextObject ] ) )
    {
	//Serch for free place:
	for( i = 0; i < MAX_TOUCHES; i++ )
	{
	    if( m_touches[ i ] == 0 )
		break;
	}
	if( i == MAX_TOUCHES )
	{
	    //No free place for new touch:
	    return;
	}
	
	//Make it busy:
	m_touches[ i ] = touch;
	
	//Get current number of busy places:
	touches_num = 0;
	for( i2 = 0; i2 < MAX_TOUCHES; i2++ )
	{
	    if( m_touches[ i2 ] )
		touches_num = i2 + 1;
	}
	
	//Send event:
	CGPoint pos = [ touch locationInView: self ];
	iphone_sundog_touches( (int)pos.x / DIV, (int)pos.y / DIV, 0, 0, i );
    }
    iphone_sundog_touches( 0, 0, 0, -1, 0 ); //Send buffer with events
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
    int num = [ touches count ];
    NSEnumerator* e = [ touches objectEnumerator ];
    UITouch* touch;
    int i;
    int touch_count = 0;
    while( ( touch = [ e nextObject ] ) )
    {
	for( i = 0; i < touches_num; i++ )
	{
	    if( m_touches[ i ] == touch )
		break;
	}
	if( i == touches_num )
	{
	    //Touch not found!
	    return;
	}
	
	//Send event:
	int dont_draw = 0;
	if( touch_count < num - 1 )
	    dont_draw = 1;
	CGPoint pos = [ touch locationInView: self ];
	iphone_sundog_touches( (int)pos.x / DIV, (int)pos.y / DIV, dont_draw, 1, i );
	
	touch_count++;
    }
    iphone_sundog_touches( 0, 0, 0, -1, 0 ); //Send buffer with events
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
    NSEnumerator* e = [ touches objectEnumerator ];
    UITouch* touch;
    int i, i2;
    while( ( touch = [ e nextObject ] ) )
    {
	for( i = 0; i < touches_num; i++ )
	{
	    if( m_touches[ i ] == touch )
		break;
	}
	if( i == touches_num )
	{
	    //Touch not found!
	    return;
	}
	m_touches[ i ] = 0;

	//Get current number of busy places:
	touches_num = 0;
	for( i2 = 0; i2 < MAX_TOUCHES; i2++ )
	{
	    if( m_touches[ i2 ] )
		touches_num = i2 + 1;
	}	
	
	//Send event:
	CGPoint pos = [ touch locationInView: self ];
	iphone_sundog_touches( (int)pos.x / DIV, (int)pos.y / DIV, 0, 2, i );
    }
    iphone_sundog_touches( 0, 0, 0, -1, 0 ); //Send buffer with events
}

- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
    NSEnumerator* e = [ touches objectEnumerator ];
    UITouch* touch;
    int i, i2;
    while( ( touch = [ e nextObject ] ) )
    {
	for( i = 0; i < touches_num; i++ )
	{
	    if( m_touches[ i ] == touch )
		break;
	}
	if( i == touches_num )
	{
	    //Touch not found!
	    return;
	}
	m_touches[ i ] = 0;

	//Get current number of busy places:
	touches_num = 0;
	for( i2 = 0; i2 < MAX_TOUCHES; i2++ )
	{
	    if( m_touches[ i2 ] )
		touches_num = i2 + 1;
	}	
	
    	//Send event:
	CGPoint pos = [ touch locationInView: self ];
	iphone_sundog_touches( (int)pos.x / DIV, (int)pos.y / DIV, 0, 2, i );
    }
    iphone_sundog_touches( 0, 0, 0, -1, 0 ); //Send buffer with events
}

CGImageRef UIGetScreenImage();

- (void)redrawAll
{
    //Invoke setNeedsDisplay method on the main thread:
    //[ self performSelectorOnMainThread: @selector(setNeedsDisplay) withObject:nil waitUntilDone: NO ];
    
#ifdef USECAMERA
    NSAutoreleasePool* pool = [ [ NSAutoreleasePool alloc ] init ];
    /*UIGraphicsBeginImageContext( [[UIScreen mainScreen] applicationFrame].size );
    CGContextClearRect( UIGraphicsGetCurrentContext(), [[UIScreen mainScreen] applicationFrame] );
    UIGraphicsEndImageContext();*/
    /*CGRect rect;
    rect.origin.x = 0;
    rect.origin.y = 0;
    rect.size.width = 256;
    rect.size.height = 256;
    UIGraphicsBeginImageContext( rect.size );
    [ g_sundog_view.layer renderInContext: UIGraphicsGetCurrentContext() ];
    UIImage* scr = UIGraphicsGetImageFromCurrentImageContext();
    NSData* pixels = (NSData*)CGDataProviderCopyData( CGImageGetDataProvider( scr.CGImage ) );
    memcpy( g_camera_buffer, [ pixels bytes ], 256 * 256 * 2 );
    [ pixels release ];
    UIGraphicsEndImageContext();*/
    
    CGImageRef scr = UIGetScreenImage();
    int img_xsize = CGImageGetWidth( scr );
    int img_ysize = CGImageGetHeight( scr );
    int img_bytes_per_pixel = CGImageGetBitsPerPixel( scr ) / 8;
    int img_bytes_per_row = CGImageGetBytesPerRow( scr );
    NSData* pixels = (NSData*)CGDataProviderCopyData( CGImageGetDataProvider( scr ) );
    int offset = ( ( img_ysize - 256 ) / 2 ) * img_bytes_per_row;
    offset += ( ( img_xsize - 256 ) / 2 ) * img_bytes_per_pixel;
    unsigned char* img_ptr = (unsigned char*)[ pixels bytes ];
    int x, y, p = 0;
    for( y = 0; y < 256; y++ )
    {
	for( x = 0; x < 256; x++ )
	{
	    g_camera_buffer[ p ] = img_ptr[ offset ];
	    p++;
	    offset += img_bytes_per_pixel;
	}
	offset += img_bytes_per_row - 256 * img_bytes_per_pixel;
    }
    [ pixels release ];
    CGImageRelease( scr );
    
    [ pool release ];
    //glReadPixels( 0, 256, 256, 256, GL_RGB, GL_UNSIGNED_SHORT_5_6_5, g_camera_buffer );    
#endif
    glBindRenderbufferOES( GL_RENDERBUFFER_OES, viewRenderbuffer );
    [ context presentRenderbuffer:GL_RENDERBUFFER_OES ];
    glBindFramebufferOES( GL_FRAMEBUFFER_OES, viewFramebuffer );
}

- (void)audiobusInit
{
#ifdef AUDIOBUS_URL
    audiobusController = [ [ ABAudiobusController alloc ] initWithAppLaunchURL:[ NSURL URLWithString:@AUDIOBUS_URL ] apiKey:@AUDIOBUS_API_KEY ];
#endif
}

- (void)audiobusDeinit
{
#ifdef AUDIOBUS_URL
    [ audiobusController release ];
    audiobusController = 0;
#endif
}

@end
