/*
    handlers_dialog.h.
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

dialog_item* DIALOG_ITEMS = 0;
const UTF8_CHAR* DIALOG_BUTTONS_TEXT = 0;
const UTF8_CHAR* DIALOG_TEXT = 0;
int* DIALOG_RESULT = 0;

struct dialog_data
{
    WINDOWPTR this_window;
    const UTF8_CHAR* text;
    int* result;
    UTF8_CHAR** buttons_text;
    WINDOWPTR* buttons;
    int buttons_num;
    dialog_item* items;
    int timer;
    ticks_t timer_start;
    COLOR base_color;
};

int dialog_item_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    dialog_data* data = (dialog_data*)user_data;
    if( data->items )
    {
	int i = 0;
	while( 1 )
	{
	    if( data->items[ i ].type == DIALOG_ITEM_NONE )
		break;
		
	    if( data->items[ i ].win == win )
	    {
		switch( data->items[ i ].type )
		{
		    case DIALOG_ITEM_NUMBER:
		    case DIALOG_ITEM_NUMBER_HEX:
			data->items[ i ].int_val = text_get_value( win, wm );
			break;
		    case DIALOG_ITEM_TEXT:
			{
			    if( data->items[ i ].str_val )
				mem_free( data->items[ i ].str_val );
			    UTF8_CHAR* str = text_get_text( win, wm );
			    if( str )
			    {
				int len = mem_get_size( str );
				data->items[ i ].str_val = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, len );
				mem_copy( data->items[ i ].str_val, str, len );
			    }
			    else
			    {
				data->items[ i ].str_val = 0;
			    }
			}
			break;
		    case DIALOG_ITEM_POPUP:
			data->items[ i ].int_val = button_get_menu_val( win, wm );
			break;
		}
		break;
	    }
	    
	    i++;
	}
    }
    return 0;
}

int dialog_button_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    dialog_data* data = (dialog_data*)user_data;
    if( data->result ) 
    {
	int i;
	for( i = 0; i < data->buttons_num; i++ )
	{
	    if( data->buttons[ i ] == win )
	    {
		*data->result = i;
		break;
	    }
	}
	if( i == data->buttons_num )
	{
	    *data->result = -1;
	}
    }
    remove_window( data->this_window, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

void dialog_timer( void* user_data, sundog_timer* timer, window_manager* wm )
{
    dialog_data* data = (dialog_data*)user_data;
    ticks_t t = time_ticks() - data->timer_start;
    t = ( t * 256 ) / time_ticks_per_second() + 256 + 128;
    int v = g_hsin_tab[ t & 255 ];
    if( t & 256 ) v = -v;
    v = v / 2 + 128;
    v /= 2;
    data->this_window->color = blend( data->base_color, wm->color2, v );
    for( int i = 0; i < data->buttons_num; i++ )
    {
	data->buttons[ i ]->color = blend( wm->button_color, wm->color2, v );
    }
    draw_window( data->this_window->parent, wm );
}

int dialog_handler( sundog_event* evt, window_manager* wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    dialog_data* data = (dialog_data*)win->data;
    int but_ysize = BUTTON_YSIZE;
    int but_xsize = BUTTON_XSIZE;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( dialog_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		data->text = DIALOG_TEXT;
		data->result = DIALOG_RESULT;
		data->items = DIALOG_ITEMS;
		
		data->timer = -1;
		if( data->text && data->text[ 0 ] == '!' )
		{
		    data->text++;
		    data->timer = add_timer( dialog_timer, data, 0, wm );
		    data->timer_start = time_ticks();
		    data->base_color = win->color;
		}

		if( data->items )
		{
		    int i = 0;
		    int y = INTERELEMENT_SPACE;
		    while( 1 )
		    {
			dialog_item* item = &data->items[ i ];
			
			if( item->type == DIALOG_ITEM_NONE )
			    break;
			
			WINDOWPTR w = 0;
			switch( item->type )
			{
			    case DIALOG_ITEM_NUMBER:
			    case DIALOG_ITEM_NUMBER_HEX:
				if( item->type == DIALOG_ITEM_NUMBER )
				    CREATE_NUMERIC_TEXT = 1;
				else
				    CREATE_NUMERIC_TEXT = 2;
				CREATE_NUMERIC_TEXT_MIN = item->min;
				CREATE_NUMERIC_TEXT_MAX = item->max;
				TEXT_CALL_HANDLER_ON_ANY_CHANGES = 1;
				w = new_window( "d.item.num", 0, y, 8, DIALOG_ITEM_NUMBER_SIZE, wm->text_background, win, text_handler, wm );
				text_set_value( w, item->int_val, wm );
				y += DIALOG_ITEM_NUMBER_SIZE + INTERELEMENT_SPACE;
				break;
			    case DIALOG_ITEM_TEXT:
				TEXT_CALL_HANDLER_ON_ANY_CHANGES = 1;
				w = new_window( "d.item.txt", 0, y, 8, DIALOG_ITEM_TEXT_SIZE, wm->text_background, win, text_handler, wm );
				if( item->str_val )
				    text_set_text( w, (const UTF8_CHAR*)item->str_val, wm );
				y += DIALOG_ITEM_TEXT_SIZE + INTERELEMENT_SPACE;
				break;
			    case DIALOG_ITEM_LABEL:
				w = new_window( (const UTF8_CHAR*)item->str_val, 0, y, 8, DIALOG_ITEM_LABEL_SIZE, win->color, win, label_handler, wm );
				y += DIALOG_ITEM_LABEL_SIZE + INTERELEMENT_SPACE;
				break;
			    case DIALOG_ITEM_POPUP:
				CREATE_BUTTON_WITH_TEXT_MENU = item->menu;
				CREATE_FLAT_BUTTON = 1;
				BUTTON_FLAGS = BUTTON_FLAG_SHOW_VALUE;
				w = new_window( (const UTF8_CHAR*)item->str_val, 0, y, 8, DIALOG_ITEM_POPUP_SIZE, wm->button_color, win, button_handler, wm );
				button_set_menu_val( w, item->int_val, wm );
				y += DIALOG_ITEM_POPUP_SIZE + INTERELEMENT_SPACE;
				break;
			}
			set_window_controller( w, 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
			set_window_controller( w, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
			set_handler( w, dialog_item_handler, data, wm );
			item->win = w;
			    
			i++;
		    }
		}
		
		data->buttons_text = 0;
		if( DIALOG_BUTTONS_TEXT )
		{
		    int buttons_num = 1;
		    int i, p, len;
		    int slen = mem_strlen( DIALOG_BUTTONS_TEXT );
		    for( i = 0; i < slen; i++ )
		    {
			if( DIALOG_BUTTONS_TEXT[ i ] == ';' )
			    buttons_num++;
		    }
		    data->buttons_text = (UTF8_CHAR**)MEM_NEW( HEAP_DYNAMIC, sizeof( UTF8_CHAR* ) * buttons_num );
		    data->buttons = (WINDOWPTR*)MEM_NEW( HEAP_DYNAMIC, sizeof( WINDOWPTR ) * buttons_num );
		    data->buttons_num = buttons_num;
		    mem_set( data->buttons_text, sizeof( UTF8_CHAR* ) * buttons_num, 0 );
		    int word_start = 0;
		    for( i = 0; i < buttons_num; i++ )
		    {
			len = 0;
			while( 1 )
			{
			    if( DIALOG_BUTTONS_TEXT[ word_start + len ] == 0 || DIALOG_BUTTONS_TEXT[ word_start + len ] == ';' )
			    {
				break;
			    }
			    len++;
			}
			if( len > 0 )
			{
			    data->buttons_text[ i ] = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, len + 1 );
			    for( p = word_start; p < word_start + len; p++ )
			    {
				data->buttons_text[ i ][ p - word_start ] = DIALOG_BUTTONS_TEXT[ p ];
			    }
			    data->buttons_text[ i ][ p - word_start ] = 0;
			}
			word_start += len + 1;
			if( word_start >= slen ) break;
		    }
		    
		    //Create buttons:
		    int xoff = 0;
		    for( i = 0; i < buttons_num; i++ )
		    {
			data->buttons[ i ] = new_window( data->buttons_text[ i ], 0, 0, 10, 10, wm->button_color, win, button_handler, wm );
			set_handler( data->buttons[ i ], dialog_button_handler, data, wm );
			set_window_controller( data->buttons[ i ], 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
			set_window_controller( data->buttons[ i ], 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );
			if( buttons_num <= 2 )
			{
			    set_window_controller( data->buttons[ i ], 0, wm, (WCMD)xoff + INTERELEMENT_SPACE, CEND );
			    set_window_controller( data->buttons[ i ], 2, wm, (WCMD)xoff + INTERELEMENT_SPACE + but_xsize, CEND );
			}
			else
			{
			    if( i == 0 )
				set_window_controller( data->buttons[ i ], 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
			    else
				set_window_controller( data->buttons[ i ], 0, wm, CPERC, (WCMD)( i * 100 ) / buttons_num, CADD, (WCMD)1, CEND );
			    if( i == buttons_num - 1 )
				set_window_controller( data->buttons[ i ], 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
			    else
				set_window_controller( data->buttons[ i ], 2, wm, CPERC, (WCMD)( ( i + 1 ) * 100 ) / buttons_num, CEND );
			}
			xoff += but_xsize + 1;
		    }
		}

		DIALOG_ITEMS = 0;
		DIALOG_BUTTONS_TEXT = 0;
		DIALOG_TEXT = 0;
		DIALOG_RESULT = 0;

		//SET FOCUS:
		set_focus_win( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    if( data->buttons_text )
	    {
		for( int i = 0; i < data->buttons_num; i++ )
		{
		    if( data->buttons_text[ i ] )
		    {
			mem_free( data->buttons_text[ i ] );
		    }
		}
		mem_free( data->buttons_text );
	    }
	    if( data->buttons )
	    {
		mem_free( data->buttons );
	    }
	    remove_timer( data->timer, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    win_draw_lock( win, wm );
	    if( data->items )
	    {
		win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
		if( data->text )
		    win_draw_string( win, data->text, INTERELEMENT_SPACE, INTERELEMENT_SPACE, wm->color3, win->color, wm );
	    }
	    else 
	    {
		wbd_lock( win );
		draw_box( 0, 0, win->xsize, win->ysize, win->color, wm );
		if( data->text )
		{
		    wm->cur_font_color = wm->color3;
		    draw_string( data->text, INTERELEMENT_SPACE, INTERELEMENT_SPACE, wm );
		}
		wbd_draw();
		wbd_unlock();
	    }
	    win_draw_unlock( win, wm );
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    if( evt->key == KEY_ESCAPE )
	    {
		dialog_button_handler( data, 0, wm );
		retval = 1;
	    }
	    if( evt->key == 'y' || evt->key == KEY_ENTER )
	    {
		if( data->buttons_num < 3 )
		{
		    dialog_button_handler( data, data->buttons[ 0 ], wm );
		    retval = 1;
		}
	    }
	    if( evt->key == 'n' )
	    {
		if( data->buttons_num < 3 )
		{
		    dialog_button_handler( data, data->buttons[ 1 ], wm );
		    retval = 1;
		}
	    }
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}
