/*
    handlers_popup.h.
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#define POPUP_ITEM_YSIZE 	wm->list_item_ysize
#define POPUP_ITEM_FLAG_SHADED	1
#define POPUP_ITEM_FLAG_BOLD	2

struct popup_data
{
    WINDOWPTR this_window;
    UTF8_CHAR* text;
    int16 lines[ 128 ];
    uchar line_flags[ 128 ];
    int lines_num;
    char cols; //Columns
    int16 col_xsize[ 8 ];
    int16 col_ysize[ 8 ];
    int current_selected;
    WINDOWPTR prev_focus;
    int* exit_flag;
    int* result;
};

const UTF8_CHAR* CREATE_TEXT_POPUP = 0;
int* SAVE_POPUP_EXIT_TO = 0;
int* SAVE_POPUP_RESULT_TO = 0;

int popup_handler( sundog_event *evt, window_manager *wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    popup_data* data = (popup_data*)win->data;
    int popup_border = 2;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( popup_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		//Set focus:
		data->prev_focus = wm->focus_win;
		set_focus_win( win, wm );

		//Data init:
		data->this_window = win;
	    	data->text = 0;
		data->current_selected = -1;
		win->action_result = -1;
		mem_set( data->line_flags, sizeof( data->line_flags ), 0 );

		//Set window size:
		int xsize = 0;
		int ysize = 0;
		int x = win->x;
		int y = win->y;
		int name_xsize = 0;
		int name_ysize = 0;
		if( win->name && win->name[ 0 ] != 0 )
		{
		    name_xsize = string_size( win, win->name, wm ) + 1;
		    name_ysize = POPUP_ITEM_YSIZE;
		}
		data->cols = 1;
		data->col_xsize[ 0 ] = 0;
		data->col_ysize[ 0 ] = name_ysize;
		if( CREATE_TEXT_POPUP )
		{
		    //Text popup:
		    data->text = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( CREATE_TEXT_POPUP ) + 1 );
		    mem_copy( data->text, CREATE_TEXT_POPUP, mem_strlen( CREATE_TEXT_POPUP ) + 1 );
		    data->lines_num = 0;
		    int cur_x_size = 0;
		    data->lines[ 0 ] = 0;
		    int nl_code = 0;
		    int text_size = mem_strlen( data->text );
		    for( int i = 0; i <= text_size; i++ )
		    {
			UTF8_CHAR c = data->text[ i ];
			if( c == '@' )
			{
			    data->line_flags[ data->lines_num ] = POPUP_ITEM_FLAG_SHADED;
			    data->text[ i ] = 0; 
			    if( data->text[ i + 1 ] == '@' )
			    {
				data->line_flags[ data->lines_num ] = POPUP_ITEM_FLAG_BOLD;
				i++;
				data->text[ i ] = 0;
			    }
			}
			else
			{
			    if( c == 0xA || c == 0 ) 
			    { 
				data->text[ i ] = 0;
				data->lines_num++;
				data->col_ysize[ data->cols - 1 ] += POPUP_ITEM_YSIZE;
				if( data->col_ysize[ data->cols - 1 ] > wm->screen_ysize - popup_border * 2 - POPUP_ITEM_YSIZE )
				{
				    data->col_xsize[ data->cols ] = 0;
				    data->col_ysize[ data->cols ] = name_ysize;
				    data->cols++;
				}
				cur_x_size = 0; 
				nl_code = 1; 
			    }
			    else
			    {
				if( c != 0xA && c != 0xD ) 
				{
				    if( nl_code == 1 ) data->lines[ data->lines_num ] = i;
				    nl_code = 0;
				    cur_x_size += char_x_size( win, c, wm );
				    if( cur_x_size > data->col_xsize[ data->cols - 1 ] ) data->col_xsize[ data->cols - 1 ] = cur_x_size + 1;
				}
				else
				{
				    data->text[ i ] = 0;
				    nl_code = 1;
				}
			    }
			}
		    }
		    for( int i = 0; i < data->cols; i++ ) 
		    {
			xsize += data->col_xsize[ i ];
		    }
		    if( xsize < name_xsize ) 
		    {
			data->col_xsize[ data->cols - 1 ] += name_xsize - xsize;
			xsize = name_xsize;
		    }
		    xsize += ( data->cols - 1 );
		    ysize += data->col_ysize[ 0 ];
		}
		xsize += popup_border * 2;
		ysize += popup_border * 2;

		//Control window position:
		if( x + xsize > win->parent->xsize && x > 0 ) x -= ( x + xsize ) - win->parent->xsize;
		if( y + ysize > win->parent->ysize && y > 0 ) y -= ( y + ysize ) - win->parent->ysize;
		if( x < 0 ) x = 0;
		if( y < 0 ) y = 0;

		win->x = x;
		win->y = y;
		win->xsize = xsize;
		win->ysize = ysize;

		data->exit_flag = SAVE_POPUP_EXIT_TO;
		data->result = SAVE_POPUP_RESULT_TO;

		CREATE_TEXT_POPUP = 0;
		SAVE_POPUP_EXIT_TO = 0;
		SAVE_POPUP_RESULT_TO = 0;
    	    
		retval = 1;
	    }
	    break;
	case EVT_BEFORECLOSE:
	    if( data->text ) mem_free( data->text );
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    {
		int prev_sel = data->current_selected;
		data->current_selected = -1;
		int cur_y = popup_border;
		int name_ysize = 0;
		if( win->name && win->name[ 0 ] != 0 )
		{
		    name_ysize = POPUP_ITEM_YSIZE;
		}
		cur_y += name_ysize;
		if( data->text && data->lines_num > 0 )
		{
		    int col_x = popup_border;
		    int line = 0;
		    for( int i = 0; i < data->cols; i++ )
		    {
			int xsize = data->col_xsize[ i ];
			int lines = ( win->ysize - name_ysize - popup_border * 2 ) / POPUP_ITEM_YSIZE;
			if( line + lines > data->lines_num ) lines = data->lines_num - line;
			if( rx >= col_x && rx < col_x + data->col_xsize[ i ] )
			{
			    if( ry >= cur_y && ry < cur_y + lines * POPUP_ITEM_YSIZE )
			    {
				data->current_selected = line + ( ry - cur_y ) / POPUP_ITEM_YSIZE;
				break;
			    }
			}
			col_x += xsize + 1;
			line += lines;
		    }
		}
		if( prev_sel != data->current_selected )
		    draw_window( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( data->current_selected >= 0 && evt->key == MOUSE_BUTTON_LEFT )
	    {
		//Successful selection:
		//win->action_result = data->current_selected;
		if( data->result )
		    *data->result = data->current_selected;
		set_focus_win( data->prev_focus, wm ); // -> EVT_UNFOCUS -> remove_window()
	    }
	    retval = 1;
	    break;
	case EVT_UNFOCUS:
	    if( data->exit_flag ) 
		*data->exit_flag = 1;
	    remove_window( win, wm );
	    recalc_regions( wm );
	    draw_window( wm->root_win, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    win_draw_lock( win, wm );
	    wbd_lock( win );
	    wm->cur_transparency = 255;
	    {
		int cur_y = popup_border;
		int cur_x = popup_border;
		draw_box( 0, 0, win->xsize, win->ysize, win->color, wm );

		//Draw name:
		int name_ysize = 0;
		if( win->name && win->name[ 0 ] != 0 )
		{
		    name_ysize = POPUP_ITEM_YSIZE;
		    wm->cur_font_color = wm->header_text_color;
		    draw_string( 
			win->name, 
			cur_x, 
			cur_y + ( POPUP_ITEM_YSIZE - char_y_size( win, wm ) ) / 2,
			wm );
		    cur_y += name_ysize;
		}

		//Draw text:
		if( data->text && data->lines_num > 0 )
		{
		    int col = 0;
		    int xsize = data->col_xsize[ col ];
		    int ysize = data->col_ysize[ col ];
		    int y = 0;
		    for( int i = 0; i < data->lines_num; i++ )
		    {
			UTF8_CHAR* text = data->text + data->lines[ i ];
			int text_y = cur_y + ( POPUP_ITEM_YSIZE - char_y_size( win, wm ) ) / 2;
			if( i == data->current_selected )
			{
			    draw_box( 
				cur_x, 
				cur_y, 
				xsize, 
				POPUP_ITEM_YSIZE, 
				blend( wm->color2, wm->color1, 32 ),
				wm );
			    wm->cur_font_color = wm->color0;
			    draw_string( text, cur_x, text_y, wm );
			}
			else
			{
			    COLOR new_color = win->color;
			    if( y & 1 )
			    {
				new_color = blend( new_color, wm->color3, 16 );
			    }
			    if( data->line_flags[ i ] & POPUP_ITEM_FLAG_SHADED )
			    {
				new_color = blend( new_color, wm->color2, 90 );
			    }
			    wm->cur_font_color = wm->color3;
			    if( new_color != win->color )
			    {
				draw_box( 
				    cur_x, 
				    cur_y, 
				    xsize, 
				    POPUP_ITEM_YSIZE, 
				    new_color,
				    wm );
			    }
			    draw_string( text, cur_x, text_y, wm );
			}
			cur_y += POPUP_ITEM_YSIZE;
			y++;
			if( cur_y - popup_border >= ysize )
			{
			    //Go to the next column:
			    y = 0;
			    cur_y = popup_border + name_ysize;
			    cur_x += xsize + 1;
			    col++;
			    xsize = data->col_xsize[ col ];
			}
		    }
		}

		draw_frame( 0, 0, win->xsize - 1, win->ysize - 1, BORDER_COLOR( win->color ), wm );

		retval = 1;
	    }
	    wbd_draw();
	    wbd_unlock();
	    win_draw_unlock( win, wm );
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}
