/*
    handlers_preferences.cpp
    This file is part of the SunDog engine.
    Copyright (C) 2011 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../../main/user_code.h"
#include "../../filesystem/v3nus_fs.h"
#include "../../time/timemanager.h"
#include "../../utils/utils.h"
#include "../wmanager.h"

WINDOWPTR g_prefs = 0;
bool g_prefs_restart_request = 0;

const UTF8_CHAR* g_prefs_section_names[ 32 ];
int g_prefs_sections = 0;
void* g_prefs_section_handlers[ 32 ]; 
int g_pref_section_ysize = 200;

int g_prefs_flags = 0;

struct prefs_data
{
    WINDOWPTR this_window;
    WINDOWPTR close;
    WINDOWPTR sections;
    int list_xsize;
    int cur_section;
    WINDOWPTR cur_section_window;
    
    int correct_ysize;
};

int prefs_close_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    if( g_prefs )
    {
	hide_window( g_prefs, wm );
	recalc_regions( wm );
	draw_window( wm->root_win, wm );
    }
    
    if( g_prefs_restart_request )
    {
	if( dialog( "!Some preferences changed.\nYou must restart the program\nto apply these changes.\nRestart now?", "Yes;No" ) == 0 )
	{
	    wm->exit_request = 1;
	    wm->restart_request = 1;
	}
    }
    
    return 0;
}

int prefs_sections_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_data* data = (prefs_data*)user_data;
    list_data* ldata = list_get_data( data->sections, wm );
    
    if( ldata )
    {
	if( list_get_last_action( win, wm ) == LIST_ACTION_ESCAPE )
	{
	    //ESCAPE KEY:
	    prefs_close_handler( data, 0, wm );
	    return 0;
	}
	int sel = list_get_selected_num( ldata );
	if( (unsigned)sel < (unsigned)g_prefs_sections )
	{
	    if( sel != data->cur_section )
	    {
		data->cur_section = sel;
		//Close old section:
		remove_window( data->cur_section_window, wm );
		//Open new section:
		data->cur_section_window = new_window( "Section", 0, 0, 1, 1, data->this_window->color, data->this_window, (int(*)(sundog_event*,window_manager*)) g_prefs_section_handlers[ data->cur_section ], wm );
		set_window_controller( data->cur_section_window, 0, wm, (WCMD)INTERELEMENT_SPACE * 2 + data->list_xsize, CEND );
		set_window_controller( data->cur_section_window, 1, wm, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->cur_section_window, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->cur_section_window, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		show_window( data->cur_section_window, wm );
		bring_to_front( data->close, wm );
		//Resize prefs window:
		int new_ysize = g_pref_section_ysize + INTERELEMENT_SPACE * 3 + BUTTON_YSIZE;
		if( new_ysize < g_wm->large_window_ysize )
		    new_ysize = g_wm->large_window_ysize;
		resize_window_with_decorator( data->this_window, 0, new_ysize, wm );
		//Show it:
		recalc_regions( wm );
		draw_window( wm->root_win, wm );
	    }
	}
    }
    
    return 0;
}

int prefs_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    prefs_data* data = (prefs_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( prefs_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		
		g_prefs_restart_request = 0;

		data->list_xsize = 16;
		g_prefs_sections = 0;
		while( 1 )
		{
		    const UTF8_CHAR* section_name = g_prefs_section_names[ g_prefs_sections ];
		    int x = string_size( win, section_name, wm ) + 1;
		    if( x > data->list_xsize ) data->list_xsize = x;
		    if( section_name[ 0 ] == '0' ) break;
		    g_prefs_sections++;
		}
		
		CREATE_SMOOTH_LIST = 1;
		CREATE_LIST_WITHOUT_SCROLLBAR = 1;
		data->sections = new_window( "Sections", 0, 0, 1, 1, wm->list_background, win, list_handler, wm );
		set_window_controller( data->sections, 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->sections, 1, wm, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->sections, 2, wm, (WCMD)INTERELEMENT_SPACE + data->list_xsize, CEND );
		set_window_controller( data->sections, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)BUTTON_YSIZE + INTERELEMENT_SPACE * 2, CEND );
		set_handler( data->sections, prefs_sections_handler, data, wm );
		list_data* l = list_get_data( data->sections, wm );
		for( int i = 0; i < g_prefs_sections; i++ )
		{
		    const UTF8_CHAR* section_name = g_prefs_section_names[ i ];
		    list_add_item( section_name, 0, l );
		}
		list_set_selected_num( 0, l );

		data->cur_section = 0;
		data->cur_section_window = new_window( "Section", 0, 0, 1, 1, win->color, win, (int(*)(sundog_event*,window_manager*)) g_prefs_section_handlers[ 0 ], wm );
		set_window_controller( data->cur_section_window, 0, wm, (WCMD)INTERELEMENT_SPACE * 2 + data->list_xsize, CEND );
		set_window_controller( data->cur_section_window, 1, wm, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->cur_section_window, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->cur_section_window, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );

		data->close = new_window( "Close", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		set_handler( data->close, prefs_close_handler, data, wm );
		int x = INTERELEMENT_SPACE;
		set_window_controller( data->close, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->close, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->close, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->close, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)BUTTON_YSIZE + INTERELEMENT_SPACE, CEND );
		
		//data->correct_ysize = win->ysize;
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    break;
	case EVT_BUTTONDOWN:
	    if( evt->key == KEY_ESCAPE )
	    {
		prefs_close_handler( data, 0, wm );
		retval = 1;
	    }
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    retval = 1;
	    g_prefs = 0;
	    break;
	    
	case EVT_SCREENRESIZE:
	case EVT_BEFORESHOW:
	    //resize_window_with_decorator( win, 0, data->correct_ysize, wm );
	    break;
    }
    return retval;
}

//
//
//

struct prefs_ui_data
{
    WINDOWPTR this_window;
    WINDOWPTR maxfps;
    WINDOWPTR control;
    WINDOWPTR color;
    WINDOWPTR scale;
    WINDOWPTR virt_kbd;
    WINDOWPTR window_pars;
};

void prefs_ui_reinit( WINDOWPTR win )
{
    prefs_ui_data* data = (prefs_ui_data*)win->data;
    
    UTF8_CHAR ts[ 32 ];
    const UTF8_CHAR* v = "";
    
    sprintf( ts, "Max FPS = %d", profile_get_int_value( "maxfps", g_wm->max_fps, 0 ) );
    button_set_text( data->maxfps, ts, g_wm );
    
    int tc = profile_get_int_value( KEY_TOUCHCONTROL, -1, 0 );
    int pc = profile_get_int_value( KEY_PENCONTROL, -1, 0 );
    if( tc == -1 && pc == -1 ) v = "Auto";
    if( tc >= 0 ) v = "Fingers";
    if( pc >= 0 ) v = "Pen or Mouse";
    sprintf( ts, "Control type = %s", v );
    button_set_text( data->control, ts, g_wm );
    
    int vk = -1;
    if( profile_get_int_value( KEY_SHOW_VIRT_KBD, -1, 0 ) != -1 ) vk = 1;
    if( profile_get_int_value( KEY_HIDE_VIRT_KBD, -1, 0 ) != -1 ) vk = 0;
    if( vk == -1 ) v = "Auto";
    if( vk == 0 ) v = "NO";
    if( vk == 1 ) v = "YES";
    sprintf( ts, "Show text keyboard = %s", v );
    button_set_text( data->virt_kbd, ts, g_wm );
}

int prefs_ui_fps_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_ui_data* data = (prefs_ui_data*)user_data;
    
    int new_fps = -1;
    int v = win->action_result;
    switch( v )
    {
	case 0: new_fps = 0; break;
	case 1: new_fps = 5; break;
	case 2: new_fps = 8; break;
	case 3: new_fps = 10; break;
	case 4: new_fps = 15; break;
	case 5: new_fps = 20; break;
	case 6: new_fps = 25; break;
	case 7: new_fps = 30; break;
	case 8: new_fps = 40; break;
	case 9: new_fps = 50; break;
	case 10: new_fps = 60; break;
	case 11: new_fps = 120; break;
	default: new_fps = -1; break;
    }
    if( new_fps == 0 )
	profile_remove_key( "maxfps", 0 );
    if( new_fps > 0 )
	profile_set_int_value( "maxfps", new_fps, 0 );
    if( new_fps != -1 )
    {
	profile_save( 0 );
	g_prefs_restart_request = 1;
    }
    
    prefs_ui_reinit( data->this_window );
    draw_window( data->this_window, wm );
    
    return 0;
}

int prefs_ui_control_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_ui_data* data = (prefs_ui_data*)user_data;

    int v = win->action_result;
    switch( v )
    {
	case 0:
	    //Auto:
	    profile_remove_key( KEY_TOUCHCONTROL, 0 );
	    profile_remove_key( KEY_PENCONTROL, 0 );
	    break;
	case 1:
	    //Fingers:
	    profile_set_int_value( KEY_TOUCHCONTROL, 1, 0 );
	    profile_remove_key( KEY_PENCONTROL, 0 );
	    break;
	case 2:
	    //Pen/Mouse:
	    profile_set_int_value( KEY_PENCONTROL, 1, 0 );
	    profile_remove_key( KEY_TOUCHCONTROL, 0 );
	    break;
	default:
	    return 0;
	    break;
    }
    profile_save( 0 );
    g_prefs_restart_request = 1;

    prefs_ui_reinit( data->this_window );
    draw_window( data->this_window, wm );
    
    return 0;
}

int prefs_ui_color_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    colortheme_open();
    return 0;
}

int prefs_ui_scale_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    ui_scale_open();
    return 0;
}

int prefs_ui_virt_kbd_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_ui_data* data = (prefs_ui_data*)user_data;

    int v = win->action_result;
    switch( v )
    {
	case 0:
	    //Auto:
	    profile_remove_key( KEY_SHOW_VIRT_KBD, 0 );
	    profile_remove_key( KEY_HIDE_VIRT_KBD, 0 );
	    break;
	case 1:
	    //Yes:
	    profile_set_int_value( KEY_SHOW_VIRT_KBD, 1, 0 );
	    profile_remove_key( KEY_HIDE_VIRT_KBD, 0 );
	    break;
	case 2:
	    //No:
	    profile_set_int_value( KEY_HIDE_VIRT_KBD, 1, 0 );
	    profile_remove_key( KEY_SHOW_VIRT_KBD, 0 );
	    break;
	default:
	    return 0;
	    break;
    }
    profile_save( 0 );
    g_prefs_restart_request = 1;

    prefs_ui_reinit( data->this_window );
    draw_window( data->this_window, wm );
    
    return 0;
}

int prefs_ui_window_pars_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    int screen_xsize = profile_get_int_value( KEY_SCREENX, g_wm->real_window_width, 0 );
    int screen_ysize = profile_get_int_value( KEY_SCREENY, g_wm->real_window_width, 0 );
    int fullscreen = 0;
    if( profile_get_int_value( KEY_FULLSCREEN, -1, 0 ) != -1 ) fullscreen = 1;
    dialog_item di[ 6 ];
    di[ 0 ].type = DIALOG_ITEM_LABEL;
    di[ 0 ].str_val = (UTF8_CHAR*)"Default window width:";
    di[ 1 ].type = DIALOG_ITEM_NUMBER;
    di[ 1 ].min = 0;
    di[ 1 ].max = 8000;
    di[ 1 ].int_val = screen_xsize;
    di[ 2 ].type = DIALOG_ITEM_LABEL;
    di[ 2 ].str_val = (UTF8_CHAR*)"Default window height:";
    di[ 3 ].type = DIALOG_ITEM_NUMBER;
    di[ 3 ].min = 0;
    di[ 3 ].max = 8000;
    di[ 3 ].int_val = screen_ysize;
    di[ 4 ].type = DIALOG_ITEM_POPUP;
    di[ 4 ].str_val = (UTF8_CHAR*)"Fullscreen";
    di[ 4 ].int_val = fullscreen;
    di[ 4 ].menu = "Disabled\nEnabled";
    di[ 5 ].type = DIALOG_ITEM_NONE;
    DIALOG_ITEMS = di;
    int d = dialog( "Window parameters", "OK;Cancel" );
    if( d == 0 )
    {
        bool changed = 0;
    	if( di[ 1 ].int_val != screen_xsize )
    	{
    	    profile_set_int_value( KEY_SCREENX, di[ 1 ].int_val, 0 );
    	    changed = 1;
    	}
    	if( di[ 3 ].int_val != screen_ysize )
    	{
    	    profile_set_int_value( KEY_SCREENY, di[ 3 ].int_val, 0 );
    	    changed = 1;
    	}
    	if( di[ 4 ].int_val != fullscreen )
    	{
    	    if( di[ 4 ].int_val )
    		profile_set_int_value( KEY_FULLSCREEN, 1, 0 );
    	    else
    		profile_remove_key( KEY_FULLSCREEN, 0 );
    	    changed = 1;
    	}
        if( changed )
        {
    	    g_prefs_restart_request = 1;
    	    profile_save( 0 );
    	}
    }
    return 0;
}

int prefs_ui_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    prefs_ui_data* data = (prefs_ui_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( prefs_ui_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		
		int y = 0;

		data->window_pars = new_window( "Window parameters", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->window_pars, prefs_ui_window_pars_handler, data, wm );
		set_window_controller( data->window_pars, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->window_pars, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		data->scale = new_window( "Scale", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->scale, prefs_ui_scale_handler, data, wm );
		set_window_controller( data->scale, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->scale, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		data->color = new_window( "Color theme", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->color, prefs_ui_color_handler, data, wm );
		set_window_controller( data->color, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->color, 2, wm, CPERC, (WCMD)100, CEND );
		if( g_prefs_flags & PREFS_FLAG_NO_COLOR_THEME )
		    data->color->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
		else
		    y += wm->text_ysize + INTERELEMENT_SPACE;

		CREATE_BUTTON_WITH_TEXT_MENU = "Auto\n5\n8\n10\n15\n20\n25\n30\n40\n50\n60\n120";
		data->maxfps = new_window( "Max Frames Per Second", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->maxfps, prefs_ui_fps_handler, data, wm );
		set_window_controller( data->maxfps, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->maxfps, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		CREATE_BUTTON_WITH_TEXT_MENU = "Auto\nFingers\nPen or Mouse";
		data->control = new_window( "Control", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->control, prefs_ui_control_handler, data, wm );
		set_window_controller( data->control, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->control, 2, wm, CPERC, (WCMD)100, CEND );
		if( g_prefs_flags & PREFS_FLAG_NO_CONTROL_TYPE )
		    data->control->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
		else
		    y += wm->text_ysize + INTERELEMENT_SPACE;

		CREATE_BUTTON_WITH_TEXT_MENU = "Auto\nYES\nNO";
		data->virt_kbd = new_window( "Show text keyboard", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->virt_kbd, prefs_ui_virt_kbd_handler, data, wm );
		set_window_controller( data->virt_kbd, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->virt_kbd, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;
		
		prefs_ui_reinit( win );
		
		g_pref_section_ysize = y;
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    retval = 1;
	    break;
    }
    return retval;
}

//
//
//

struct prefs_audio_data
{
    WINDOWPTR this_window;
    WINDOWPTR buf;
    WINDOWPTR freq;
    WINDOWPTR driver;
    WINDOWPTR device;
    WINDOWPTR device_in;
    WINDOWPTR more;
};

void prefs_audio_reinit( WINDOWPTR win )
{
    prefs_audio_data* data = (prefs_audio_data*)win->data;
    
    UTF8_CHAR ts[ 256 ];

    UTF8_CHAR* drv = profile_get_str_value( KEY_AUDIODRIVER, 0, 0 );
    if( mem_strstr( drv, "asio" ) )
    {
	if( data->more->flags & WIN_FLAG_ALWAYS_INVISIBLE )
	{
	    data->more->flags &= ~WIN_FLAG_ALWAYS_INVISIBLE;
	    show_window( data->more, g_wm );
	    recalc_regions( g_wm );
        }
    }
    else
    {
	if( ( data->more->flags & WIN_FLAG_ALWAYS_INVISIBLE ) == 0 )
	{
	    data->more->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
	    hide_window( data->more, g_wm );
	    recalc_regions( g_wm );
	}
    }
    if( drv == 0 )
    {
	sprintf( ts, "Driver = Auto" );
    }
    else
    {
	sprintf( ts, "Device = %s", drv );
	UTF8_CHAR** names = 0;
	UTF8_CHAR** infos = 0;
	int drivers = sound_stream_get_drivers( &names, &infos );
	if( ( drivers > 0 ) && names && infos )
	{
	    for( int d = 0; d < drivers; d++ )
	    {
		if( mem_strcmp( names[ d ], drv ) == 0 )
		{
		    sprintf( ts, "Driver = %s", infos[ d ] );
		    break;
		}
	    }
	    for( int d = 0; d < drivers; d++ )
	    {
		mem_free( names[ d ] );
		mem_free( infos[ d ] );
	    }
	    mem_free( names );
	    mem_free( infos );
	}
    }
    button_set_text( data->driver, ts, g_wm );

    UTF8_CHAR* dev = profile_get_str_value( KEY_AUDIODEVICE, 0, 0 );
    if( dev == 0 )
	sprintf( ts, "Output = Auto" );
    else
    {
	sprintf( ts, "Output = %s", dev );
	UTF8_CHAR** names = 0;
	UTF8_CHAR** infos = 0;
	int devices = sound_stream_get_devices( profile_get_str_value( KEY_AUDIODRIVER, 0, 0 ), &names, &infos, 0 );
	if( ( devices > 0 ) && names && infos )
	{
	    for( int d = 0; d < devices; d++ )
	    {
		if( mem_strcmp( names[ d ], dev ) == 0 )
		{
		    ts[ 0 ] = 0;
		    mem_strcat_s( ts, sizeof( ts ), "Output = " );
		    mem_strcat_s( ts, sizeof( ts ), infos[ d ] );
		    break;
		}
	    }
	    for( int d = 0; d < devices; d++ )
	    {
		mem_free( names[ d ] );
		mem_free( infos[ d ] );
	    }
	    mem_free( names );
	    mem_free( infos );
	}
    }
    button_set_text( data->device, ts, g_wm );

    dev = profile_get_str_value( KEY_AUDIODEVICE_IN, 0, 0 );
    if( dev == 0 )
	sprintf( ts, "Input = Auto" );
    else
    {
	sprintf( ts, "Input = %s", dev );
	UTF8_CHAR** names = 0;
	UTF8_CHAR** infos = 0;
	int devices = sound_stream_get_devices( profile_get_str_value( KEY_AUDIODRIVER, 0, 0 ), &names, &infos, 1 );
	if( ( devices > 0 ) && names && infos )
	{
	    for( int d = 0; d < devices; d++ )
	    {
		if( mem_strcmp( names[ d ], dev ) == 0 )
		{
		    ts[ 0 ] = 0;
		    mem_strcat_s( ts, sizeof( ts ), "Input = " );
		    mem_strcat_s( ts, sizeof( ts ), infos[ d ] );
		    break;
		}
	    }
	    for( int d = 0; d < devices; d++ )
	    {
		mem_free( names[ d ] );
		mem_free( infos[ d ] );
	    }
	    mem_free( names );
	    mem_free( infos );
	}
    }
    button_set_text( data->device_in, ts, g_wm );
    
    int freq = 44100;
    if( g_snd_initialized )
        freq = g_snd.freq;
    int size = profile_get_int_value( KEY_SOUNDBUFFER, 0, 0 );
    if( size == 0 )
	sprintf( ts, "Buffer = Auto" );
    else
	sprintf( ts, "Buffer = %d; %d ms", size, ( size * 1000 ) / freq );
    button_set_text( data->buf, ts, g_wm );

    freq = profile_get_int_value( KEY_FREQ, 0, 0 );
    if( freq == 0 )
	sprintf( ts, "Frequency = Auto" );
    else
	sprintf( ts, "Frequency = %d Hz", freq );
    button_set_text( data->freq, ts, g_wm );
}

int prefs_audio_driver_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_audio_data* data = (prefs_audio_data*)user_data;
    
    size_t menu_size = 8192;
    UTF8_CHAR* menu = (UTF8_CHAR*)MEM_NEW2( menu_size );
    menu[ 0 ] = 0;
    
    mem_strcat_s( menu, menu_size, "Auto" );
    
    UTF8_CHAR** names = 0;
    UTF8_CHAR** infos = 0;
    int drivers = sound_stream_get_drivers( &names, &infos );
    if( ( drivers > 0 ) && names && infos )
    {
	mem_strcat_s( menu, menu_size, "\n" );
	
	for( int d = 0; d < drivers; d++ )
	{
	    mem_strcat_s( menu, menu_size, infos[ d ] );
	    if( d != drivers - 1 )
		mem_strcat_s( menu, menu_size, "\n" );
	}

	int sel = popup_menu( "Driver", menu, win->screen_x, win->screen_y, wm->menu_color );
	if( (unsigned)sel < (unsigned)drivers + 1 )
	{
	    if( sel == 0 )
	    {
		//Auto:
		if( profile_get_str_value( KEY_AUDIODRIVER, 0, 0 ) != 0 )
		{
		    profile_remove_key( KEY_AUDIODRIVER, 0 );
		    profile_remove_key( KEY_AUDIODEVICE, 0 );
		    profile_remove_key( KEY_AUDIODEVICE_IN, 0 );
            	    g_prefs_restart_request = 1;
        	    profile_save( 0 );
		}
	    }
	    else
	    {
		if( mem_strcmp( profile_get_str_value( KEY_AUDIODRIVER, "", 0 ), names[ sel - 1 ] ) )
		{
		    profile_set_str_value( KEY_AUDIODRIVER, names[ sel - 1 ], 0 );
		    profile_remove_key( KEY_AUDIODEVICE, 0 );
		    profile_remove_key( KEY_AUDIODEVICE_IN, 0 );
		    g_prefs_restart_request = 1;
            	    profile_save( 0 );
		}
	    }
	}

	for( int d = 0; d < drivers; d++ )
	{
	    mem_free( names[ d ] );
	    mem_free( infos[ d ] );
	}
	mem_free( names );
	mem_free( infos );
    }
    
    mem_free( menu );    

    prefs_audio_reinit( data->this_window );
    draw_window( data->this_window, wm );

    return 0;
}

int prefs_audio_device_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_audio_data* data = (prefs_audio_data*)user_data;
    
    bool input = 0;
    const UTF8_CHAR* key = KEY_AUDIODEVICE;
    if( win == data->device_in )
    {
	input = 1;
	key = KEY_AUDIODEVICE_IN;
    }
    
    size_t menu_size = 8192;
    UTF8_CHAR* menu = (UTF8_CHAR*)MEM_NEW2( menu_size );
    menu[ 0 ] = 0;
    
    mem_strcat_s( menu, menu_size, "Auto" );
    
    UTF8_CHAR** names = 0;
    UTF8_CHAR** infos = 0;
    int devices = sound_stream_get_devices( profile_get_str_value( KEY_AUDIODRIVER, 0, 0 ), &names, &infos, input );
    if( ( devices > 0 ) && names && infos )
    {
	mem_strcat_s( menu, menu_size, "\n" );
	
	for( int d = 0; d < devices; d++ )
	{
	    mem_strcat_s( menu, menu_size, infos[ d ] );
	    if( d != devices - 1 )
		mem_strcat_s( menu, menu_size, "\n" );
	}

	int sel;
	if( input )
	    sel = popup_menu( "Input device", menu, win->screen_x, win->screen_y, wm->menu_color );
	else
	    sel = popup_menu( "Device", menu, win->screen_x, win->screen_y, wm->menu_color );
	if( (unsigned)sel < (unsigned)devices + 1 )
	{
	    if( sel == 0 )
	    {
		//Auto:
		if( profile_get_str_value( key, 0, 0 ) != 0 )
		{
		    profile_remove_key( key, 0 );
            	    g_prefs_restart_request = 1;
        	    profile_save( 0 );
		}
	    }
	    else
	    {
		if( mem_strcmp( profile_get_str_value( key, "", 0 ), names[ sel - 1 ] ) )
		{
		    profile_set_str_value( key, names[ sel - 1 ], 0 );
		    g_prefs_restart_request = 1;
            	    profile_save( 0 );
		}
	    }
	}

	for( int d = 0; d < devices; d++ )
	{
	    mem_free( names[ d ] );
	    mem_free( infos[ d ] );
	}
	mem_free( names );
	mem_free( infos );
    }
    
    mem_free( menu );    

    prefs_audio_reinit( data->this_window );
    draw_window( data->this_window, wm );

    return 0;
}

int g_sound_buf_size_table[] = 
{
    128,
    256, 
    512,
    768,
    1024,
    1280,
    1536,
    1792,
    2048,
    2560,
    3072,
    4096,
    -1
};

int prefs_audio_buf_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_audio_data* data = (prefs_audio_data*)user_data;
    
    size_t menu_size = 8192;
    UTF8_CHAR* menu = (UTF8_CHAR*)MEM_NEW2( menu_size );
    menu[ 0 ] = 0;
    mem_strcat_s( menu, menu_size, "Auto" );
    
    int i = 0;
    int size = 0;

#if defined(PALMOS) || defined(OSX) || defined(IPHONE)
#else
    int freq = 44100;
    if( g_snd_initialized )
        freq = g_snd.freq;
    UTF8_CHAR ts[ 256 ];
    while( 1 )
    {
	size = g_sound_buf_size_table[ i ];
	if( size == -1 ) break;
	sprintf( ts, "\n%d; %d ms", size, ( size * 1000 ) / freq ); mem_strcat_s( menu, menu_size, ts );
	i++;
    }
#endif
    
    size = 0;
    int v = popup_menu( "Audio buffer size", menu, win->screen_x, win->screen_y, wm->menu_color );
    if( v > 0 && v < i + 1 )
	size = g_sound_buf_size_table[ v - 1 ];
    
    mem_free( menu );
    
    if( v == 0 || size > 0 )
    {
	if( v == 0 )
	{
	    if( profile_get_int_value( KEY_SOUNDBUFFER, -1, 0 ) != -1 )
	    {
		profile_remove_key( KEY_SOUNDBUFFER, 0 );
		g_prefs_restart_request = 1;
		profile_save( 0 );
	    }
	}
	if( size > 0 )
	{
	    if( profile_get_int_value( KEY_SOUNDBUFFER, -1, 0 ) != size )
	    {
		profile_set_int_value( KEY_SOUNDBUFFER, size, 0 );
		g_prefs_restart_request = 1;
		profile_save( 0 );
	    }
	}
    }

    prefs_audio_reinit( data->this_window );
    draw_window( data->this_window, wm );
    
    return 0;
}

int prefs_audio_freq_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_audio_data* data = (prefs_audio_data*)user_data;
    
    int freq = 0;
    int v = win->action_result;
    switch( v )
    {
	case 1: freq = 44100; break;
	case 2: freq = 48000; break;
	case 3: freq = 96000; break;
	case 4: freq = 192000; break;
    }
    
    if( v == 0 || freq > 0 )
    {
	if( v == 0 )
	{
	    if( profile_get_int_value( KEY_FREQ, -1, 0 ) != -1 )
	    {
		profile_remove_key( KEY_FREQ, 0 );
		g_prefs_restart_request = 1;
		profile_save( 0 );
	    }
	}
	if( freq > 0 )
	{
	    if( profile_get_int_value( KEY_FREQ, -1, 0 ) != freq )
	    {
		profile_set_int_value( KEY_FREQ, freq, 0 );
		g_prefs_restart_request = 1;
		profile_save( 0 );
	    }
	}
    }

    prefs_audio_reinit( data->this_window );
    draw_window( data->this_window, wm );
    
    return 0;
}

int prefs_audio_more_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_audio_data* data = (prefs_audio_data*)user_data;

    UTF8_CHAR* drv = profile_get_str_value( KEY_AUDIODRIVER, 0, 0 );
    if( mem_strstr( drv, "asio" ) )
    {
	dialog_item di[ 5 ];
        di[ 0 ].type = DIALOG_ITEM_LABEL;
        di[ 0 ].str_val = (UTF8_CHAR*)"First output channel:";
        di[ 1 ].type = DIALOG_ITEM_NUMBER;
        di[ 1 ].min = 0;
        di[ 1 ].max = 64;
        di[ 1 ].int_val = profile_get_int_value( "audio_ch", 0, 0 );
        di[ 2 ].type = DIALOG_ITEM_LABEL;
        di[ 2 ].str_val = (UTF8_CHAR*)"First input channel:";
        di[ 3 ].type = DIALOG_ITEM_NUMBER;
        di[ 3 ].min = 0;
        di[ 3 ].max = 64;
        di[ 3 ].int_val = profile_get_int_value( "audio_ch_in", 0, 0 );
        di[ 4 ].type = DIALOG_ITEM_NONE;
        DIALOG_ITEMS = di;
        int d = dialog( "ASIO options", "OK;Cancel" );
        if( d == 0 )
        {
    	    bool changed = 0;
    	    if( profile_get_int_value( "audio_ch", 0, 0 ) != di[ 1 ].int_val )
    	    {
    		profile_set_int_value( "audio_ch", di[ 1 ].int_val, 0 );
    		changed = 1;
    	    }
    	    if( profile_get_int_value( "audio_ch_in", 0, 0 ) != di[ 3 ].int_val )
    	    {
    		profile_set_int_value( "audio_ch_in", di[ 3 ].int_val, 0 );
    		changed = 1;
    	    }
    	    if( changed )
    	    {
    		g_prefs_restart_request = 1;
        	profile_save( 0 );
    	    }
        }
    }
    
    return 0;
}   

int prefs_audio_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    prefs_audio_data* data = (prefs_audio_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( prefs_audio_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		
		int y = 0;

		BUTTON_FLAGS = BUTTON_FLAG_LEFT_ALIGNMENT_ON_OVERFLOW;
		data->driver = new_window( "Driver", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->driver, prefs_audio_driver_handler, data, wm );
		set_window_controller( data->driver, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->driver, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		BUTTON_FLAGS = BUTTON_FLAG_LEFT_ALIGNMENT_ON_OVERFLOW;
		data->device = new_window( "Output device", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->device, prefs_audio_device_handler, data, wm );
		set_window_controller( data->device, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->device, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		BUTTON_FLAGS = BUTTON_FLAG_LEFT_ALIGNMENT_ON_OVERFLOW;
		data->device_in = new_window( "Input device", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->device_in, prefs_audio_device_handler, data, wm );
		set_window_controller( data->device_in, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->device_in, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		data->buf = new_window( "Audio buffer size", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->buf, prefs_audio_buf_handler, data, wm );
		set_window_controller( data->buf, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->buf, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

#if defined(ONLY44100)
		CREATE_BUTTON_WITH_TEXT_MENU = "Auto";
#else
		CREATE_BUTTON_WITH_TEXT_MENU = "Auto\n44100\n48000\n96000\n192000";
#endif
		data->freq = new_window( "Sampling frequency", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->freq, prefs_audio_freq_handler, data, wm );
		set_window_controller( data->freq, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->freq, 2, wm, CPERC, (WCMD)100, CEND );
		y += wm->text_ysize + INTERELEMENT_SPACE;

		data->more = new_window( "More options", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
		set_handler( data->more, prefs_audio_more_handler, data, wm );
		set_window_controller( data->more, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->more, 2, wm, CPERC, (WCMD)100, CEND );
		data->more->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
		y += wm->text_ysize + INTERELEMENT_SPACE;

		prefs_audio_reinit( win );
		
		g_pref_section_ysize = y;
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
                if( g_snd_initialized )
                {
                    UTF8_CHAR ts[ 256 ];
                    int y = data->freq->y + data->freq->ysize + INTERELEMENT_SPACE;
                    if( data->more->visible )
                        y = data->more->y + data->more->ysize + INTERELEMENT_SPACE;
                    ts[ 0 ] = 0;
                    mem_strcat_s( ts, 256, "Current driver: " );
                    mem_strcat_s( ts, 256, sound_stream_get_driver_info() );
                    win_draw_string( win, ts, 0, y, wm->color2, win->color, wm );
                    sprintf( ts, "Current frequency: %d Hz", g_snd.freq );
                    win_draw_string( win, ts, 0, y + char_y_size( win, wm ), wm->color2, win->color, wm );
                    sprintf( ts, "Current latency: %d; %d ms", g_snd.latency, ( g_snd.latency * 1000 ) / g_snd.freq );
                    win_draw_string( win, ts, 0, y + char_y_size( win, wm ) * 2, wm->color2, win->color, wm );
                }
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    retval = 1;
	    break;
    }
    return retval;
}
