#ifndef __WINDOWMANAGER_STRUCT__
#define __WINDOWMANAGER_STRUCT__

#include "../core/core.h"
#include "../memory/memory.h"
#include "../filesystem/v3nus_fs.h"
#include "../time/timemanager.h"
#include "regions/device.h"

#define EVENTS	64 //must be 2/4/8/16/32...
#define TIMERS	16

#define WM_INLINE inline

#if defined(WIN) || defined(WINCE)
    #include <windows.h>
#endif

#ifdef UNIX
    #include <pthread.h>
#endif

#ifdef OPENGL
#ifdef IPHONE
    #include <OpenGLES/ES1/gl.h>
    #include <OpenGLES/ES1/glext.h>
#endif
#ifdef OSX
    #include <OpenGL/gl.h>
    #include <OpenGL/glext.h>
    #include <OpenGL/glu.h>
    #include <OpenGL/OpenGL.h>
#endif
#if defined(LINUX) || defined(FREEBSD) || defined(WIN)
#ifdef OPENGLES
    #include <GLES/gl.h>
    #include <GLES/glext.h>
    #include <EGL/egl.h>
#else
    #include <GL/gl.h>
    #include <GL/glext.h>
    #include <GL/glu.h>
#endif
#endif //LINUX,BSD,WINDOWS
#endif //OPENGL

#ifdef X11
#ifdef OPENGL
#ifndef OPENGLES
    #include <GL/glx.h>
#endif
#endif
    #include <X11/Xlib.h>
    #include <X11/Xutil.h>
    #include <X11/Xatom.h>
    #include <X11/keysym.h>
    #include <X11/XKBlib.h>
    #include <stdlib.h>
#endif

#if defined(DIRECTDRAW) && defined(UNIX)
    #include "SDL/SDL.h"
#endif

#if defined(DIRECTDRAW) && defined(WIN)
    #include "ddraw.h"
#endif

#ifdef COLOR8BITS
    #define COLOR	uchar
    #define COLORSIGNED	signed char
    #define COLORPTR	uchar*
    #define COLORLEN	(int)1
    #define COLORBITS	8
    #define COLORMASK	0xFF
    #define CLEARCOLOR	0
#endif
#ifdef COLOR16BITS
    #define COLOR	uint16
    #define COLORSIGNED	int16
    #define COLORPTR  	uint16*
    #define COLORLEN  	(int)2
    #define COLORBITS 	16
    #define COLORMASK 	0xFFFF
    #define CLEARCOLOR 	0
#endif
#ifdef COLOR32BITS
    #define COLOR     	uint
    #define COLORSIGNED	int
    #define COLORPTR  	uint*
    #define COLORLEN  	(int)4
    #define COLORBITS 	32
    #define COLORMASK 	0x00FFFFFF
    #define CLEARCOLOR 	0x00000000
#endif

#ifdef OPENGL
    #define SCREEN_ZOOM_AVAILABLE 1
#endif
#if defined(DIRECTDRAW) && defined(UNIX)
    #define SCREEN_ZOOM_AVAILABLE 1    
#endif
#ifdef WINCE
    #define SCREEN_ZOOM_AVAILABLE 1
#endif

#define CW 				COLORMASK
#define CB 				CLEARCOLOR

#define WIN_INIT_FLAG_SCALABLE	    	( 1 << 0 )
#define WIN_INIT_FLAG_NOBORDER	    	( 1 << 1 )
#define WIN_INIT_FLAG_FULLSCREEN    	( 1 << 2 )
#define WIN_INIT_FLAG_FULL_CPU_USAGE	( 1 << 3 )
#define WIN_INIT_FLAGOFF_ANGLE		4
#define WIN_INIT_FLAG_TOUCHCONTROL	( 1 << 6 )
#define WIN_INIT_FLAG_OPTIMIZE_MOVE_EVENTS	( 1 << 7 )
#define WIN_INIT_FLAG_NOSCREENROTATE	( 1 << 8 )
#define WIN_INIT_FLAG_LANDSCAPE_ONLY	( 1 << 9 )
#define WIN_INIT_FLAG_PORTRAIT_ONLY     ( 1 << 10 )
#define WIN_INIT_FLAG_NOCURSOR		( 1 << 11 )
#define WIN_INIT_FLAG_VSYNC		( 1 << 12 )
#define WIN_INIT_FLAG_NOAUTOZOOM	( 1 << 13 )
#define WIN_INIT_FLAGOFF_ZOOM		14

enum 
{
    EVT_NULL = 0,
    EVT_GETDATASIZE,
    EVT_AFTERCREATE,
    EVT_BEFORECLOSE,
    EVT_BEFORESHOW,
    EVT_AFTERRESIZE,
    EVT_DRAW,
    EVT_FRAME,
    EVT_FOCUS,
    EVT_UNFOCUS, //When user click on other window
    EVT_MOUSEBUTTONDOWN,
    EVT_MOUSEBUTTONUP,
    EVT_MOUSEMOVE,
    EVT_TOUCHBEGIN,
    EVT_TOUCHEND,
    EVT_TOUCHMOVE,
    EVT_BUTTONDOWN,
    EVT_BUTTONUP,
    EVT_SCREENRESIZE, //After main screen resize
    EVT_SCREENFOCUS,
    EVT_SCREENUNFOCUS,
    EVT_CLOSEREQUEST, //Close Window request. Can be ignored
    EVT_QUIT, //Close Application request. E.g. when user click Main Window Close button. Can be ignored
              //Available not for all systems.
    EVT_USERDEFINED1
};

//Event flags:
#define EVT_FLAG_SHIFT 			1
#define EVT_FLAG_CTRL 			2
#define EVT_FLAG_ALT 			4
#define EVT_FLAG_MODE 			8
#define EVT_FLAG_MODS			( EVT_FLAG_SHIFT | EVT_FLAG_CTRL | EVT_FLAG_ALT | EVT_FLAG_MODE )
#define EVT_FLAG_DOUBLECLICK		16
#define EVT_FLAG_DONTDRAW		32
#define EVT_FLAG_HANDLING		64

//Mouse buttons:
#define MOUSE_BUTTON_LEFT 		1
#define MOUSE_BUTTON_MIDDLE 		2
#define MOUSE_BUTTON_RIGHT 		4
#define MOUSE_BUTTON_SCROLLUP 		8
#define MOUSE_BUTTON_SCROLLDOWN 	16

//Virtual key codes (ASCII):
//(scancode and unicode may be set for each virtual key)
// 00      : empty;
// 01 - 1F : control codes;
// 20 - 7E : standart ASCII symbols (all letters are small - from 61);
// 7F      : not defined;
// 80 - xx : additional key codes
#define KEY_BACKSPACE   		0x08
#define KEY_TAB         		0x09
#define KEY_ENTER       		0x0D
#define KEY_ESCAPE      		0x1B
#define KEY_SPACE       		0x20

//Additional key codes:
enum 
{
    KEY_F1 = 0x80,
    KEY_F2,
    KEY_F3,
    KEY_F4,
    KEY_F5,
    KEY_F6,
    KEY_F7,
    KEY_F8,
    KEY_F9,
    KEY_F10,
    KEY_F11,
    KEY_F12,
    KEY_UP,
    KEY_DOWN,
    KEY_LEFT,
    KEY_RIGHT,
    KEY_INSERT,
    KEY_DELETE,
    KEY_HOME,
    KEY_END,
    KEY_PAGEUP,
    KEY_PAGEDOWN,
    KEY_CAPS,
    KEY_SHIFT,
    KEY_CTRL,
    KEY_ALT,
    KEY_MENU,
    KEY_UNKNOWN, //virtual key code = code - KEY_UNKNOWN
};

//Scroll flags:
#define SF_LEFT         		1
#define SF_RIGHT        		2
#define SF_UP           		4
#define SF_DOWN         		8
#define SF_STATICWIN    		16

#define WINDOWPTR			sundog_window*
#define WINDOW				sundog_window

#define DECOR_BORDER_SIZE		g_wm->decor_border_size
#define DECOR_HEADER_SIZE		g_wm->decor_header_size
#define DECOR_FLAG_CENTERED		( 1 << 0 )
#define DECOR_FLAG_CHECK_SIZE		( 1 << 1 )
#define DECOR_FLAG_STATIC		( 1 << 2 )
#define DECOR_FLAG_NOBORDER		( 1 << 3 )
#define DECOR_FLAG_NOHEADER		( 1 << 4 )
#define DECOR_FLAG_NORESIZE		( 1 << 5 )
#define DECOR_FLAG_FULLSCREEN		( 1 << 6 )
#define DECOR_FLAG_WITH_CLOSE		( 1 << 7 )
#define DECOR_FLAG_WITH_MINIMIZE	( 1 << 8 )

#define SCROLLBAR_SIZE	    		g_wm->scrollbar_size
#define BUTTON_YSIZE			g_wm->button_ysize
#define BUTTON_XSIZE  			g_wm->button_xsize
#define POPUP_BUTTON_YSIZE		g_wm->popup_button_ysize
#define CONTROLLER_YSIZE		g_wm->controller_ysize
#define INTERELEMENT_SPACE  		2

#define BORDER_TRANSP			32
#define BORDER_COLOR_WITHOUT_TRANSP	g_wm->color3
#define PUSHED_TRANSP			40
#define PUSHED_COLOR_WITHOUT_TRASNP	g_wm->color3
#define PUSHED_COLOR( orig ) 		blend( orig, PUSHED_COLOR_WITHOUT_TRASNP, PUSHED_TRANSP )
#define BORDER_COLOR( orig )		blend( orig, BORDER_COLOR_WITHOUT_TRANSP, BORDER_TRANSP )
#define SELECTED_BUTTON_COLOR		blend( g_wm->button_color, g_wm->selection_color, 80 )
#define LABEL_TRANSP			128

#define DOUBLE_CLICK_PERIOD 		200

typedef size_t                          WCMD;
#define CWIN		    		(WCMD)30000
#define CX1		    		(WCMD)30001
#define CX2		    		(WCMD)30002
#define CY1		    		(WCMD)30003
#define CY2		    		(WCMD)30004
#define CXSIZE		    		(WCMD)30005
#define CYSIZE		    		(WCMD)30006
#define CSUB		    		(WCMD)30007
#define CADD		    		(WCMD)30008
#define CPERC		    		(WCMD)30009
#define CBACKVAL0	    		(WCMD)30010
#define CBACKVAL1	    		(WCMD)30011
#define CMAXVAL		    		(WCMD)30012
#define CMINVAL		    		(WCMD)30013
#define CPUTR0		    		(WCMD)30014
#define CGETR0		    		(WCMD)30015
#define CR0		    		(WCMD)30016
#define CEND		    		(WCMD)30017

#define IMAGE_NATIVE_RGB		1
#define IMAGE_ALPHA8			2
#define IMAGE_STATIC_SOURCE		4
#define IMAGE_INTERPOLATION		8
#define IMAGE_CLEAN			16

struct sundog_image
{
    int			xsize;
    int			ysize;
    int			int_xsize; //internal xsize
    int			int_ysize; //internal ysize
    void*		data;
    COLOR		color;
    COLOR		backcolor;
    uchar		transp; //0..255
#ifdef OPENGL
    unsigned int	gl_texture_id;
#endif
    int			flags;
};

#define WBD_FLAG_ONE_COLOR		1
#define WBD_FLAG_ONE_TRANSP		2

struct sundog_vertex
{
    int16 x;
    int16 y;
    COLOR c;
    uchar t;
};

struct sundog_polygon
{
    int vnum;
    sundog_vertex* v;
};

#define WIN_FLAG_ALWAYS_INVISIBLE	( 1 << 0 )
#define WIN_FLAG_ALWAYS_ON_TOP		( 1 << 1 )
#define WIN_FLAG_ALWAYS_UNFOCUSED	( 1 << 2 )
#define WIN_FLAG_TRASH			( 1 << 3 )
#define WIN_FLAG_DOUBLECLICK		( 1 << 4 )
#define WIN_FLAG_DONT_USE_CONTROLLERS	( 1 << 5 )
#define WIN_FLAG_UNFOCUS_HANDLING	( 1 << 6 )
#define WIN_FLAG_FOCUS_HANDLING		( 1 << 7 )

struct sundog_window
{
    int		    	visible;
    uint16	    	flags;
    uint16	    	id;
    const UTF8_CHAR*	name; //window name
    int16	    	x, y; //x/y (relative)
    int16	    	screen_x, screen_y; //real x/y
    int16	    	xsize, ysize; //window size
    COLOR	    	color;
    int	    		font;
    MWCLIPREGION*	reg;
    WINDOWPTR	    	parent;
    WINDOWPTR*		childs;
    int		    	childs_num;
    int		    	(*win_handler)( void*, void* ); //window handler: (*event, *window_manager)
    void*		data;

    WCMD*		x1com; //Controllers of window coordinates (special commands)
    WCMD*		y1com;
    WCMD*		x2com;
    WCMD*		y2com;
    int16	    	controllers_defined;
    int16	    	controllers_calculated;

    int		    	(*action_handler)( void*, void*, void* );
    int		    	action_result;
    void*		handler_data;

    ticks_hr_t	    	click_time;
};

//SunDog event handling:
// EVENT 
//   |
// user_event_handler()
//   | (if not handled)
// window 
//   | (if not handled)
// window children
//   | (if not handled)
// handler_of_unhandled_events()

struct sundog_event
{
    uint16          	type; //event type
    uint16	    	flags;
    ticks_hr_t	   	time;
    WINDOWPTR	    	win;
    int16           	x;
    int16           	y;
    uint16	    	key; //virtual key code: ASCII (0..127) or additional (see KEY_xxx defines)
    uint16	    	scancode; //device dependent (or touch count for multitouch devices)
    uint16	    	pressure; //key pressure (0..1024)
    uint16	    	reserved;
    UTF32_CHAR	    	unicode; //unicode if possible, or zero
};

struct window_manager;
struct sundog_timer;

struct sundog_timer
{
    void            	(*handler)( void*, sundog_timer*, window_manager* );
    void*		data;
    ticks_t	    	deadline;
    ticks_t	    	delay;
    int			id;
};

enum
{
    DIALOG_ITEM_NONE = 0,
    DIALOG_ITEM_NUMBER,
    DIALOG_ITEM_NUMBER_HEX,
    DIALOG_ITEM_TEXT,
    DIALOG_ITEM_LABEL,
    DIALOG_ITEM_POPUP,
};

#define DIALOG_ITEM_TEXT_SIZE g_wm->text_ysize
#define DIALOG_ITEM_NUMBER_SIZE DIALOG_ITEM_TEXT_SIZE
#define DIALOG_ITEM_LABEL_SIZE char_y_size( win, g_wm )
#define DIALOG_ITEM_POPUP_SIZE POPUP_BUTTON_YSIZE

struct dialog_item
{
    int			type;
    int			min;
    int			max;
    int			int_val;
    UTF8_CHAR*		str_val;
    const UTF8_CHAR*	menu;
    
    WINDOWPTR		win;
};

struct new_fdialog_struct
{
    bool		open;
    bool		first_start;
    bool		save_mode;
    const UTF8_CHAR**	dir_names;
    const UTF8_CHAR**	dir_paths;
    const UTF8_CHAR*	name;
    const UTF8_CHAR*	mask;
    const UTF8_CHAR*	default_filename;
    int			dir_num;
    int			file_num;
};

//Control type: 0 - mouse, pen, stylus; 1 - touch/multitouch screen.
#define PENCONTROL			0
#define TOUCHCONTROL			1

struct window_manager
{
    int            	wm_initialized;
    bool		restart_request;
    bool 	    	exit_request;
    int 	    	exit_code;
    int             	argc; //Minimum - 1
    UTF8_CHAR**		argv; // "program name"; "arg1"; "arg2"; ...

    int		    	flags;
    
    int            	events_count; //Number of events to execute.
    int	        	current_event_num;
    sundog_event    	events[ EVENTS ];
    sundog_mutex    	events_mutex;

    WINDOWPTR*		trash;

    uint	    	window_counter; //For ID calculation.

    WINDOWPTR	    	root_win;

    int			creg0; //Temp register for window's controller calculation.

    sundog_timer    	timers[ TIMERS ];
    int		    	timers_num;
    int			timers_id_counter;

    int             	screen_xsize;
    int             	screen_ysize;
    int			screen_angle;
    bool		screen_angle_lock;
    int 		screen_zoom; //pixel size
    int			screen_ppi; //pixels per inch
    float		screen_scale; //UI scale factor
    bool		screen_buffer_preserved; //Screen buffer preserved after a new frame 
						 //(TRUE for framebuffer modes and for single-buffer OpenGL).
						 //If FALSE - all windows will be redrawed every frame.
    int		    	screen_lock_counter;
    int	            	screen_is_active;
    int			screen_changed;
    
    bool		show_virtual_keyboard;
    
    int			default_font;

    int			color_theme;
    COLOR	    	color0;
    COLOR		color1;
    COLOR		color2;
    COLOR	    	color3;
    bool		color3_is_brighter;
    int			color1_darkness;
    COLOR	    	yellow;
    COLOR	    	green;
    COLOR	    	red;
    COLOR	    	blue;
    COLOR		header_text_color;
    COLOR		alternative_text_color;
    COLOR	    	dialog_color;
    COLOR	    	decorator_color;
    COLOR	    	decorator_border;
    COLOR	    	button_color;
    COLOR	    	menu_color;
    COLOR	    	selection_color;
    COLOR	    	text_background;
    COLOR	    	list_background;
    COLOR	    	scroll_color;
    COLOR	    	scroll_background_color;
    
    sundog_image*	texture0;

    int			control_type;
    
    int                 max_fps; //Max FPS, provided by window manager. Less value - less CPU usage.
    bool		frame_event_request;
    
    int			normal_window_xsize;
    int			normal_window_ysize;
    int			large_window_xsize;
    int			large_window_ysize;
    int		    	decor_border_size;
    int		    	decor_header_size;
    int		    	scrollbar_size; //Scrollbar, divider, normal button
    int		    	button_xsize; //Big button
    int		    	button_ysize;
    int			popup_button_ysize;
    int			controller_ysize;
    int			list_item_ysize;
    int			text_ysize; //Text editor
    int			corners_size;
    int			corners_len;
    					  
    // WBD (Window Buffer Draw):
    WINDOWPTR		cur_window;
    COLOR*		g_screen; //Graphics screen pointer.
    COLOR           	cur_font_color;
    int		    	cur_font_zoom; //256 - normal; 512 - x2
    int		    	cur_transparency;
    int			cur_flags;

    int		    	pen_x;
    int		    	pen_y;
    WINDOWPTR	    	focus_win;
    WINDOWPTR	    	prev_focus_win;
    uint16	    	focus_win_id;
    uint16	    	prev_focus_win_id;
    WINDOWPTR	    	last_unfocused_window;

    WINDOWPTR	    	handler_of_unhandled_events;
    
    //DEVICE DEPENDENT PART:

    int		    	fb_xpitch; //framebuffer xpitch
    int		    	fb_ypitch; //...
    int		    	fb_offset; //...
    int			fb_xsize;
    int			fb_ysize;
    int		    	real_window_width;
    int		    	real_window_height;
#ifdef OPENGL
    float		gl_xscale;
    float		gl_yscale;
    unsigned int	gl_fb_texture_id;
    sundog_mutex        gl_mutex;
#endif //OPENGL
#if defined(DIRECTDRAW) && defined(WIN)
    LPDIRECTDRAWSURFACE	lpDDSPrimary; //DirectDraw primary surface (for win32).
    LPDIRECTDRAWSURFACE	lpDDSBack; //DirectDraw back surface (for win32).
#endif
#ifdef GDI
    uint           	gdi_bitmap_info[ 512 ]; //GDI bitmap info (for win32).
#endif
#ifdef WIN
    HDC	            	hdc; //Graphics content handler.
    HWND	    	hwnd;
    HINSTANCE       	hCurrentInst;
    HINSTANCE       	hPreviousInst;
    LPSTR           	lpszCmdLine;
    int             	nCmdShow;
#ifdef OPENGL
    BOOL 		(WINAPI *wglSwapIntervalEXT)( int interval );
#endif
#endif
#ifdef WINCE
    int 		hires;
    int			vd; //video-driver
    char		rfb[ 32 ]; //raw framebuffer info
    HDC             	hdc; //graphics content handler
    HINSTANCE       	hCurrentInst;
    HINSTANCE       	hPreviousInst;
    LPWSTR          	lpszCmdLine;
    int             	nCmdShow;
    int		    	os_version;
    int		    	gx_suspended;
#endif
#ifdef UNIX
    pthread_mutex_t 	sdl_lock_mutex;
#ifdef DIRECTDRAW
    SDL_Surface*	sdl_screen;
#endif
#ifdef X11
    Display*		dpy;
    Colormap	    	cmap;
    Window          	win;
    Visual*		win_visual;
    GC              	win_gc;
    XImage*		win_img;
    XImage*		win_img_back_pattern;
    char*		win_img_back_pattern_data;
    int 	    	win_img_depth;
    char*		win_buffer;
    int	            	win_depth;
#ifdef OPENGL
    int 		(*glXSwapIntervalEXT)( Display* dpy, GLXDrawable drw, int interval );
    void		(*glXSwapIntervalMESA)( GLint );
    void		(*glXSwapIntervalSGI)( GLint );
#endif //OPENGL
#endif //X11
#endif //UNIX
};

#endif
