#ifndef __WINDOWMANAGER__
#define __WINDOWMANAGER__

#include "../utils/utils.h"
#include "struct.h"

#define FONTS_NUM 2
extern uchar g_font0[ 8 * 256 + 257 ];
extern uchar g_font1[ 12 * 256 + 257 ];
extern sundog_image* g_font_img[ FONTS_NUM ];
extern int g_font_cxsize[ FONTS_NUM ];
extern int g_font_cysize[ FONTS_NUM ];
extern window_manager* g_wm;

#define COLOR_THEMES 25
extern uchar g_color_themes[ 3 * 4 * COLOR_THEMES ];

extern COLORPTR framebuffer;

//###################################
//## MAIN FUNCTIONS:               ##
//###################################

int win_init( 
    const UTF8_CHAR* windowname, 
    int xsize, 
    int ysize, 
    int flags, 
    int argc, 
    UTF8_CHAR** argv,
    window_manager* wm ); //init window manager
void win_colors_init( window_manager* wm );
void win_reinit( window_manager* wm );
void win_lock_unlock_orientation( window_manager* wm );
void win_zoom_init( window_manager* wm );
void win_vsync( bool vsync, window_manager* wm );
void win_close( window_manager* wm ); //close window manager

//###################################
//## WINDOWS:                      ##
//###################################

WINDOWPTR new_window( 
    const UTF8_CHAR *name, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR color,
    WINDOWPTR parent, 
    int (*)( sundog_event*, window_manager* ),
    window_manager* wm );
void set_window_controller( WINDOWPTR win, int ctrl_num, window_manager* wm, ... );
void remove_window( WINDOWPTR win, window_manager* wm );
void add_child( WINDOWPTR win, WINDOWPTR child, window_manager* wm );
void remove_child( WINDOWPTR win, WINDOWPTR child, window_manager* wm );
void set_handler( WINDOWPTR win, int (*handler)(void*,WINDOWPTR,window_manager*), void* handler_data, window_manager* wm );
bool is_window_visible( WINDOWPTR win, window_manager* wm );
void draw_window( WINDOWPTR win, window_manager* wm );
void show_window( WINDOWPTR win, window_manager* wm );
void hide_window( WINDOWPTR win, window_manager* wm );
void bring_to_front( WINDOWPTR win, window_manager* wm );
void recalc_regions( window_manager* wm );

void set_focus_win( WINDOWPTR win, window_manager *wm );

int send_events(
    sundog_event* events,
    int events_num,
    window_manager* wm );
int send_event( 
    WINDOWPTR win, 
    int type, 
    int flags, 
    int x, 
    int y, 
    int key, 
    int scancode, 
    int pressure, 
    UTF32_CHAR unicode,
    window_manager* wm );
int check_event( sundog_event *evt, window_manager *wm );
void handle_event( sundog_event *evt, window_manager *wm );
int handle_event_by_window( sundog_event *evt, window_manager *wm );

int EVENT_LOOP_BEGIN( sundog_event *evt, window_manager *wm );
int EVENT_LOOP_END( window_manager *wm );

//###################################
//## STATUS MESSAGES:              ##
//###################################

void show_status_message( const UTF8_CHAR* msg, int t );

//###################################
//## TIMERS:                       ##
//###################################

int add_timer( void (*)( void*, sundog_timer*, window_manager* ), void* data, ticks_t delay, window_manager* wm );
void reset_timer( int timer, window_manager* wm );
void remove_timer( int timer, window_manager* wm );

//###################################
//## WINDOWS DECORATIONS:          ##
//###################################

int decorator_handler( sundog_event *evt, window_manager *wm );
void change_decorator(
    WINDOWPTR dec,
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    int flags,
    window_manager* wm );
void resize_window_with_decorator( WINDOWPTR win, int xsize, int ysize, window_manager* wm );
WINDOWPTR new_window_with_decorator( 
    const UTF8_CHAR *name, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR color,
    WINDOWPTR parent, 
    int (*)( sundog_event*, window_manager* ),
    int flags, //DECOR_FLAG_xxx
    window_manager *wm );

//###################################
//## DRAWING FUNCTIONS:            ##
//###################################

int char_x_size( WINDOWPTR win, UTF32_CHAR c, window_manager* wm );
int char_y_size( WINDOWPTR win, window_manager* wm );
int string_size( WINDOWPTR win, const UTF8_CHAR* str, window_manager* wm );

void win_draw_lock( WINDOWPTR win, window_manager *wm );    //Only draw_window() and event_handler (during the EVT_DRAW handling) do this automaticaly!
void win_draw_unlock( WINDOWPTR win, window_manager *wm );  //Only draw_window() and event_handler (during the EVT_DRAW handling) do this automaticaly!

void win_draw_box( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, window_manager* wm );
void win_draw_frame( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, window_manager* wm );
void win_draw_frame3d( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, int inout, window_manager* wm );
void win_draw_bitmap_ext( 
    WINDOWPTR win, 
    int x, 
    int y, 
    int dest_xsize, 
    int dest_ysize,
    int source_x,
    int source_y,
    sundog_image* data, 
    window_manager* wm );
void win_draw_bitmap( 
    WINDOWPTR win, 
    int x, 
    int y, 
    sundog_image* data, 
    window_manager* wm );
void win_draw_char( WINDOWPTR win, UTF32_CHAR c, int x, int y, COLOR f, COLOR b, window_manager* wm );
void win_draw_string( WINDOWPTR win, const UTF8_CHAR *str, int x, int y, COLOR f, COLOR b, window_manager* wm );
bool line_clip( int* x1, int* y1, int* x2, int* y2, int clip_x1, int clip_y1, int clip_x2, int clip_y2 );
bool line_clip_with_transp( int* x1, int* y1, int* t1, int* x2, int* y2, int* t2, int clip_x1, int clip_y1, int clip_x2, int clip_y2 );
void win_draw_line( WINDOWPTR win, int x1, int y1, int x2, int y2, COLOR color, window_manager* wm );

void screen_changed( void );

WM_INLINE COLOR get_color( uchar r, uchar g, uchar b )
{
    COLOR res = 0; //resulted color

#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = (COLOR)( (int)( ( r + g + b + b ) >> 2 ) );
    #else
	res += ( b & 192 );
	res += ( g & 224 ) >> 2;
	res += ( r >> 5 );
    #endif
#endif
    
#ifdef COLOR16BITS

#ifdef RGB556
    res += ( r >> 3 ) << 11;
    res += ( g >> 3 ) << 6;
    res += ( b >> 2 );
#endif
#ifdef RGB555
    res += ( r >> 3 ) << 10;
    res += ( g >> 3 ) << 5;
    res += ( b >> 3 );
#endif
#ifdef RGB565
    res += ( r >> 3 ) << 11;
    res += ( g >> 2 ) << 5;
    res += ( b >> 3 );
    #ifdef PALMOS
    #ifndef DIRECTDRAW
    res = ( res >> 8 ) + ( ( res & 255 ) << 8 );
    #endif
    #endif
#endif

#endif
    
#ifdef COLOR32BITS
    #ifdef OPENGL
	res += r;
	res += g << 8;
	res += b << 16;
    #else
	res += r << 16;
	res += g << 8;
	res += b;
    #endif
#endif
    
    return res;
}

WM_INLINE COLOR get_color_from_string( UTF8_CHAR* str )
{
    uint c = 0;
    if( mem_strlen( str ) != 7 ) return COLORMASK;
    for( int i = 1; i < 7; i++ )
    {
	c <<= 4;
	if( str[ i ] < 58 ) c += str[ i ] - '0';
	else if( str[ i ] > 64 && str[ i ] < 91 ) c += str[ i ] - 'A' + 10;
	else c += str[ i ] - 'a' + 10;
    }
    return get_color( ( c >> 16 ) & 255, ( c >> 8 ) & 255, c & 255 );
}

#if defined(PALMOS) && !defined(DIRECTDRAW)
#define PALMOS_COLOR_CONV c = ( c >> 8 ) + ( ( c & 255 ) << 8 );
#else
#define PALMOS_COLOR_CONV /**/
#endif

#ifdef RGB556
#define GET_RED16 \
    res = ( ( c >> 11 ) << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif
#ifdef RGB555
#define GET_RED16 \
    res = ( ( c >> 10 ) << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif
#ifdef RGB565
#define GET_RED16 \
    res = ( ( c >> 11 ) << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif

WM_INLINE int red( COLOR c )
{
    int res = 0;
    
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = c;
    #else
	res = ( c << 5 ) & 255;
	if( res ) res |= 0x1F;
    #endif
#endif
    
#ifdef COLOR16BITS
    PALMOS_COLOR_CONV;
    GET_RED16;
#endif

#ifdef COLOR32BITS
    #ifdef OPENGL
	res = c & 255;
    #else
	res = ( c >> 16 ) & 255;
    #endif
#endif

    return res;
}

#ifdef RGB556
#define GET_GREEN16 \
    res = ( ( c >> 6 ) << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif
#ifdef RGB555
#define GET_GREEN16 \
    res = ( ( c >> 5 ) << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif
#ifdef RGB565
#define GET_GREEN16 \
    res = ( ( c >> 5 ) << 2 ) & 0xFC; \
    if( res ) res |= 3;
#endif

WM_INLINE int green( COLOR c )
{
    int res = 0;
    
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = c;
    #else
	res = ( c << 2 ) & 0xE0;
	if( res ) res |= 0x1F;
    #endif
#endif
    
#ifdef COLOR16BITS
    PALMOS_COLOR_CONV;
    GET_GREEN16;
#endif

#ifdef COLOR32BITS
    res = ( c >> 8 ) & 255;
#endif
    
    return res;
}

#ifdef RGB556
#define GET_BLUE16 \
    res = ( c << 2 ) & 0xFC; \
    if( res ) res |= 3;
#endif
#ifdef RGB555
#define GET_BLUE16 \
    res = ( c << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif
#ifdef RGB565
#define GET_BLUE16 \
    res = ( c << 3 ) & 0xF8; \
    if( res ) res |= 7;
#endif

WM_INLINE int blue( COLOR c )
{
    int res = 0;
    
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	res = c;
    #else
	res = ( c & 192 );
	if( res ) res |= 0x3F;
    #endif
#endif
    
#ifdef COLOR16BITS
    PALMOS_COLOR_CONV;
    GET_BLUE16;
#endif
    
#ifdef COLOR32BITS
    #ifdef OPENGL
	res = ( c >> 16 ) & 255;
    #else
	res = c & 255;
    #endif
#endif
    
    return res;
}

WM_INLINE COLOR fast_blend( COLOR c1, COLOR c2, uchar val )
{
    uint ival = 256 - val;
#ifdef COLOR32BITS
    uint v1_1 = c1 & 0xFF00FF;
    uint v1_2 = c1 & 0x00FF00;
    uint v2_1 = c2 & 0xFF00FF;
    uint v2_2 = c2 & 0x00FF00;
    uint res = 
	( ( ( ( v1_1 * ival ) + ( v2_1 * val ) ) >> 8 ) & 0xFF00FF ) |
	( ( ( ( v1_2 * ival ) + ( v2_2 * val ) ) >> 8 ) & 0x00FF00 );
    return res;
#endif
#ifdef COLOR16BITS
    #ifdef PALMOS
    #ifndef DIRECTDRAW
	c1 = ( c1 >> 8 ) + ( ( c1 & 255 ) << 8 );
	c2 = ( c2 >> 8 ) + ( ( c2 & 255 ) << 8 );
    #endif
    #endif
    #ifdef FAST_BLEND
        ival >>= 3;
        val >>= 3;
        #ifdef RGB556
            uint res = 
                ( ( ( ( ( c1 & 0xF83F ) * ival ) + ( ( c2 & 0xF83F ) * val ) ) >> 5 ) & 0xF83F ) |
                ( ( ( ( ( c1 & 0x7C0 ) * ival ) + ( ( c2 & 0x7C0 ) * val ) ) >> 5 ) & 0x7C0 );
        #endif
        #ifdef RGB555
            uint res = 
                ( ( ( ( ( c1 & 0x7C1F ) * ival ) + ( ( c2 & 0x7C1F ) * val ) ) >> 5 ) & 0x7C1F ) |
                ( ( ( ( ( c1 & 0x1E0 ) * ival ) + ( ( c2 & 0x1E0 ) * val ) ) >> 5 ) & 0x1E0 );
        #endif
        #ifdef RGB565
            uint res = 
                ( ( ( ( ( c1 & 0xF81F ) * ival ) + ( ( c2 & 0xF81F ) * val ) ) >> 5 ) & 0xF81F ) |
                ( ( ( ( ( c1 & 0x7E0 ) * ival ) + ( ( c2 & 0x7E0 ) * val ) ) >> 5 ) & 0x7E0 );
        #endif
    #else
        #ifdef RGB556
            int r1 = ( c1 >> 8 ) | 7;
            int g1 = ( ( c1 >> 3 ) & 255 ) | 7;
            int b1 = ( ( c1 << 2 ) & 255 ) | 3;
            int r2 = ( c2 >> 8 ) | 7;
            int g2 = ( ( c2 >> 3 ) & 255 ) | 7;
            int b2 = ( ( c2 << 2 ) & 255 ) | 3;
        #endif
        #ifdef RGB555
            int r1 = ( c1 >> 7 ) | 7;
            int g1 = ( ( c1 >> 2 ) & 255 ) | 7;
            int b1 = ( ( c1 << 3 ) & 255 ) | 7;
            int r2 = ( c2 >> 7 ) | 7;
            int g2 = ( ( c2 >> 2 ) & 255 ) | 7;
            int b2 = ( ( c2 << 3 ) & 255 ) | 7;
        #endif
        #ifdef RGB565
            int r1 = ( c1 >> 8 ) | 7;
            int g1 = ( ( c1 >> 3 ) & 255 ) | 3;
            int b1 = ( ( c1 << 3 ) & 255 ) | 7;
            int r2 = ( c2 >> 8 ) | 7;
            int g2 = ( ( c2 >> 3 ) & 255 ) | 3;
            int b2 = ( ( c2 << 3 ) & 255 ) | 7;
        #endif
            int r3 = r1 * ival + r2 * val;
            int g3 = g1 * ival + g2 * val;
            int b3 = b1 * ival + b2 * val;
        #ifdef RGB556
            r3 >>= 8 + 3;
            g3 >>= 8 + 3;
            b3 >>= 8 + 2;
            uint res = ( r3 << 11 ) | ( g3 << 6 ) | b3;
        #endif
        #ifdef RGB555
            r3 >>= 8 + 3;
            g3 >>= 8 + 3;
            b3 >>= 8 + 3;
            uint res = ( r3 << 10 ) | ( g3 << 5 ) | b3;
        #endif
        #ifdef RGB565
            r3 >>= 8 + 3;
            g3 >>= 8 + 2;
            b3 >>= 8 + 3;
            uint res = ( r3 << 11 ) | ( g3 << 5 ) | b3;
        #endif
    #endif
    #if defined(PALMOS) && !defined(DIRECTDRAW)
	return ( res >> 8 ) + ( ( res & 255 ) << 8 );
    #else    
	return res;
    #endif
#endif
#ifdef COLOR8BITS
    #ifdef GRAYSCALE
	return (COLOR)( ( c1 * ival + c2 * val ) >> 8 );
    #else
	int b1 = c1 >> 5; b1 &= ~1; if( b1 ) b1 |= 1; //because B is 2 bits only
	int g1 = ( c1 >> 3 ) & 7; 
	int r1 = c1 & 7; 
	int b2 = c2 >> 5; b2 &= ~1; if( b2 ) b2 |= 1;
	int g2 = ( c2 >> 3 ) & 7; 
	int r2 = c2 & 7; 
	int r3 = r1 * ival + r2 * val;
	int g3 = g1 * ival + g2 * val;
	int b3 = b1 * ival + b2 * val;
	b3 >>= 9;
	g3 >>= 8;
	r3 >>= 8;
	return ( b3 << 6 ) | ( g3 << 3 ) | r3;
    #endif
#endif
    return 0;
}

WM_INLINE COLOR blend( COLOR c1, COLOR c2, int val )
{
    if( val <= 0 ) return c1;
    if( val >= 255 ) return c2;
    return fast_blend( c1, c2, (uchar)val );
}

//###################################
//## FONTS:                        ##
//###################################

#define utf32_to_win1251( dest, src ) \
{ \
    while( 1 ) \
    { \
	if( src > 0x7F ) \
	{ \
    	    if( ( src >> 8 ) == 0x04 ) \
    	    { \
    		/* Russian: */ \
		UTF32_CHAR src_t = src & 255; \
    		if( src_t >= 0x10 && src_t <= 0x4F ) \
		{ \
	    	    dest = src_t + 0xB0; \
		    break; \
		} \
	    } \
	} \
	else \
	{ \
	    if( src >= 0x10 ) { dest = src; break; } \
	} \
	dest = '*'; \
	break; \
    } \
}

//###################################
//## IMAGE FUNCTIONS:              ##
//###################################

sundog_image* new_image( 
    int xsize, 
    int ysize, 
    void* src,
    int src_x,
    int src_y,
    int src_xsize,
    int src_ysize,
    int flags );
void flush_image( sundog_image* img );
sundog_image* resize_image( int resize_flags, int new_xsize, int new_ysize, sundog_image* img );
void remove_image( sundog_image* img );
bool image_can_be_redrawn_safely( sundog_image* img );

//###################################
//## WBD FUNCTIONS:                ##
//###################################

#define IMG_PREC 11 /* fixed point precision */

//See code in wbd.cpp
void wbd_lock( WINDOWPTR win );
void wbd_unlock( void );
void wbd_draw( void );

void draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager *wm );
void draw_line_with_transp( int x1, int y1, int t1, int x2, int y2, int t2, COLOR color );
void draw_frame( int x, int y, int x_size, int y_size, COLOR color, window_manager *wm );
void draw_box( int x, int y, int x_size, int y_size, COLOR color, window_manager *wm );
void draw_hgradient( int x, int y, int xsize, int ysize, COLOR c, int t1, int t2 );
void draw_vgradient( int x, int y, int xsize, int ysize, COLOR c, int t1, int t2 );
void draw_glass_rect( int x, int y, int xsize, int ysize, int parts );
void draw_corners( int x, int y, int xsize, int ysize, int size, int len, COLOR c );
void draw_corners2( int x, int y, int xsize, int ysize, int size, int len, COLOR c );
void draw_polygon( sundog_polygon* p );
int draw_char( UTF32_CHAR c, int x, int y, window_manager *wm );
void draw_string( const UTF8_CHAR *str, int x, int y, window_manager *wm );
void draw_string_limited( const UTF8_CHAR* str, int x, int y, int limit, window_manager* wm );
void draw_pixel( int x, int y, COLOR color, window_manager *wm );
void draw_image_scaled(
    int src_x, int src_y, //fixed point
    int src_xsize, int src_ysize, //fixed point
    int dest_x, int dest_y,
    int dest_xsize, int dest_ysize,
    sundog_image *img,
    window_manager *wm );

int wbd_char_x_size( UTF32_CHAR c, window_manager* wm );
int wbd_char_y_size( window_manager* wm );
int wbd_string_size( const UTF8_CHAR* str, window_manager* wm );

//###################################
//## STANDARD WINDOW HANDLERS:     ##
//###################################

int null_handler( sundog_event* evt, window_manager* wm );
int desktop_handler( sundog_event* evt, window_manager* wm );
extern int DIVIDER_VSCROLL_MIN;
extern int DIVIDER_VSCROLL_MAX;
extern bool VERTICAL_DIVIDER;
extern bool DIVIDER_WITH_TIME;
int divider_handler( sundog_event* evt, window_manager* wm );
void bind_divider_to( WINDOWPTR win, WINDOWPTR bind_to, int bind_num, window_manager* wm );
void set_divider_push( WINDOWPTR win, WINDOWPTR push_win, window_manager* wm );
void set_divider_vscroll_parameters( WINDOWPTR win, int min, int max, window_manager* wm );
int get_divider_vscroll_value( WINDOWPTR win, window_manager* wm );
int label_handler( sundog_event* evt, window_manager* wm );
extern int TEXT_ZOOM;
extern bool TEXT_NO_VIRTUAL_KEYBOARD;
extern bool TEXT_CALL_HANDLER_ON_ANY_CHANGES;
extern bool TEXT_RO;
extern int CREATE_NUMERIC_TEXT;
extern int CREATE_NUMERIC_TEXT_MIN;
extern int CREATE_NUMERIC_TEXT_MAX;
extern bool NUMERIC_TEXT_HIDE_ZERO;
int text_handler( sundog_event* evt, window_manager* wm );
void text_set_text( WINDOWPTR win, const UTF8_CHAR* text, window_manager* wm );
void text_set_cursor_position( WINDOWPTR win, int cur_pos, window_manager* wm );
void text_set_value( WINDOWPTR win, int val, window_manager* wm );
UTF8_CHAR* text_get_text( WINDOWPTR win, window_manager* wm );
int text_get_cursor_position( WINDOWPTR win, window_manager* wm );
int text_get_value( WINDOWPTR win, window_manager* wm );
void text_changed( WINDOWPTR win, window_manager* wm );
void text_set_zoom( WINDOWPTR win, int zoom, window_manager* wm );
extern sundog_image* CREATE_BITMAP_BUTTON;
extern int CREATE_BITMAP_BUTTON_XSIZE;
extern int CREATE_BITMAP_BUTTON_YSIZE;
extern int CREATE_BITMAP_BUTTON_XOFF1;
extern int CREATE_BITMAP_BUTTON_YOFF1;
extern int CREATE_BITMAP_BUTTON_XOFF2;
extern int CREATE_BITMAP_BUTTON_YOFF2;
extern const UTF8_CHAR* CREATE_BUTTON_WITH_TEXT_MENU;
extern bool CREATE_BUTTON_WITH_AUTOREPEAT;
extern char CREATE_FLAT_BUTTON;
extern int (*BUTTON_END_HANDLER)( void*, WINDOWPTR );
extern void* BUTTON_END_HANDLER_DATA;
extern uchar BUTTON_FLAGS;
extern const UTF8_CHAR text_up[ 2 ];
extern const UTF8_CHAR text_down[ 2 ];
extern const UTF8_CHAR text_left[ 2 ];
extern const UTF8_CHAR text_right[ 2 ];
#define BUTTON_FLAG_HANDLER_CALLED_FROM_TIMER	1
#define BUTTON_FLAG_AUTOBACKCOLOR		2
#define BUTTON_FLAG_SHOW_VALUE			4
#define BUTTON_FLAG_LEFT_ALIGNMENT_ON_OVERFLOW	8
int button_handler( sundog_event* evt, window_manager* wm );
void button_set_menu( WINDOWPTR win, const UTF8_CHAR* menu, window_manager* wm );
void button_set_menu_val( WINDOWPTR win, int val, window_manager* wm );
int button_get_menu_val( WINDOWPTR win, window_manager* wm );
void button_set_text( WINDOWPTR win, UTF8_CHAR* text, window_manager* wm );
void button_set_text_color( WINDOWPTR win, COLOR c, window_manager* wm );
void button_set_text_transp( WINDOWPTR win, uchar transp, window_manager* wm );
int button_get_text_transp( WINDOWPTR win, window_manager* wm );
void button_set_bitmap(
    WINDOWPTR win,
    sundog_image* img,
    int img_xsize,
    int img_ysize,
    int img_xoff1,
    int img_yoff1,
    int img_xoff2,
    int img_yoff2
);
void button_set_flags( WINDOWPTR win, uchar flags );
uchar button_get_flags( WINDOWPTR win );
extern bool CREATE_NUMBERED_LIST;
extern bool CREATE_SMOOTH_LIST;
extern bool CREATE_LIST_WITHOUT_SCROLLBAR;
extern bool CREATE_LIST_WITHOUT_EXTENSIONS;
#define LIST_ACTION_UPDOWN	1
#define LIST_ACTION_ENTER	2
#define LIST_ACTION_CLICK	3
#define LIST_ACTION_RCLICK	4
#define LIST_ACTION_DOUBLECLICK	5
#define LIST_ACTION_ESCAPE	6
int list_handler( sundog_event *evt, window_manager *wm );
list_data* list_get_data( WINDOWPTR win, window_manager *wm );
int list_get_last_action( WINDOWPTR win, window_manager *wm );
void list_select_item( WINDOWPTR win, int item_num, window_manager *wm );
extern bool CREATE_VERTICAL_SCROLLBAR;
extern bool CREATE_REVERSE_SCROLLBAR;
extern bool CREATE_COMPACT_SCROLLBAR;
extern bool CREATE_FLAT_SCROLLBAR;
extern int CREATE_HEX_SCROLLBAR;
extern int (*SCROLLBAR_BEGIN_HANDLER)( void*, WINDOWPTR, int );
extern int (*SCROLLBAR_END_HANDLER)( void*, WINDOWPTR, int );
extern void* SCROLLBAR_BEGIN_HANDLER_DATA;
extern void* SCROLLBAR_END_HANDLER_DATA;
void draw_scrollbar_horizontal_selection( WINDOWPTR win, int x );
void draw_scrollbar_vertical_selection( WINDOWPTR win, int y );
int scrollbar_handler( sundog_event *evt, window_manager *wm );
void scrollbar_set_parameters( WINDOWPTR win, int cur, int max_value, int page_size, int step_size, window_manager *wm );
void scrollbar_set_value( WINDOWPTR win, int val, window_manager *wm );
int scrollbar_get_value( WINDOWPTR win, window_manager *wm );
void scrollbar_set_name( WINDOWPTR win, const UTF8_CHAR* name, window_manager* wm );
void scrollbar_set_values( WINDOWPTR win, const UTF8_CHAR* values, window_manager* wm );
void scrollbar_set_showing_offset( WINDOWPTR win, int offset, window_manager *wm );
void scrollbar_set_hex_mode( WINDOWPTR win, int hex, window_manager *wm );
int keyboard_handler( sundog_event *evt, window_manager *wm );
void hide_keyboard_for_text_window( window_manager* wm );
void show_keyboard_for_text_window( WINDOWPTR text, window_manager *wm );
enum files_preview_status
{
    FPREVIEW_NONE = 0,
    FPREVIEW_OPEN,
    FPREVIEW_CLOSE,
    FPREVIEW_FILE_SELECTED,
};
struct files_preview_data
{
    files_preview_status status;
    UTF8_CHAR* name;
    WINDOWPTR win;
};
extern const UTF8_CHAR* FILES_PROPS;
extern const UTF8_CHAR* FILES_MASK;
extern const UTF8_CHAR* FILES_PREVIEW;
extern int (*FILES_PREVIEW_HANDLER)( files_preview_data*, window_manager* );
extern const UTF8_CHAR* FILES_USER_BUTTON;
extern int (*FILES_USER_BUTTON_HANDLER)( void* user_data, WINDOWPTR win, window_manager* );
extern void* FILES_USER_BUTTON_DATA;
extern UTF8_CHAR* FILES_RESULTED_FILENAME;
extern const UTF8_CHAR* FILES_DEF_FILENAME;
extern bool g_files_preview_flag;
int files_handler( sundog_event *evt, window_manager *wm );
extern new_fdialog_struct* NEW_FILE_DIALOG_STRUCT;
int new_fdialog_handler( sundog_event* evt, window_manager* wm );
extern dialog_item* DIALOG_ITEMS;
extern const UTF8_CHAR* DIALOG_BUTTONS_TEXT;
extern const UTF8_CHAR* DIALOG_TEXT;
extern int* DIALOG_RESULT;
int dialog_handler( sundog_event *evt, window_manager *wm );
extern const UTF8_CHAR* CREATE_TEXT_POPUP;
extern int* SAVE_POPUP_EXIT_TO;
extern int* SAVE_POPUP_RESULT_TO;
int popup_handler( sundog_event *evt, window_manager *wm );

extern WINDOWPTR g_colortheme;
extern int g_prev_color_theme;
extern bool g_color_theme_changed;
int colortheme_handler( sundog_event* evt, window_manager* wm );

extern WINDOWPTR g_ui_scale;            
extern bool g_ui_scale_changed;
int ui_scale_handler( sundog_event* evt, window_manager* wm );

extern WINDOWPTR g_prefs;
extern bool g_prefs_restart_request;
extern const UTF8_CHAR* g_prefs_section_names[ 32 ];
extern int g_prefs_sections;
extern void* g_prefs_section_handlers[ 32 ];
extern int g_pref_section_ysize;
extern int g_prefs_flags;
#define PREFS_FLAG_NO_COLOR_THEME		1
#define PREFS_FLAG_NO_CONTROL_TYPE		2
int prefs_ui_handler( sundog_event* evt, window_manager* wm );
int prefs_audio_handler( sundog_event* evt, window_manager* wm );
int prefs_handler( sundog_event* evt, window_manager* wm );

//###################################
//## DIALOGS & BUILT-IN WINDOWS:   ##
//###################################

extern void (*fdialog_open_wifi)( void );
UTF8_CHAR* dialog_open_file( const UTF8_CHAR* name, const UTF8_CHAR* mask, const UTF8_CHAR* id, const UTF8_CHAR* def_filename );
int dialog( const UTF8_CHAR* name, const UTF8_CHAR* buttons );
int popup_menu( const UTF8_CHAR* name, const UTF8_CHAR* items, int x, int y, COLOR color );
void prefs_clear( void );
void prefs_add_sections( const UTF8_CHAR** names, void** handlers, int num );
void prefs_add_default_sections( void );
void prefs_open( void );
void colortheme_open( void );
void ui_scale_open( void );

//###################################
//## DEVICE DEPENDENT:             ##
//###################################

void device_event_handler( window_manager* );
int device_start( const UTF8_CHAR*, int xsize, int ysize, window_manager* ); 
                                                  //device start (before main loop)
void device_end( window_manager* );	          //device end   (after main loop)

void device_draw_line( int x1, int y1, int x2, int y2, COLOR color );
void device_draw_line_with_transp( int x1, int y1, int t1, int x2, int y2, int t2, COLOR color );
void device_draw_box( int x, int y, int xsize, int ysize, COLOR color );
void device_draw_polygon( sundog_polygon* p );
void device_draw_bitmap( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    int src_x, int src_y,
    sundog_image* data );
void device_draw_bitmap_scaled( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    float src_x, float src_y,
    float src_xs, float src_ys,
    sundog_image* data );
void device_screen_lock( WINDOWPTR win, window_manager* wm );
void device_screen_unlock( WINDOWPTR win, window_manager* wm );
void device_vsync( bool vsync );
void device_redraw_framebuffer( void );

#ifdef OPENGL
void gl_lock( void );
void gl_unlock( void );
void gl_resize( window_manager* wm );
#else
#define gl_lock() {}
#define gl_unlock() {}
#define gl_resize() {}
#endif

#endif
