package nightradio.pixilang;

import android.app.NativeActivity;
import android.content.Context;
import android.content.Intent;
import android.net.Uri;
import android.os.PowerManager;
import android.util.Log;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileInputStream;

public class MyNativeActivity extends NativeActivity 
{
	Boolean wlInit;
	PowerManager pm;
	PowerManager.WakeLock wl;
	
    public MyNativeActivity() 
    {
    	super();
    	wlInit = false;
    }
    
    public String GetInternalCacheDir() 
    {
    	try { 
    		File f = this.getApplicationContext().getCacheDir();
    		return f.toString();
    	}
    	catch( Exception e ) 
    	{
    		Log.e( "GetInternalCacheDir", "getCacheDir() error", e ); 
    		return null;
    	}
    }

    public String GetInternalFilesDir() 
    {
    	try { 
    		File f = this.getApplicationContext().getFilesDir();    	
    		return f.toString();
    	}
    	catch( Exception e ) 
    	{
    		Log.e( "GetInternalFilesDir", "getFilesDir() error", e ); 
    		return null;
    	}
    }

    public String GetExternalCacheDir() 
    {
    	try {
    		File f = this.getApplicationContext().getExternalCacheDir();    	
    		return f.toString();
    	}
    	catch( Exception e ) 
    	{
    		Log.e( "GetExternalCacheDir", "getExternalCacheDir() error", e ); 
    		return null;
    	}
    }
    
    public String GetExternalFilesDir() 
    {
    	try {
    		File f = this.getApplicationContext().getExternalFilesDir( null );    	
    		return f.toString();
    	}
    	catch( Exception e ) 
    	{
    		Log.e( "GetExternalFilesDir", "getExternalFilesDir() error", e ); 
    		return null;
    	}
    }
    
    public int WakeLock()
    {
    	if( !wlInit )
    	{
    		pm = (PowerManager)getSystemService( Context.POWER_SERVICE );
    		wl = pm.newWakeLock( PowerManager.SCREEN_DIM_WAKE_LOCK, "WakeLock" );
    		wlInit = true;
    	}
    	if( wlInit ) wl.acquire();
    	return 0;
    }
    
    public int WakeUnlock()
    {
    	if( wlInit ) wl.release();
    	return 0;
    }
    
    public int CopyResources()
    {
    	int rv = 0;
    	int changes = 0;
    	int len1 = 0;
    	int len2 = 0;
    	int noExternalHash = 0;
    	BufferedInputStream h1 = null;
    	FileInputStream h2 = null;
    	
    	//Read internal hash:
    	h1 = new BufferedInputStream( getResources().openRawResource( R.raw.hash ) );
    	byte[] buf1 = new byte[ 128 ];
    	try { 
    		len1 = h1.read( buf1 );
    		h1.close();
    	} catch( Exception e ) { 
	    	Log.e( "CopyResources", "Open hash (local)", e ); 
	    	return -1;
	    }
    	
    	//Try to open external hash:
    	String externalFilesDir = GetExternalFilesDir();
    	if( externalFilesDir == null )
    	{
    		Log.e( "CopyResources", "Can't get external files dir" );
    		return -2;
    	}
    	try {
    		h2 = new FileInputStream( externalFilesDir + "/hash" );
    	} catch( FileNotFoundException e ) {
    		Log.i( "CopyResources", "Hash not found on external storage" );
    		noExternalHash = 1;
    		changes++;
    	}
    	
    	if( noExternalHash == 0 )
    	{
    		//Compare internal and external hashes:
    		try {
    			byte[] buf2 = new byte[ 128 ];
    			len2 = h2.read( buf2 );
    			if( len1 != len2 ) changes++;
    			for( int i = 0; i < len1; i++ )
        		{
        			if( buf1[ i ] != buf2[ i ] ) changes++;
        		}
    			h2.close();
    		} catch( Exception e ) { 
    	    	Log.e( "CopyResources", "Compare hashes", e ); 
    	    }
    	}
    	
    	if( changes > 0 )
    	{   		
            rv = 0;
            	
            //Copy internal hash to external:
            try {
    			FileOutputStream fout = new FileOutputStream( externalFilesDir + "/hash" );
    			fout.write( buf1, 0, len1 );
    			fout.close();
    		} catch( Exception e ) { 
    	    	Log.e( "CopyResources", "Write hash (external)", e ); 
    	    }

   			//Unpack all files:
   			BufferedInputStream zipInputStream = new BufferedInputStream( getResources().openRawResource( R.raw.files ) );
   			String unzipLocation = externalFilesDir + "/";
   			Decompress d = new Decompress( zipInputStream, unzipLocation );
   			d.unzip();
    	}
    	else
    	{
    		Log.i( "CopyResources", "All files in place" );
    	}
    	
    	return rv;
    }
    
    public int OpenURL( String s )
    {
    	Intent i = new Intent( Intent.ACTION_VIEW );  
    	i.setData( Uri.parse( s ) );  
    	startActivity( i );
    	return 0;
    }
}