/*
    pixilang_vm_builtin_fns.cpp
    This file is part of the Pixilang programming language.
    
    [ MIT license ]

    Copyright (c) 2006 - 2012, Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to 
    deal in the Software without restriction, including without limitation the 
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in 
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
*/

#include "pixilang_vm_builtin_fns.h"

const UTF8_CHAR* g_pix_fn_names[ FN_NUM ] = 
{
    "new",
    "remove",
    "remove_with_alpha",
    "resize",
    "clean",
    "clone",
    "copy",
    "get_size",
    "get_xsize",
    "get_ysize",
    "get_esize",
    "get_type",
    "get_flags",
    "set_flags",
    "reset_flags",
    "get_prop",
    "set_prop",
    "remove_props",
    "convert_type",
    
    "num_to_str",
    "str_to_num",
    
    "load",
    "save",
    "get_real_path",
    "new_flist",
    "remove_flist",
    "get_flist_name",
    "get_flist_type",
    "flist_next",
    "get_file_size",
    "remove_file",
    "rename_file",
    "copy_file",
    "create_directory",
    
    "frame",
    "vsync",
    "set_screen",
    "get_screen",
    "set_zbuf",
    "get_zbuf",
    "get_color",
    "get_red",
    "get_green",
    "get_blue",
    "get_blend",
    "transp",
    "get_transp",
    "clear",
    "clone",
    "dot",
    "dot3d",
    "get_dot",
    "get_dot3d",
    "line",
    "line3d",
    "box",
    "fbox",
    "pixi",
    "triangles3d",
    "sort_triangles3d",
    "set_key_color",
    "get_key_color",
    "set_alpha",
    "get_alpha",
    "print",
    "get_text_xsize",
    "get_text_ysize",
    "set_font",
    "get_font",
    "effector",
    "split_rgb",

    "set_gl_callback",
    "remove_gl_data",
    
    "unpack_frame",
    "pack_frame",
    "create_anim",
    "remove_anim",
    "clone_frame",
    "remove_frame",
    "play",
    "stop",
    
    "t_reset",
    "t_rotate",
    "t_translate",
    "t_scale",
    "t_push_matrix",
    "t_pop_matrix",
    "t_get_matrix",
    "t_set_matrix",
    "t_mul_matrix",
    "t_point",
    
    "set_audio_callback",
    "get_note_freq",
    
    "midi_open_client",
    "midi_close_client",
    "midi_get_device",
    "midi_open_port",
    "midi_reopen_port",
    "midi_close_port",
    "midi_get_event",
    "midi_get_event_time",
    "midi_next_event",
    "midi_send_event",
    
    "start_timer",
    "get_timer",
    "get_year",
    "get_month",
    "get_day",
    "get_hours",
    "get_minutes",
    "get_seconds",
    "get_ticks",
    "get_tps",
    "sleep",
    
    "get_event",
    "set_quit_action",
    
    "thread_create",
    "thread_destroy",
    "mutex_create",
    "mutex_destroy",
    "mutex_lock",
    "mutex_trylock",
    "mutex_unlock",
    
    "acos",
    "acosh",
    "asin",
    "asinh",
    "atan",
    "atanh",
    "ceil",
    "cos",
    "cosh",
    "exp",
    "exp2",
    "expm1",
    "abs",
    "floor",
    "mod",
    "log",
    "log2",
    "log10",
    "pow",
    "sin",
    "sinh",
    "sqrt",
    "tan",
    "tanh",
    "rand",
    "rand_seed",
    
    "op_cn",
    "op_cc",
    "op_ccn",
    "generator",
    "wavetable_generator",
    "sampler",
    "envelope2p",
    "fft",
    "replace_values",
    
    "file_dialog",
    "prefs_dialog",

    "open_url",
    
    "dlopen",
    "dlclose",
    "dlsym",
    "dlcall",

    "fopen",
    "fclose",
    "fputc",
    "fputs",
    "fwrite",
    "fgetc",
    "fgets",
    "fread",
    "feof",
    "fflush",
    "fseek",
    "ftell",
    "strcat",
    "strcmp"
    "strlen",
    "strstr",
    "sprintf",
    "printf",
    "system",
    "argc",
    "argv",
    "exit",
    
    "system_copy",
    "system_paste",
    "open_webserver",
};

pix_builtin_fn g_pix_fns[ FN_NUM ] = 
{
    fn_new_pixi,
    fn_remove_pixi,
    fn_remove_pixi_with_alpha,
    fn_resize_pixi,
    fn_clean_pixi,
    fn_clone_pixi,
    fn_copy_pixi,
    fn_get_pixi_info, //size
    fn_get_pixi_info, //xsize
    fn_get_pixi_info, //ysize
    fn_get_pixi_info, //esize
    fn_get_pixi_info, //type
    fn_get_pixi_flags,
    fn_set_pixi_flags,
    fn_reset_pixi_flags,
    fn_get_pixi_prop_OR_set_pixi_prop, //get_prop
    fn_get_pixi_prop_OR_set_pixi_prop, //set_prop
    fn_remove_pixi_props,
    fn_convert_pixi_type,
    
    fn_num_to_string,
    fn_string_to_num,
    
    fn_load,
    fn_save,
    fn_get_real_path,
    fn_new_flist,
    fn_remove_flist,
    fn_get_flist_name,
    fn_get_flist_type,
    fn_flist_next,
    fn_get_file_size,
    fn_remove_file,
    fn_rename_file,
    fn_copy_file,
    fn_create_directory,
    
    fn_frame,
    fn_vsync,
    fn_set_screen,
    fn_get_screen,
    fn_set_zbuf,
    fn_get_zbuf,
    fn_clear_zbuf,
    fn_get_color,
    fn_get_red,
    fn_get_green,
    fn_get_blue,
    fn_get_blend,
    fn_transp,
    fn_get_transp,
    fn_clear,
    fn_dot,
    fn_dot, //3d
    fn_get_dot,
    fn_get_dot, //3d
    fn_line,
    fn_line, //3d
    fn_box,
    fn_box, //fbox
    fn_pixi,
    fn_triangles3d,
    fn_sort_triangles3d,
    fn_set_key_color,
    fn_get_key_color,
    fn_set_alpha,
    fn_get_alpha,
    fn_print,
    fn_get_text_xsize,
    fn_get_text_ysize,
    fn_set_font,
    fn_get_font,
    fn_effector,
    fn_split_rgb,
    
    fn_set_gl_callback,
    fn_remove_gl_data,
    
    fn_pixi_unpack_frame,
    fn_pixi_pack_frame,
    fn_pixi_create_anim,
    fn_pixi_remove_anim,
    fn_pixi_clone_frame,
    fn_pixi_remove_frame,
    fn_pixi_play,
    fn_pixi_stop,
    
    fn_t_reset,
    fn_t_rotate,
    fn_t_translate,
    fn_t_scale,
    fn_t_push_matrix,
    fn_t_pop_matrix,
    fn_t_get_matrix,
    fn_t_set_matrix,
    fn_t_mul_matrix,
    fn_t_point,
    
    fn_set_audio_callback,
    fn_enable_audio_input,
    fn_get_note_freq,
    
    fn_midi_open_client,
    fn_midi_close_client,
    fn_midi_get_device,
    fn_midi_open_port,
    fn_midi_reopen_port,
    fn_midi_close_port,
    fn_midi_get_event,
    fn_midi_get_event_time,
    fn_midi_next_event,
    fn_midi_send_event,
    
    fn_start_timer,
    fn_get_timer,
    fn_get_year,
    fn_get_month,
    fn_get_day,
    fn_get_hours,
    fn_get_minutes,
    fn_get_seconds,
    fn_get_ticks,
    fn_get_tps,
    fn_sleep,
    
    fn_get_event,
    fn_set_quit_action,
    
    fn_thread_create,
    fn_thread_destroy,
    fn_mutex_create,
    fn_mutex_destroy,
    fn_mutex_lock,
    fn_mutex_trylock,
    fn_mutex_unlock,
    
    fn_acos,
    fn_acosh,
    fn_asin,
    fn_asinh,
    fn_atan,
    fn_atanh,
    fn_ceil,
    fn_cos,
    fn_cosh,
    fn_exp,
    fn_exp2,
    fn_expm1,
    fn_abs,
    fn_floor,
    fn_mod,
    fn_log,
    fn_log2,
    fn_log10,
    fn_pow,
    fn_sin,
    fn_sinh,
    fn_sqrt,
    fn_tan,
    fn_tanh,
    fn_rand,
    fn_rand_seed,
    
    fn_op_cn,
    fn_op_cc,
    fn_op_ccn,
    fn_generator,
    fn_wavetable_generator,
    fn_sampler,
    fn_envelope2p,
    fn_fft,
    fn_replace_values,
    
    fn_file_dialog,
    fn_prefs_dialog,

    fn_system_copy_OR_open_url, //open_url
    
    fn_dlopen,
    fn_dlclose,
    fn_dlsym,
    fn_dlcall,

    fn_fopen,
    fn_fclose,
    fn_fputc,
    fn_fputs,
    fn_fgets_OR_fwrite_OR_fread, //fwrite
    fn_fgetc,
    fn_fgets_OR_fwrite_OR_fread, //fgets
    fn_fgets_OR_fwrite_OR_fread, //fread
    fn_feof_OF_fflush_OR_ftell, //feof
    fn_feof_OF_fflush_OR_ftell, //fflush
    fn_fseek,
    fn_feof_OF_fflush_OR_ftell, //ftell
    fn_strcat,
    fn_strcmp_OR_strstr, //strcmp
    fn_strlen,
    fn_strcmp_OR_strstr, //strstr
    fn_sprintf,
    fn_sprintf,
    fn_system,
    fn_argc,
    fn_argv,
    fn_exit,
    
    fn_system_copy_OR_open_url, //copy
    fn_system_paste,
    fn_system_webserver
};

//
// Containers (memory management)
//

void fn_new_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_INT xsize = 1;
    PIX_INT ysize = 1;
    int type = 32;
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( xsize, 0, PIX_INT );
    if( pars_num >= 2 ) GET_VAL_FROM_STACK( ysize, 1, PIX_INT );
    if( pars_num >= 3 ) GET_VAL_FROM_STACK( type, 2, int );
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_new_container( -1, xsize, ysize, type, 0, 0, vm );
}

void fn_remove_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    if( pars_num >= 1 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_remove_container( cnum, vm );
    }
}

void fn_remove_pixi_with_alpha( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    if( pars_num >= 1 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	PIX_CID a = pix_vm_get_container_alpha( cnum, vm );
	pix_vm_remove_container( a, vm );
	pix_vm_remove_container( cnum, vm );
    }
}

void fn_resize_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 2 ) return;
    PIX_CID cnum;
    PIX_INT xsize = -1;
    PIX_INT ysize = -1;
    int type = -1;
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    GET_VAL_FROM_STACK( xsize, 1, PIX_INT );
    if( pars_num >= 3 ) GET_VAL_FROM_STACK( ysize, 2, PIX_INT );
    if( pars_num >= 4 ) GET_VAL_FROM_STACK( type, 3, int );
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_resize_container( cnum, xsize, ysize, type, vm );
}

void fn_clean_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    if( pars_num > 0 ) 
    {
	PIX_INT offset = 0;
	PIX_INT size = -1;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	if( pars_num > 1 )
	{
	    if( pars_num > 2 ) { GET_VAL_FROM_STACK( offset, 2, PIX_INT ); }
	    if( pars_num > 3 ) { GET_VAL_FROM_STACK( size, 3, PIX_INT ); }
	    pix_vm_clean_container( cnum, stack_types[ sp + 1 ], stack[ sp + 1 ], offset, size, vm );
	}
	else 
	{
	    PIX_VAL v;
	    v.i = 0;
	    pix_vm_clean_container( cnum, 0, v, 0, -1, vm );
	}
    }
}

void fn_clone_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    if( pars_num >= 1 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	stack[ sp + ( pars_num - 1 ) ].i = pix_vm_clone_container( cnum, vm );
    }
    else
    {
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

void fn_copy_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_INT rv = 0;
    
    if( pars_num >= 2 )
    {
	PIX_CID cnum1; //destination
	PIX_CID cnum2; //source
	GET_VAL_FROM_STACK( cnum1, 0, PIX_CID );
	GET_VAL_FROM_STACK( cnum2, 1, PIX_CID );
	if( (unsigned)cnum1 >= (unsigned)PIX_VM_CONTAINERS ) goto copy_end;
	pix_vm_container* cont1 = vm->c[ cnum1 ];
#ifdef OPENGL
	if( cnum2 == PIX_GL_SCREEN )
	{
	    if( cont1->size * g_pix_container_type_sizes[ cont1->type ] <= vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i * vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i * COLORLEN )
	    {
		int format = -1;
		int type = -1;
		if( COLORLEN == 4 ) { format = GL_RGBA; type = GL_UNSIGNED_BYTE; }
		if( COLORLEN == 2 ) { format = GL_RGB; type = GL_UNSIGNED_SHORT_5_6_5; }
		if( format >= 0 && type >= 0 )
		{
		    glReadPixels( 0, 0, vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i, vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i, format, type, cont1->data );
		    COLORPTR src1 = (COLORPTR)cont1->data;
		    COLORPTR src2 = src1 + vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i * ( vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i - 1 );
		    for( int y = 0; y < vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i / 2; y++ )
		    {
			for( int x = 0; x < vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i; x++ )
			{
			    COLOR temp = *src2;
			    *src2 = *src1;
			    *src1 = temp;
			    src1++;
			    src2++;
			}
			src2 -= vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i * 2;
		    }
		}
		goto copy_end;		
	    }
	}
#endif
	if( (unsigned)cnum2 >= (unsigned)PIX_VM_CONTAINERS ) goto copy_end;
	pix_vm_container* cont2 = vm->c[ cnum2 ];
	if( cont1 == 0 || cont2 == 0 ) goto copy_end;
	
	size_t dest_offset;
	size_t src_offset;
	size_t count;
	size_t dest_step;
	size_t src_step;
	if( pars_num >= 3 ) { GET_VAL_FROM_STACK( dest_offset, 2, size_t ); } else { dest_offset = 0; }
	if( pars_num >= 4 ) { GET_VAL_FROM_STACK( src_offset, 3, size_t ); } else { src_offset = 0; }
	if( pars_num >= 5 ) { GET_VAL_FROM_STACK( count, 4, size_t ); } else { count = cont1->size; }
	if( pars_num >= 6 ) { GET_VAL_FROM_STACK( dest_step, 5, size_t ); } else { dest_step = 1; }
	if( pars_num >= 7 ) { GET_VAL_FROM_STACK( src_step, 6, size_t ); } else { src_step = 1; }
	if( dest_offset >= cont1->size ) goto copy_end;
	if( src_offset >= cont2->size ) goto copy_end;
	if( count == 0 ) goto copy_end;
	size_t end_bound = dest_offset + ( count - 1 ) * dest_step;
	if( end_bound >= cont1->size ) 
	    count = ( cont1->size - dest_offset ) / dest_step;
	end_bound = src_offset + ( count - 1 ) * src_step;
	if( end_bound >= cont2->size )
	    count = ( cont2->size - src_offset ) / src_step;
	
	rv = (PIX_INT)count;
	
	if( cont1->type == cont2->type && dest_step == 1 && src_step == 1 )
	{
	    mem_copy( 
		     (char*)cont1->data + dest_offset * g_pix_container_type_sizes[ cont1->type ], 
		     (char*)cont2->data + src_offset * g_pix_container_type_sizes[ cont1->type ], 
		     count * g_pix_container_type_sizes[ cont1->type ] 
		    );
	}
	else 
	{
	    for( PIX_INT i = dest_offset, i2 = src_offset; i < dest_offset + count * dest_step; i += dest_step, i2 += src_step )
	    {
		char val_type;
		PIX_VAL val;
		switch( cont2->type )
		{
		    case PIX_CONTAINER_TYPE_INT8: val_type = 0; val.i = ( (signed char*)cont2->data )[ i2 ]; break;
		    case PIX_CONTAINER_TYPE_INT16: val_type = 0; val.i = ( (signed short*)cont2->data )[ i2 ]; break;
		    case PIX_CONTAINER_TYPE_INT32: val_type = 0; val.i = ( (signed int*)cont2->data )[ i2 ]; break;
#ifdef PIX_INT64_ENABLED
		    case PIX_CONTAINER_TYPE_INT64: val_type = 0; val.i = ( (signed long long*)cont2->data )[ i2 ]; break;
#endif
		    case PIX_CONTAINER_TYPE_FLOAT32: val_type = 1; val.f = ( (float*)cont2->data )[ i2 ]; break;
#ifdef PIX_FLOAT64_ENABLED
		    case PIX_CONTAINER_TYPE_FLOAT64: val_type = 1; val.f = ( (double*)cont2->data )[ i2 ]; break;
#endif
		    case PIX_CONTAINER_TYPE_DYNAMIC: val_type = cont2->data_types[ i ]; val = ( (PIX_VAL*)cont2->data )[ i2 ]; break;
		    default: val_type = 0; val.i = 0; val.f = 0; break;
		}
		switch( cont1->type )
		{
		    case PIX_CONTAINER_TYPE_INT8: if( val_type ) ( (signed char*)cont1->data )[ i ] = val.f; else ( (signed char*)cont1->data )[ i ] = val.i; break;
		    case PIX_CONTAINER_TYPE_INT16: if( val_type ) ( (signed short*)cont1->data )[ i ] = val.f; else ( (signed short*)cont1->data )[ i ] = val.i; break;
		    case PIX_CONTAINER_TYPE_INT32: if( val_type ) ( (signed int*)cont1->data )[ i ] = val.f; else ( (signed int*)cont1->data )[ i ] = val.i; break;
#ifdef PIX_INT64_ENABLED
		    case PIX_CONTAINER_TYPE_INT64: if( val_type ) ( (signed long long*)cont1->data )[ i ] = val.f; else ( (signed long long*)cont1->data )[ i ] = val.i; break;
#endif
		    case PIX_CONTAINER_TYPE_FLOAT32: if( val_type ) ( (float*)cont1->data )[ i ] = val.f; else ( (float*)cont1->data )[ i ] = val.i; break;
#ifdef PIX_FLOAT64_ENABLED
		    case PIX_CONTAINER_TYPE_FLOAT64: if( val_type ) ( (double*)cont1->data )[ i ] = val.f; else ( (double*)cont1->data )[ i ] = val.i; break;
#endif
		    case PIX_CONTAINER_TYPE_DYNAMIC: cont1->data_types[ i ] = val_type; ( (PIX_VAL*)cont1->data )[ i ] = val; break;
		    default: break;
		}
	    }
	}
    }
    
copy_end:

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_get_pixi_info( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    PIX_INT rv = 1;
    if( pars_num >= 1 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	if( vm->c && (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    switch( fn_num )
	    {
		case FN_GET_PIXI_SIZE: rv = c->xsize * c->ysize; break;
		case FN_GET_PIXI_XSIZE: rv = c->xsize; break;
		case FN_GET_PIXI_YSIZE: rv = c->ysize; break;
		case FN_GET_PIXI_ESIZE: rv = g_pix_container_type_sizes[ c->type ]; break;
		case FN_GET_PIXI_TYPE: rv = c->type; break;
	    }
	}
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_get_pixi_flags( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    int flags = 0;
    if( pars_num >= 1 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	if( vm->c && (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    flags = c->flags;
	}
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = flags;
}

void fn_set_pixi_flags( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    int flags;
    if( pars_num >= 2 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( flags, 1, int );
	if( vm->c && (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    volatile int new_flags = c->flags | flags;
	    c->flags = new_flags;
	}
    }
}

void fn_reset_pixi_flags( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID cnum;
    int flags;
    if( pars_num >= 2 ) 
    {
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( flags, 1, int );
	if( vm->c && (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    volatile int new_flags = c->flags & ~flags;
	    c->flags = new_flags;
	}
    }
}

void fn_get_pixi_prop_OR_set_pixi_prop( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 2 ) 
    {
	PIX_CID cnum;
	PIX_CID prop_name;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( prop_name, 1, PIX_CID );
	bool need_to_free = 0;
	UTF8_CHAR* prop_name_str = pix_vm_make_cstring_from_container( prop_name, &need_to_free, vm );
	switch( fn_num )
	{
	    case FN_GET_PIXI_PROP:
		{	    
		    if( pars_num > 2 )
		    {
			stack[ sp + ( pars_num - 1 ) ] = stack[ sp + 2 ];
			stack_types[ sp + ( pars_num - 1 ) ] = stack_types[ sp + 2 ];
		    }
		    else
		    {
			stack_types[ sp + ( pars_num - 1 ) ] = 0;
			stack[ sp + ( pars_num - 1 ) ].i = 0;
		    }
    
		    pix_sym* sym = pix_vm_get_container_property( cnum, prop_name_str, -1, vm );
		    
		    if( sym )
		    {
			if( sym->type == SYMTYPE_NUM_F )
			    stack_types[ sp + ( pars_num - 1 ) ] = 1;
			stack[ sp + ( pars_num - 1 ) ] = sym->val;
		    }

		}
		break;
	    case FN_SET_PIXI_PROP:
		{
		    pix_vm_set_container_property( cnum, prop_name_str, -1, stack_types[ sp + 2 ], stack[ sp + 2 ], vm );
		}
		break;
	}
	if( need_to_free ) mem_free( prop_name_str );
    }
}

void fn_remove_pixi_props( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER; 
   
    if( pars_num >= 1 ) 
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c && c->opt_data )
	{
	    pix_symtab_deinit( &c->opt_data->props );
	}
    }
}

void fn_convert_pixi_type( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    PIX_INT rv = 1;
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	int type;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( type, 1, int );
	
	rv = pix_vm_convert_container_type( cnum, type, vm );
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//
// Working with strings
//
	
void fn_num_to_string( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
		
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	if( vm->c && (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    UTF8_CHAR ts[ 128 ];
	    if( stack_types[ sp + 1 ] == 0 )
	    {
		//int:
		sprintf( ts, "%d", (int)stack[ sp + 1 ].i );
	    }
	    else 
	    {
		//float:
		sprintf( ts, "%f", (float)stack[ sp + 1 ].f );
	    }
	    size_t size = c->size * g_pix_container_type_sizes[ c->type ];
	    PIX_INT ts_len = (PIX_INT)mem_strlen( ts );
	    if( ts_len > size )
	    {
		if( pix_vm_resize_container( cnum, ts_len, 1, PIX_CONTAINER_TYPE_INT8, vm ) ) return;
		size = c->size * g_pix_container_type_sizes[ c->type ];
	    }
	    mem_copy( c->data, ts, ts_len );
	    if( ts_len < size ) ((UTF8_CHAR*)c->data)[ ts_len ] = 0;
	}
    }
}

void fn_string_to_num( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    bool ok = 0;
    if( pars_num >= 1 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	if( vm->c && (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    if( c && c->data )
	    {
		size_t size = c->size * g_pix_container_type_sizes[ c->type ];
		UTF8_CHAR* str = (UTF8_CHAR*)c->data;
		UTF8_CHAR ts[ 128 + 1 ];
		ts[ 0 ] = 0;
		bool float_num = 0;
		int i = 0;
		while( 1 )
		{
		    UTF8_CHAR cc = str[ i ];
		    if( cc == '.' ) float_num = 1;
		    if( cc == 0 ) break;
		    ts[ i ] = cc;
		    i++;
		    if( i == 128 ) break;
		    if( i == size ) break;
		}
		ts[ i ] = 0;
		if( float_num )
		{
		    stack_types[ sp + ( pars_num - 1 ) ] = 1;
		    stack[ sp + ( pars_num - 1 ) ].f = atof( ts );
		    ok = 1;
		}
		else 
		{
		    PIX_INT val = 0;
		    bool minus = 0;
		    for( int i2 = 0; i2 < i; i2++ )
		    {
			UTF8_CHAR cc = ts[ i2 ];
			if( cc == '-' ) 
			{
			    minus = 1;
			}
			else
			{
			    val *= 10;
			    val += cc - '0';
			}
		    }
		    if( minus ) val = -val;
		    stack_types[ sp + ( pars_num - 1 ) ] = 0;
		    stack[ sp + ( pars_num - 1 ) ].i = val;
		    ok = 1;
		}
	    }
	}
    }
    if( ok == 0 )
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = 0;
    }
}

//
// Working with files
//

void fn_load( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num == 0 ) return;
		
    PIX_CID name;
    GET_VAL_FROM_STACK( name, 0, PIX_CID );

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;

    bool need_to_free = 0;
    UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
    if( ts == 0 ) return;
	
    UTF8_CHAR* full_path = pix_compose_full_path( vm->base_path, ts );
    if( full_path )
    {
	stack[ sp + ( pars_num - 1 ) ].i = pix_vm_load( full_path, vm );
	mem_free( full_path );
    }
    
    if( need_to_free ) mem_free( ts );
}

void fn_save( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 3 ) return;
    
    PIX_CID cnum;
    PIX_CID name;
    int format;
    int par1 = 0;
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    GET_VAL_FROM_STACK( name, 1, PIX_CID );
    GET_VAL_FROM_STACK( format, 2, int );
    if( pars_num > 3 ) GET_VAL_FROM_STACK( par1, 3, int );

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;

    bool need_to_free = 0;
    UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
    if( ts == 0 ) return;
    
    UTF8_CHAR* full_path = pix_compose_full_path( vm->base_path, ts );
    if( full_path )
    {
	stack[ sp + ( pars_num - 1 ) ].i = pix_vm_save( cnum, full_path, format, par1, vm );
	mem_free( full_path );
    }
    
    if( need_to_free ) mem_free( ts );
}

void fn_get_real_path( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name;
    pix_vm_container* name_cont;
            
    bool err = 0;
            
    //Get parameters:
    while( 1 )
    {
        if( pars_num < 1 ) { err = 1; break; }
        GET_VAL_FROM_STACK( name, 0, PIX_CID );
        if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
        name_cont = vm->c[ name ];
        break;
    }
        
    //Execute:
    if( err == 0 )
    {
        bool need_to_free = 0;
        UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
        if( ts )
        {
	    UTF8_CHAR* path = v3_make_filename( (const UTF8_CHAR*)ts );
	    if( path == 0 )
	    {
		err = 1;
	    }
	    else
	    {
		int path_len = (int)mem_strlen( path );
        	PIX_CID path_cnum = pix_vm_new_container( -1, path_len, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
        	if( path_cnum >= 0 )
        	{
            	    pix_vm_container* path_cont = vm->c[ path_cnum ];
            	    mem_copy( path_cont->data, path, path_len );
        	}
        	mem_free( path );
    		stack_types[ sp + ( pars_num - 1 ) ] = 0;
    		stack[ sp + ( pars_num - 1 ) ].i = path_cnum;
    	    }
    	    if( need_to_free ) mem_free( ts );
    	}
    }
    
    if( err != 0 )
    {
        stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

struct flist_data
{
    UTF8_CHAR* path;
    UTF8_CHAR* mask;
    find_struct fs;
    UTF8_CHAR* cur_file;
    int cur_type;
};

void fn_new_flist( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    flist_data* f = 0;    
    UTF8_CHAR* path = 0;
    UTF8_CHAR* mask = 0;
    PIX_CID path_cnum = -1;
    PIX_CID mask_cnum = -1;
    
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( path_cnum, 0, PIX_CID );
    if( pars_num >= 2 ) GET_VAL_FROM_STACK( mask_cnum, 1, PIX_CID );
    
    bool need_to_free1 = 0;
    path = pix_vm_make_cstring_from_container( path_cnum, &need_to_free1, vm );
    bool need_to_free2 = 0;
    mask = pix_vm_make_cstring_from_container( mask_cnum, &need_to_free2, vm );
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;
    
    if( path )
    {
	f = (flist_data*)MEM_NEW2( sizeof( flist_data ) );
	mem_zero( f );
	f->path = (UTF8_CHAR*)MEM_NEW2( mem_strlen( path ) + 1 ); f->path[ 0 ] = 0; mem_strcat( f->path, path );
	if( mask )
	{    
	    f->mask = (UTF8_CHAR*)MEM_NEW2( mem_strlen( mask ) + 1 ); f->mask[ 0 ] = 0; mem_strcat( f->mask, mask );
	}
	f->fs.start_dir = (const UTF8_CHAR*)f->path;
	f->fs.mask = (const UTF8_CHAR*)f->mask;
	if( find_first( &f->fs ) )
	{
	    f->cur_file = f->fs.name;
	    f->cur_type = f->fs.type;
	}
	stack[ sp + ( pars_num - 1 ) ].i = pix_vm_new_container( -1, mem_get_size( f ), 1, PIX_CONTAINER_TYPE_INT8, f, 0, vm );
    }
    
    if( need_to_free1 ) mem_free( path );
    if( need_to_free2 ) mem_free( mask );
}

void fn_remove_flist( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;    
    PIX_CID flist_cnum = -1;
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( flist_cnum, 0, PIX_CID );
    flist_data* f = (flist_data*)pix_vm_get_container_data( flist_cnum, vm );
    if( f )
    {
	find_close( &f->fs );
	mem_free( f->path );
	mem_free( f->mask );
	pix_vm_remove_container( flist_cnum, vm );
    }
}

void fn_get_flist_name( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    PIX_CID flist_cnum = -1;
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( flist_cnum, 0, PIX_CID );
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;
    flist_data* f = (flist_data*)pix_vm_get_container_data( flist_cnum, vm );
    if( f && f->cur_file )
    {
	int name_len = (int)mem_strlen( f->cur_file );
        PIX_CID name_cnum = pix_vm_new_container( -1, name_len, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
        if( name_cnum >= 0 )
        {
    	    pix_vm_container* name_cont = vm->c[ name_cnum ];
            mem_copy( name_cont->data, f->cur_file, name_len );
	    stack[ sp + ( pars_num - 1 ) ].i = name_cnum;
        }
    }
}

void fn_get_flist_type( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    PIX_CID flist_cnum = -1;
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( flist_cnum, 0, PIX_CID );
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;
    flist_data* f = (flist_data*)pix_vm_get_container_data( flist_cnum, vm );
    if( f )
    {
	stack[ sp + ( pars_num - 1 ) ].i = f->cur_type;
    }
}

void fn_flist_next( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    PIX_CID flist_cnum = -1;
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( flist_cnum, 0, PIX_CID );
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = 0;
    flist_data* f = (flist_data*)pix_vm_get_container_data( flist_cnum, vm );
    if( f )
    {
	if( find_next( &f->fs ) )
	{
	    f->cur_file = f->fs.name;
	    f->cur_type = f->fs.type;
	    stack[ sp + ( pars_num - 1 ) ].i = 1;
	}
    }
}

void fn_get_file_size( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name;
    pix_vm_container* name_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 1 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name, 0, PIX_CID );
	if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name ] == 0 ) { err = 1; break; }
	name_cont = vm->c[ name ];
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free = 0;
	UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
	
	UTF8_CHAR* full_path = pix_compose_full_path( vm->base_path, ts );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)v3_get_file_size( full_path );
	mem_free( full_path );
	
	if( need_to_free ) mem_free( ts );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = 0;
    }
}

//Remove a file:
void fn_remove_file( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name;
    pix_vm_container* name_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 1 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name, 0, PIX_CID );
	if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name ] == 0 ) { err = 1; break; }
	name_cont = vm->c[ name ];
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free = 0;
	UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
	
	UTF8_CHAR* full_path = pix_compose_full_path( vm->base_path, ts );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)v3_remove( full_path );
	mem_free( full_path );
	
	if( need_to_free ) mem_free( ts );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

//Rename a file:
void fn_rename_file( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name1;
    PIX_CID name2;
    pix_vm_container* name1_cont;
    pix_vm_container* name2_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 2 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name1, 0, PIX_CID );
	GET_VAL_FROM_STACK( name2, 1, PIX_CID );
	if( (unsigned)name1 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name1 ] == 0 ) { err = 1; break; }
	name1_cont = vm->c[ name1 ];
	if( (unsigned)name2 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name2 ] == 0 ) { err = 1; break; }
	name2_cont = vm->c[ name2 ];
	if( name1_cont->size == 0 ) { err = 1; break; }
	if( name2_cont->size == 0 ) { err = 1; break; }
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free1 = 0;
	bool need_to_free2 = 0;
	UTF8_CHAR* ts1 = pix_vm_make_cstring_from_container( name1, &need_to_free1, vm );
	UTF8_CHAR* ts2 = pix_vm_make_cstring_from_container( name2, &need_to_free2, vm );
	
	UTF8_CHAR* full_path1 = pix_compose_full_path( vm->base_path, ts1 );
	UTF8_CHAR* full_path2 = pix_compose_full_path( vm->base_path, ts2 );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)v3_rename( full_path1, full_path2 );
	mem_free( full_path1 );
	mem_free( full_path2 );
	
	if( need_to_free1 ) mem_free( ts1 );
	if( need_to_free2 ) mem_free( ts2 );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

//Copy a file:
void fn_copy_file( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name1;
    PIX_CID name2;
    pix_vm_container* name1_cont;
    pix_vm_container* name2_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 2 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name1, 0, PIX_CID );
	GET_VAL_FROM_STACK( name2, 1, PIX_CID );
	if( (unsigned)name1 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name1 ] == 0 ) { err = 1; break; }
	name1_cont = vm->c[ name1 ];
	if( (unsigned)name2 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name2 ] == 0 ) { err = 1; break; }
	name2_cont = vm->c[ name2 ];
	if( name1_cont->size == 0 ) { err = 1; break; }
	if( name2_cont->size == 0 ) { err = 1; break; }
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free1 = 0;
	bool need_to_free2 = 0;
	UTF8_CHAR* ts1 = pix_vm_make_cstring_from_container( name1, &need_to_free1, vm );
	UTF8_CHAR* ts2 = pix_vm_make_cstring_from_container( name2, &need_to_free2, vm );
	
	UTF8_CHAR* full_path1 = pix_compose_full_path( vm->base_path, ts1 );
	UTF8_CHAR* full_path2 = pix_compose_full_path( vm->base_path, ts2 );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_copy_file( (const UTF8_CHAR*)full_path2, (const UTF8_CHAR*)full_path1 );
	mem_free( full_path1 );
	mem_free( full_path2 );
	
	if( need_to_free1 ) mem_free( ts1 );
	if( need_to_free2 ) mem_free( ts2 );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

void fn_create_directory( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name;
    uint mode = 0;
#ifdef UNIX
    mode = S_IRWXU | S_IRWXG | S_IRWXO;
#endif
    pix_vm_container* name_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 1 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name, 0, PIX_CID );
        if( pars_num > 1 ) GET_VAL_FROM_STACK( mode, 1, uint );
	if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name ] == 0 ) { err = 1; break; }
	name_cont = vm->c[ name ];
	if( name_cont->size == 0 ) { err = 1; break; }
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free = 0;
	UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
	
	UTF8_CHAR* full_path = pix_compose_full_path( vm->base_path, ts );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_mkdir( (const UTF8_CHAR*)full_path, mode );
	mem_free( full_path );
	
	if( need_to_free ) mem_free( ts );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

//
// Graphics
//
	
void fn_frame( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
		
    if( vm->ready == 0 ) { time_sleep( 50 ); return; }
    int delay = 0;
    if( pars_num >= 1 ) GET_VAL_FROM_STACK( delay, 0, int );
    PIX_CID scr = vm->screen;
    if( (unsigned)scr < (unsigned)PIX_VM_CONTAINERS && vm->c[ scr ] && vm->screen_ptr != 0 )
    {
	pix_vm_container* c = vm->c[ scr ];
	vm->screen_change_x = 0;
	vm->screen_change_y = 0;
	vm->screen_change_xsize = c->xsize;
	vm->screen_change_ysize = c->ysize;
	if( pars_num >= 2 ) GET_VAL_FROM_STACK( vm->screen_change_x, 1, int );
	if( pars_num >= 3 ) GET_VAL_FROM_STACK( vm->screen_change_y, 2, int );
	if( pars_num >= 4 ) GET_VAL_FROM_STACK( vm->screen_change_xsize, 3, int );
	if( pars_num >= 5 ) GET_VAL_FROM_STACK( vm->screen_change_ysize, 4, int );
	vm->last_displayed_screen = vm->screen;
	vm->screen_redraw_request++;
	if( time_ticks() - vm->fps_time > time_ticks_per_second() )
	{
	    vm->fps = vm->fps_counter;
	    vm->vars[ PIX_GVAR_FPS ].i = vm->fps;
	    vm->var_types[ PIX_GVAR_FPS ] = 0;
	    vm->fps_time = time_ticks();
	    vm->fps_counter = 0;
	}
	vm->fps_counter++;
	if( delay ) time_sleep( delay );
	ticks_t t = time_ticks();
	ticks_t t_timeout = time_ticks_per_second() / 4;
	while( vm->screen_redraw_request != vm->screen_redraw_answer )
	{
	    time_sleep( 8 );
	    if( time_ticks() - t >= t_timeout ) break;
	}
    }
    else 
    {
	//No screen.
	if( delay == 0 ) delay = 50; 
	time_sleep( delay );
    }
}

void fn_vsync( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num < 1 ) return;
    bool vsync;
    GET_VAL_FROM_STACK( vsync, 0, bool );
    vm->vsync_request = vsync + 1;
    while( vm->vsync_request != 0 ) { time_sleep( 50 ); }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = 0;
}

void fn_set_screen( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_CID scr;
	GET_VAL_FROM_STACK( scr, 0, PIX_CID );
	pix_vm_gfx_set_screen( scr, vm );
    }
}

void fn_get_screen( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = vm->screen;
}

void fn_set_zbuf( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
		
    if( pars_num >= 1 )
    {
	PIX_CID z;
	GET_VAL_FROM_STACK( z, 0, PIX_CID );
#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    if( vm->zbuf != PIX_GL_ZBUF && z == PIX_GL_ZBUF )
	    {
		glDepthFunc( GL_LESS );
		glEnable( GL_DEPTH_TEST );
	    }
	    if( vm->zbuf == PIX_GL_ZBUF && z == -1 )
	    {
		glDisable( GL_DEPTH_TEST );
	    }
	}
#endif
	vm->zbuf = z;
    }
}

void fn_get_zbuf( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = vm->zbuf;
}

void fn_clear_zbuf( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    PIX_VAL v;
    v.i = 0x80000000;
    pix_vm_clean_container( vm->zbuf, 0, v, 0, -1, vm );

#ifdef OPENGL
    if( vm->screen == PIX_GL_SCREEN )
    {
	if( vm->zbuf == PIX_GL_ZBUF )
	{
	    glClear( GL_DEPTH_BUFFER_BIT );
	}
    }
#endif
}

void fn_get_color( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
	    
    if( pars_num >= 3 )
    {
	int r, g, b;
	GET_VAL_FROM_STACK( r, 0, int );
	GET_VAL_FROM_STACK( g, 1, int );
	GET_VAL_FROM_STACK( b, 2, int );
	if( r < 0 ) r = 0;
	if( g < 0 ) g = 0;
	if( b < 0 ) b = 0;
	if( r > 255 ) r = 255;
	if( g > 255 ) g = 255;
	if( b > 255 ) b = 255;
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (COLORSIGNED)get_color( r, g, b );
    }
}

void fn_get_red( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_INT c;
	GET_VAL_FROM_STACK( c, 0, PIX_INT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)red( c );
    }
}

void fn_get_green( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_INT c;
	GET_VAL_FROM_STACK( c, 0, PIX_INT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)green( c );
    }
}

void fn_get_blue( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_INT c;
	GET_VAL_FROM_STACK( c, 0, PIX_INT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)blue( c );
    }
}

void fn_get_blend( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 3 )
    {
	PIX_INT c1, c2, v;
	GET_VAL_FROM_STACK( c1, 0, PIX_INT );
	GET_VAL_FROM_STACK( c2, 1, PIX_INT );
	GET_VAL_FROM_STACK( v, 2, PIX_INT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (COLORSIGNED)blend( c1, c2, (int)v );
    }
}

void fn_transp( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_INT t;
	GET_VAL_FROM_STACK( t, 0, PIX_INT );
	if( t < 0 ) t = 0;
	if( t > 255 ) t = 255;
	vm->transp = (uchar)t;
    }
}

void fn_get_transp( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = vm->transp;
}

void fn_clear( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int t = vm->transp;
    if( t == 0 ) return;

#ifdef OPENGL
    if( vm->screen == PIX_GL_SCREEN )
    {
	COLOR c = CLEARCOLOR;
	if( pars_num >= 1 )
	{
	    GET_VAL_FROM_STACK( c, 0, COLOR );
	}
	if( t >= 255 ) 
	{
	    glClearColor( (float)red( c ) / 255, (float)green( c ) / 255, (float)blue( c ) / 255, 1 );
	    glClear( GL_COLOR_BUFFER_BIT );
	}
	else
	{
	    glPushMatrix();
	    glLoadIdentity();
	    glColor4ub( red( c ), green( c ), blue( c ), (uchar)t );
	    int16 v[ 8 ];
	    v[ 0 ] = 0; v[ 1 ] = 0;
	    v[ 2 ] = vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i; v[ 3 ] = 0;
	    v[ 4 ] = 0; v[ 5 ] = vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i;
	    v[ 6 ] = vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i; v[ 7 ] = vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i;
	    glEnableClientState( GL_VERTEX_ARRAY );
	    glVertexPointer( 2, GL_SHORT, 0, v );
	    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
	    glPopMatrix();
	}
	return;
    }
#endif
    
    if( vm->screen_ptr == 0 ) return;
    COLOR c = CLEARCOLOR;
    if( pars_num >= 1 )
    {
	GET_VAL_FROM_STACK( c, 0, COLOR );
    }
    COLORPTR p = vm->screen_ptr;
    COLORPTR p_end = vm->screen_ptr + vm->screen_xsize * vm->screen_ysize;
    if( t == 255 )
    {
	for( ; p < p_end; p++ )
	{
	    *p = c;
	}
    }
    else 
    {
	for( ; p < p_end; p++ )
	{
	    *p = fast_blend( *p, c, t );
	}
    }
}

void fn_dot( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( vm->screen_ptr == 0 ) return;
    if( vm->transp == 0 ) return;
		
    COLOR c = COLORMASK;
    PIX_INT x, y, z;

    if( fn_num == FN_DOT )
    {
	if( pars_num < 2 ) return;
	if( pars_num >= 3 )
	{
	    GET_VAL_FROM_STACK( c, 2, COLOR );
	}
    }
    else 
    {
	if( pars_num < 3 ) return;
	if( pars_num >= 4 )
	{
	    GET_VAL_FROM_STACK( c, 3, COLOR );
	}
    }

#ifdef OPENGL
    if( vm->screen == PIX_GL_SCREEN )
    {
	glColor4ub( red( c ), green( c ), blue( c ), vm->transp );
	float v[ 3 ];
	if( fn_num == FN_DOT )
	{
	    GET_VAL_FROM_STACK( v[ 0 ], 0, float );
	    GET_VAL_FROM_STACK( v[ 1 ], 1, float );
	    v[ 2 ] = 0;
	}
	else
	{
	    GET_VAL_FROM_STACK( v[ 0 ], 0, float );
	    GET_VAL_FROM_STACK( v[ 1 ], 1, float );
	    GET_VAL_FROM_STACK( v[ 2 ], 2, float );
	}
	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer( 3, GL_FLOAT, 0, v );
	glDrawArrays( GL_POINTS, 0, 1 );
	return;
    }
#endif
		    
    if( vm->t_enabled )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	PIX_FLOAT v[ 3 ];
	GET_VAL_FROM_STACK( v[ 0 ], 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( v[ 1 ], 1, PIX_FLOAT );
	if( fn_num == FN_DOT )
	    v[ 2 ] = 0;
	else 
	    GET_VAL_FROM_STACK( v[ 2 ], 2, PIX_FLOAT );
	pix_vm_gfx_vertex_transform( v, m );
	x = v[ 0 ];
	y = v[ 1 ];
	z = v[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
    }
    else 
    {
	GET_VAL_FROM_STACK( x, 0, PIX_INT );
	GET_VAL_FROM_STACK( y, 1, PIX_INT );
	if( fn_num == FN_GET_DOT )
	    z = 0;
	else 
	{
	    PIX_FLOAT zf;
	    GET_VAL_FROM_STACK( zf, 1, PIX_FLOAT );
	    z = zf * ( 1 << PIX_FIXED_MATH_PREC );
	}
    }
    x += vm->screen_xsize / 2;
    y += vm->screen_ysize / 2;
    if( (unsigned)x < (unsigned)vm->screen_xsize &&
	(unsigned)y < (unsigned)vm->screen_ysize )
    {
	int* zbuf = pix_vm_gfx_get_zbuf( vm );
	if( zbuf )
	{
	    int p = y * vm->screen_xsize + x;
	    if( z > zbuf[ p ] ) 
	    {
		if( vm->transp == 255 )
		{
		    vm->screen_ptr[ p ] = c;
		}
		else
		{
		    vm->screen_ptr[ p ] = fast_blend( vm->screen_ptr[ p ], c, vm->transp );
		}
		zbuf[ p ] = z;
	    }
	}
	else 
	{
	    COLORPTR p = vm->screen_ptr + y * vm->screen_xsize + x;
	    if( vm->transp == 255 )
	    {
		*p = c;
	    }
	    else
	    {
		*p = fast_blend( *p, c, vm->transp );
	    }
	}
    }
}

void fn_get_dot( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( vm->screen_ptr == 0 ) return;
    
    PIX_INT x, y;
    
    if( fn_num == FN_GET_DOT )
    {
	if( pars_num < 2 ) return;
    }
    else 
    {
	if( pars_num < 3 ) return;
    }
    
    if( vm->t_enabled )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	PIX_FLOAT v[ 3 ];
	GET_VAL_FROM_STACK( v[ 0 ], 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( v[ 1 ], 1, PIX_FLOAT );
	if( fn_num == FN_GET_DOT )
	    v[ 2 ] = 0;
	else 
	    GET_VAL_FROM_STACK( v[ 2 ], 2, PIX_FLOAT );
	pix_vm_gfx_vertex_transform( v, m );
	x = v[ 0 ];
	y = v[ 1 ];
    }
    else 
    {
	GET_VAL_FROM_STACK( x, 0, PIX_INT );
	GET_VAL_FROM_STACK( y, 1, PIX_INT );
    }
    x += vm->screen_xsize / 2;
    y += vm->screen_ysize / 2;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    if( (unsigned)x < (unsigned)vm->screen_xsize &&
	(unsigned)y < (unsigned)vm->screen_ysize )
    {
	COLORPTR p = vm->screen_ptr + y * vm->screen_xsize + x;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)*p;
    }
    else 
    {
	stack[ sp + ( pars_num - 1 ) ].i = 0;
    }
}

void fn_line( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( vm->screen_ptr == 0 ) return;
    if( vm->transp == 0 ) return;

    COLOR c = COLORMASK;
    PIX_INT x1, y1, z1, x2, y2, z2;
    
    if( fn_num == FN_LINE )
    {
	if( pars_num < 4 ) return;
	if( pars_num >= 5 )
	{
	    GET_VAL_FROM_STACK( c, 4, COLOR );
	}	
    }
    else
    {
	if( pars_num < 6 ) return;
	if( pars_num >= 7 )
	{
	    GET_VAL_FROM_STACK( c, 6, COLOR );
	}
    }

#ifdef OPENGL
    if( vm->screen == PIX_GL_SCREEN )
    {
	float v[ 6 ];
	if( fn_num == FN_LINE )
	{
	    GET_VAL_FROM_STACK( v[ 0 ], 0, float );
	    GET_VAL_FROM_STACK( v[ 1 ], 1, float );
	    v[ 2 ] = 0;
	    GET_VAL_FROM_STACK( v[ 3 ], 2, float );
	    GET_VAL_FROM_STACK( v[ 4 ], 3, float );
	    v[ 5 ] = 0;
	}
	else
	{
	    GET_VAL_FROM_STACK( v[ 0 ], 0, float );
	    GET_VAL_FROM_STACK( v[ 1 ], 1, float );
	    GET_VAL_FROM_STACK( v[ 2 ], 2, float );
	    GET_VAL_FROM_STACK( v[ 3 ], 3, float );
	    GET_VAL_FROM_STACK( v[ 4 ], 4, float );
	    GET_VAL_FROM_STACK( v[ 5 ], 5, float );
	}
	glColor4ub( red( c ), green( c ), blue( c ), vm->transp );
	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer( 3, GL_FLOAT, 0, v );
	glDrawArrays( GL_LINES, 0, 2 );
	return;
    }
#endif
    
    int* zbuf = pix_vm_gfx_get_zbuf( vm );

    if( fn_num == FN_LINE )
    {
	if( vm->t_enabled )
	{
	    PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	    PIX_FLOAT v[ 3 ];
	    
	    GET_VAL_FROM_STACK( v[ 0 ], 0, PIX_FLOAT );
	    GET_VAL_FROM_STACK( v[ 1 ], 1, PIX_FLOAT );
	    v[ 2 ] = 0;
	    pix_vm_gfx_vertex_transform( v, m );
	    x1 = v[ 0 ];
	    y1 = v[ 1 ];
	    if( zbuf ) z1 = v[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );

	    GET_VAL_FROM_STACK( v[ 0 ], 2, PIX_FLOAT );
	    GET_VAL_FROM_STACK( v[ 1 ], 3, PIX_FLOAT );
	    v[ 2 ] = 0;
	    pix_vm_gfx_vertex_transform( v, m );
	    x2 = v[ 0 ];
	    y2 = v[ 1 ];
	    if( zbuf ) z2 = v[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	}
	else 
	{
	    GET_VAL_FROM_STACK( x1, 0, PIX_INT );
	    GET_VAL_FROM_STACK( y1, 1, PIX_INT );
	    GET_VAL_FROM_STACK( x2, 2, PIX_INT );
	    GET_VAL_FROM_STACK( y2, 3, PIX_INT );
	    if( zbuf )
	    {
		z1 = 0;
		z2 = 0;
	    }
	}
    }
    else 
    {
	if( vm->t_enabled )
	{
	    PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	    PIX_FLOAT v[ 3 ];
	    
	    GET_VAL_FROM_STACK( v[ 0 ], 0, PIX_FLOAT );
	    GET_VAL_FROM_STACK( v[ 1 ], 1, PIX_FLOAT );
	    GET_VAL_FROM_STACK( v[ 2 ], 2, PIX_FLOAT );
	    pix_vm_gfx_vertex_transform( v, m );
	    x1 = v[ 0 ];
	    y1 = v[ 1 ];
	    if( zbuf ) z1 = v[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    
	    GET_VAL_FROM_STACK( v[ 0 ], 3, PIX_FLOAT );
	    GET_VAL_FROM_STACK( v[ 1 ], 4, PIX_FLOAT );
	    GET_VAL_FROM_STACK( v[ 2 ], 5, PIX_FLOAT );
	    pix_vm_gfx_vertex_transform( v, m );
	    x2 = v[ 0 ];
	    y2 = v[ 1 ];
	    if( zbuf ) z2 = v[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	}
	else 
	{
	    PIX_FLOAT zz;
	    GET_VAL_FROM_STACK( x1, 0, PIX_INT );
	    GET_VAL_FROM_STACK( y1, 1, PIX_INT );
	    GET_VAL_FROM_STACK( x2, 3, PIX_INT );
	    GET_VAL_FROM_STACK( y2, 4, PIX_INT );
	    if( zbuf ) 
	    {
		GET_VAL_FROM_STACK( zz, 2, PIX_FLOAT );
		z1 = zz * ( 1 << PIX_FIXED_MATH_PREC );
		GET_VAL_FROM_STACK( zz, 5, PIX_FLOAT );
		z2 = zz * ( 1 << PIX_FIXED_MATH_PREC );
	    }
	}
    }
    
    int screen_hxsize = vm->screen_xsize / 2;
    int screen_hysize = vm->screen_ysize / 2;
    
    x1 += screen_hxsize;
    y1 += screen_hysize;
    x2 += screen_hxsize;
    y2 += screen_hysize;
    if( zbuf )
    {
	pix_vm_gfx_draw_line_zbuf( x1, y1, z1, x2, y2, z2, c, zbuf, vm );
    }
    else 
    {
	pix_vm_gfx_draw_line( x1, y1, x2, y2, c, vm );
    }
}

void fn_box( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( vm->screen_ptr == 0 ) return;
    if( vm->transp == 0 ) return;
		
    COLOR c = COLORMASK;
    if( pars_num < 4 ) return;
    if( pars_num >= 5 )
    {
	GET_VAL_FROM_STACK( c, 4, COLOR );
    }

    int screen_hxsize = vm->screen_xsize / 2;
    int screen_hysize = vm->screen_ysize / 2;
    
    if( vm->t_enabled || vm->zbuf >= 0 || vm->screen == PIX_GL_SCREEN )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	PIX_FLOAT x, y, xsize, ysize;
	
	GET_VAL_FROM_STACK( x, 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( y, 1, PIX_FLOAT );
	GET_VAL_FROM_STACK( xsize, 2, PIX_FLOAT );
	GET_VAL_FROM_STACK( ysize, 3, PIX_FLOAT );
	
	PIX_FLOAT v[ 12 ];
	v[ 0 ] = x;
	v[ 1 ] = y;
	v[ 2 ] = 0;
	v[ 3 ] = x + xsize;
	v[ 4 ] = y;
	v[ 5 ] = 0;
	v[ 6 ] = x;
	v[ 7 ] = y + ysize;
	v[ 8 ] = 0;
	v[ 9 ] = x + xsize;
	v[ 10 ] = y + ysize;
	v[ 11 ] = 0;
#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    glColor4ub( red( c ), green( c ), blue( c ), vm->transp );
	    glEnableClientState( GL_VERTEX_ARRAY );
	    float* vv;
	    float vvv[ 15 ];
	    if( fn_num == FN_BOX )
	    {
		vvv[ 0 ] = v[ 0 ];
		vvv[ 1 ] = v[ 1 ];
		vvv[ 2 ] = v[ 2 ];
		vvv[ 3 ] = v[ 3 ];
		vvv[ 4 ] = v[ 4 ];
		vvv[ 5 ] = v[ 5 ];
		vvv[ 6 ] = v[ 9 ];
		vvv[ 7 ] = v[ 10 ];
		vvv[ 8 ] = v[ 11 ];
		vvv[ 9 ] = v[ 6 ];
		vvv[ 10 ] = v[ 7 ];
		vvv[ 11 ] = v[ 8 ];
		vvv[ 12 ] = v[ 0 ];
		vvv[ 13 ] = v[ 1 ];
		vvv[ 14 ] = v[ 2 ];
		glVertexPointer( 3, GL_FLOAT, 0, vvv );
		glDrawArrays( GL_LINE_STRIP, 0, 5 );
	    }
	    else
	    {
		if( sizeof( float ) == sizeof( PIX_FLOAT ) )
	    	    vv = (float*)v;
		else
		{
		    vvv[ 0 ] = v[ 0 ];
		    vvv[ 1 ] = v[ 1 ];
		    vvv[ 2 ] = v[ 2 ];
		    vvv[ 3 ] = v[ 3 ];
		    vvv[ 4 ] = v[ 4 ];
		    vvv[ 5 ] = v[ 5 ];
		    vvv[ 6 ] = v[ 6 ];
		    vvv[ 7 ] = v[ 7 ];
		    vvv[ 8 ] = v[ 8 ];
		    vvv[ 9 ] = v[ 9 ];
		    vvv[ 10 ] = v[ 10 ];
		    vvv[ 11 ] = v[ 11 ];
		    vv = vvv;
		}
		glVertexPointer( 3, GL_FLOAT, 0, vv );
		glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
	    }
	    return;
	}
#endif
	if( vm->t_enabled )
	{
	    pix_vm_gfx_vertex_transform( v, m );
	    pix_vm_gfx_vertex_transform( v + 3, m );
	    pix_vm_gfx_vertex_transform( v + 6, m );
	    pix_vm_gfx_vertex_transform( v + 9, m );
	}
	v[ 0 ] += screen_hxsize;
	v[ 1 ] += screen_hysize;
	v[ 3 ] += screen_hxsize;
	v[ 4 ] += screen_hysize;
	v[ 6 ] += screen_hxsize;
	v[ 7 ] += screen_hysize;
	v[ 9 ] += screen_hxsize;
	v[ 10 ] += screen_hysize;
	if( fn_num == FN_BOX )
	{
	    pix_vm_gfx_draw_line( v[ 0 ], v[ 1 ], v[ 3 ], v[ 4 ], c, vm );
	    pix_vm_gfx_draw_line( v[ 3 ], v[ 4 ], v[ 9 ], v[ 10 ], c, vm );
	    pix_vm_gfx_draw_line( v[ 9 ], v[ 10 ], v[ 6 ], v[ 7 ], c, vm );
	    pix_vm_gfx_draw_line( v[ 6 ], v[ 7 ], v[ 0 ], v[ 1 ], c, vm );
	}
	else 
	{
	    pix_vm_ivertex iv1;
	    pix_vm_ivertex iv2;
	    pix_vm_ivertex iv3;
	    iv1.x = v[ 0 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv1.y = v[ 1 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv1.z = v[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv2.x = v[ 3 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv2.y = v[ 4 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv2.z = v[ 5 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv3.x = v[ 6 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv3.y = v[ 7 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv3.z = v[ 8 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    if( vm->zbuf >= 0 )
		pix_vm_gfx_draw_triangle_zbuf( &iv1, &iv2, &iv3, c, vm );
	    else 
		pix_vm_gfx_draw_triangle( &iv1, &iv2, &iv3, c, vm );
	    iv1.x = v[ 9 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv1.y = v[ 10 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv1.z = v[ 11 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    if( vm->zbuf >= 0 )
		pix_vm_gfx_draw_triangle_zbuf( &iv3, &iv2, &iv1, c, vm );
	    else 
		pix_vm_gfx_draw_triangle( &iv3, &iv2, &iv1, c, vm );
	}
    }
    else 
    {
	PIX_INT x, y, xsize, ysize;
	
	GET_VAL_FROM_STACK( x, 0, PIX_INT );
	GET_VAL_FROM_STACK( y, 1, PIX_INT );
	GET_VAL_FROM_STACK( xsize, 2, PIX_INT );
	GET_VAL_FROM_STACK( ysize, 3, PIX_INT );
	
	x += screen_hxsize;
	y += screen_hysize;
	
	if( fn_num == FN_BOX )
	    pix_vm_gfx_draw_box( x, y, xsize, ysize, c, vm );
	else 
	    pix_vm_gfx_draw_fbox( x, y, xsize, ysize, c, vm );
    }
}

void fn_pixi( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( vm->screen_ptr == 0 ) return;

    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    PIX_FLOAT x, y, xscale, yscale;
    PIX_INT tx, ty, txsize, tysize;
    COLOR c = get_color( 255, 255, 255 );
    x = 0;
    y = 0;
    xscale = 1;
    yscale = 1;
    tx = 0;
    ty = 0;
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );

    if( (unsigned)cnum >= (unsigned)PIX_VM_CONTAINERS ) return;
    pix_vm_container* cont = vm->c[ cnum ];
    if( cont == 0 ) return;
    
    if( cont->opt_data && cont->opt_data->hdata )
    {
	//Try to auto-play animation:
	int f = pix_vm_container_hdata_autoplay_control( cnum, vm );
	if( f >= 0 ) pix_vm_container_hdata_unpack_frame( cnum, vm );
    }

    txsize = cont->xsize;
    tysize = cont->ysize;

    while( 1 )
    {
	if( pars_num >= 2 ) GET_VAL_FROM_STACK( x, 1, PIX_FLOAT ) else break;
        if( pars_num >= 3 ) GET_VAL_FROM_STACK( y, 2, PIX_FLOAT ) else break;
        if( pars_num >= 4 ) GET_VAL_FROM_STACK( c, 3, COLOR ) else break;
	if( pars_num >= 5 ) GET_VAL_FROM_STACK( xscale, 4, PIX_FLOAT ) else break;
        if( pars_num >= 6 ) GET_VAL_FROM_STACK( yscale, 5, PIX_FLOAT ) else break;
	if( xscale == 0 ) return;
	if( yscale == 0 ) return;
        if( pars_num >= 7 ) GET_VAL_FROM_STACK( tx, 6, PIX_INT ) else break;
        if( pars_num >= 8 ) GET_VAL_FROM_STACK( ty, 7, PIX_INT ) else break;
        if( pars_num >= 9 ) GET_VAL_FROM_STACK( txsize, 8, PIX_INT ) else break;
        if( pars_num >= 10 ) GET_VAL_FROM_STACK( tysize, 9, PIX_INT ) else break;
        break;
    }
    
    PIX_FLOAT xsize = (PIX_FLOAT)txsize * xscale;
    PIX_FLOAT ysize = (PIX_FLOAT)tysize * yscale;
        
    pix_vm_gfx_draw_container( cnum, x - xsize / 2, y - ysize / 2, 0, xsize, ysize, tx, ty, txsize, tysize, c, vm );
}

void fn_triangles3d( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num < 2 ) return;
    
    PIX_CID vertices;
    PIX_CID triangles;
    PIX_CID vert_colors = -1;
    
    GET_VAL_FROM_STACK( vertices, 0, PIX_CID );
    GET_VAL_FROM_STACK( triangles, 1, PIX_CID );
    if( pars_num >= 3 )
    {
	GET_VAL_FROM_STACK( vert_colors, 2, PIX_CID );
    }

    pix_vm_container* v_cont = pix_vm_get_container( vertices, vm );
    pix_vm_container* t_cont = pix_vm_get_container( triangles, vm );
    pix_vm_container* c_cont = 0;
    if( v_cont == 0 ) return;
    if( t_cont == 0 ) return;
    if( v_cont->data == 0 ) return;
    if( t_cont->data == 0 ) return;
    if( v_cont->size < 8 ) return;
    if( t_cont->size < 8 ) return;
    if( vert_colors != -1 )
    {
	c_cont = pix_vm_get_container( vert_colors, vm );
	if( c_cont == 0 ) return;
	if( c_cont->data == 0 ) return;
	if( c_cont->size < ( v_cont->size / 8 ) * 2 )
	{
	    dprint( "triangles3d: vert_colors container size must be greater or equal to ( number of vertices * 2 )\n" );
	    return;
	}
    }

    int screen_hxsize = vm->screen_xsize / 2;
    int screen_hysize = vm->screen_ysize / 2;

    size_t vnum = v_cont->size / 8;
    size_t tnum = t_cont->size / 8;
    for( size_t t = 0; t < tnum; t++ )
    {
	size_t ts = pix_vm_get_container_int_element( triangles, t * 8 + 7, vm );
	PIX_INT v1num = pix_vm_get_container_int_element( triangles, ts * 8 + 0, vm );
	PIX_INT v2num = pix_vm_get_container_int_element( triangles, ts * 8 + 1, vm );
	PIX_INT v3num = pix_vm_get_container_int_element( triangles, ts * 8 + 2, vm );
	COLOR color = (COLOR)pix_vm_get_container_int_element( triangles, ts * 8 + 3, vm );
	PIX_INT texture = pix_vm_get_container_int_element( triangles, ts * 8 + 4, vm );
	PIX_INT transp = pix_vm_get_container_int_element( triangles, ts * 8 + 5, vm );
	if( transp <= 0 ) continue;
	if( transp >= 255 ) transp = 255;
	uchar prev_transp = vm->transp;
	uchar new_transp = ( (int)transp * (int)vm->transp ) / 256;
	if( new_transp == 0 ) continue;
	vm->transp = new_transp;
	if( (unsigned)v1num >= vnum ) continue;
	if( (unsigned)v2num >= vnum ) continue;
	if( (unsigned)v3num >= vnum ) continue;
	PIX_FLOAT v1[ 5 ];
	PIX_FLOAT v2[ 5 ];
	PIX_FLOAT v3[ 5 ];
	v1[ 0 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 0, vm );
	v1[ 1 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 1, vm );
	v1[ 2 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 2, vm );
	v2[ 0 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 0, vm );
	v2[ 1 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 1, vm );
	v2[ 2 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 2, vm );
	v3[ 0 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 0, vm );
	v3[ 1 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 1, vm );
	v3[ 2 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 2, vm );
	if( texture >= 0 )
	{
	    v1[ 3 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 3, vm );
	    v1[ 4 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 4, vm );
	    v2[ 3 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 3, vm );
	    v2[ 4 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 4, vm );
	    v3[ 3 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 3, vm );
	    v3[ 4 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 4, vm );
	}
	COLOR v_colors[ 3 ];
	uchar v_transp[ 3 ];
	if( c_cont )
	{
	    v_colors[ 0 ] = pix_vm_get_container_int_element( vert_colors, v1num * 2 + 0, vm );
	    v_colors[ 1 ] = pix_vm_get_container_int_element( vert_colors, v2num * 2 + 0, vm );
	    v_colors[ 2 ] = pix_vm_get_container_int_element( vert_colors, v3num * 2 + 0, vm );
	    v_transp[ 0 ] = pix_vm_get_container_int_element( vert_colors, v1num * 2 + 1, vm );
	    v_transp[ 1 ] = pix_vm_get_container_int_element( vert_colors, v2num * 2 + 1, vm );
	    v_transp[ 2 ] = pix_vm_get_container_int_element( vert_colors, v3num * 2 + 1, vm );
	}
#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    float v[ 3 * 3 ];
	    v[ 0 ] = v1[ 0 ]; v[ 1 ] = v1[ 1 ]; v[ 2 ] = v1[ 2 ];
	    v[ 3 ] = v2[ 0 ]; v[ 4 ] = v2[ 1 ]; v[ 5 ] = v2[ 2 ];
	    v[ 6 ] = v3[ 0 ]; v[ 7 ] = v3[ 1 ]; v[ 8 ] = v3[ 2 ];
	    glColor4ub( red( color ), green( color ), blue( color ), vm->transp );
	    glEnableClientState( GL_VERTEX_ARRAY );
	    glVertexPointer( 3, GL_FLOAT, 0, v );
	    if( c_cont )
	    {
		glEnableClientState( GL_COLOR_ARRAY );
		uchar cc[ 3 * 4 ];
		cc[ 0 ] = red( v_colors[ 0 ] ); cc[ 1 ] = green( v_colors[ 0 ] ); cc[ 2 ] = blue( v_colors[ 0 ] ); cc[ 3 ] = v_transp[ 0 ];
		cc[ 4 ] = red( v_colors[ 1 ] ); cc[ 5 ] = green( v_colors[ 1 ] ); cc[ 6 ] = blue( v_colors[ 1 ] ); cc[ 7 ] = v_transp[ 1 ];
		cc[ 8 ] = red( v_colors[ 2 ] ); cc[ 9 ] = green( v_colors[ 2 ] ); cc[ 10 ] = blue( v_colors[ 2 ] ); cc[ 11 ] = v_transp[ 2 ];
	        glColorPointer( 4, GL_UNSIGNED_BYTE, 0, cc );
	    }
	    if( texture == -1 )
	    {
		//No texture:
		glDrawArrays( GL_TRIANGLES, 0, 3 );
	    }
	    else
	    {
		//With texture:
		pix_vm_create_container_gl_data( texture, vm );
    		pix_vm_container_gl_data* gl = pix_vm_get_container_gl_data( texture, vm );
    		if( gl == 0 ) goto triangle_draw_end;
		glEnable( GL_TEXTURE_2D );
    		glEnableClientState( GL_TEXTURE_COORD_ARRAY );
    		float t[ 2 * 3 ];
    		t[ 0 ] = v1[ 3 ] / gl->xsize; t[ 1 ] = v1[ 4 ] / gl->ysize;
    		t[ 2 ] = v2[ 3 ] / gl->xsize; t[ 3 ] = v2[ 4 ] / gl->ysize;
    		t[ 4 ] = v3[ 3 ] / gl->xsize; t[ 5 ] = v3[ 4 ] / gl->ysize;
    		glTexCoordPointer( 2, GL_FLOAT, 0, t );
    		glBindTexture( GL_TEXTURE_2D, gl->texture_id );
		glDrawArrays( GL_TRIANGLES, 0, 3 );
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    		glDisable( GL_TEXTURE_2D );
	    }
	    if( c_cont )
	    {
		glDisableClientState( GL_COLOR_ARRAY );
	    }
	    goto triangle_draw_end;
	}
#endif
	if( vm->t_enabled )
	{
	    pix_vm_gfx_vertex_transform( v1, vm->t_matrix + ( vm->t_matrix_sp * 16 ) );
	    pix_vm_gfx_vertex_transform( v2, vm->t_matrix + ( vm->t_matrix_sp * 16 ) );
	    pix_vm_gfx_vertex_transform( v3, vm->t_matrix + ( vm->t_matrix_sp * 16 ) );
	}
	v1[ 0 ] += screen_hxsize;
	v1[ 1 ] += screen_hysize;
	v2[ 0 ] += screen_hxsize;
	v2[ 1 ] += screen_hysize;
	v3[ 0 ] += screen_hxsize;
	v3[ 1 ] += screen_hysize;
	if( texture == -1 )
	{
	    //No texture:
	    pix_vm_ivertex iv1;
	    pix_vm_ivertex iv2;
	    pix_vm_ivertex iv3;
	    iv1.x = v1[ 0 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv1.y = v1[ 1 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv1.z = v1[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv2.x = v2[ 0 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv2.y = v2[ 1 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv2.z = v2[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv3.x = v3[ 0 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv3.y = v3[ 1 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    iv3.z = v3[ 2 ] * ( 1 << PIX_FIXED_MATH_PREC );
	    if( vm->zbuf >= 0 )
		pix_vm_gfx_draw_triangle_zbuf( &iv1, &iv2, &iv3, color, vm );
	    else 
		pix_vm_gfx_draw_triangle( &iv1, &iv2, &iv3, color, vm );
	}
	else
	{
	    pix_vm_gfx_draw_triangle_t( v1, v2, v3, texture, color, vm );
	}
triangle_draw_end:
	vm->transp = prev_transp;
    }
}

void fn_sort_triangles3d( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num < 2 ) return;
    
    PIX_CID vertices;
    PIX_CID triangles;
    
    GET_VAL_FROM_STACK( vertices, 0, PIX_CID );
    GET_VAL_FROM_STACK( triangles, 1, PIX_CID );

    pix_vm_container* v_cont = pix_vm_get_container( vertices, vm );
    pix_vm_container* t_cont = pix_vm_get_container( triangles, vm );
    if( v_cont == 0 ) return;
    if( t_cont == 0 ) return;
    if( v_cont->data == 0 ) return;
    if( t_cont->data == 0 ) return;
    if( v_cont->size < 8 ) return;
    if( t_cont->size < 8 ) return;
    
    size_t vnum = v_cont->size / 8;
    size_t tnum = t_cont->size / 8;
    PIX_FLOAT* zz = (PIX_FLOAT*)MEM_NEW2( tnum * sizeof( PIX_FLOAT ) );
    int* order = (int*)MEM_NEW2( tnum * sizeof( int ) );
    if( zz == 0 ) return;
    if( order == 0 ) return;
    for( size_t t = 0; t < tnum; t++ )
    {
	order[ t ] = (int)t;
	zz[ t ] = 0;
	PIX_INT v1num = pix_vm_get_container_int_element( triangles, t * 8 + 0, vm );
	PIX_INT v2num = pix_vm_get_container_int_element( triangles, t * 8 + 1, vm );
	PIX_INT v3num = pix_vm_get_container_int_element( triangles, t * 8 + 2, vm );
	if( (unsigned)v1num >= vnum ) continue;
	if( (unsigned)v2num >= vnum ) continue;
	if( (unsigned)v3num >= vnum ) continue;
	PIX_FLOAT v1[ 3 ];
	PIX_FLOAT v2[ 3 ];
	PIX_FLOAT v3[ 3 ];
	v1[ 0 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 0, vm );
	v1[ 1 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 1, vm );
	v1[ 2 ] = pix_vm_get_container_float_element( vertices, v1num * 8 + 2, vm );
	v2[ 0 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 0, vm );
	v2[ 1 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 1, vm );
	v2[ 2 ] = pix_vm_get_container_float_element( vertices, v2num * 8 + 2, vm );
	v3[ 0 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 0, vm );
	v3[ 1 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 1, vm );
	v3[ 2 ] = pix_vm_get_container_float_element( vertices, v3num * 8 + 2, vm );
	if( vm->t_enabled )
	{
	    pix_vm_gfx_vertex_transform( v1, vm->t_matrix + ( vm->t_matrix_sp * 16 ) );
	    pix_vm_gfx_vertex_transform( v2, vm->t_matrix + ( vm->t_matrix_sp * 16 ) );
	    pix_vm_gfx_vertex_transform( v3, vm->t_matrix + ( vm->t_matrix_sp * 16 ) );
	}
	zz[ t ] = ( v1[ 2 ] + v2[ 2 ] + v3[ 2 ] ) / 3;
    }
    while( 1 )
    {
	bool sort = 0;
	for( size_t t = 0; t < tnum - 1; t++ )
	{
	    if( zz[ t ] > zz[ t + 1 ] )
	    {
		PIX_FLOAT tz = zz[ t ];
		zz[ t ] = zz[ t + 1 ];
		zz[ t + 1 ] = tz;
		int to = order[ t ];
		order[ t ] = order[ t + 1 ];
		order[ t + 1 ] = to;
		sort = 1;
	    }
	}
	if( sort == 0 ) break;
    }
    for( size_t t = 0; t < tnum; t++ )
    {
	pix_vm_set_container_int_element( triangles, t * 8 + 7, order[ t ], vm );
    }
    mem_free( zz );
    mem_free( order );
}

void fn_set_key_color( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    bool key_color = 0;
    PIX_INT key;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    if( pars_num > 1 )
    {
	key_color = 1;
	GET_VAL_FROM_STACK( key, 1, PIX_INT );
    }
    
    if( key_color )
    {
	pix_vm_set_container_key_color( cnum, key, vm );
    }
    else 
    {
	pix_vm_remove_container_key_color( cnum, vm );
    }
}

void fn_get_key_color( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_get_container_key_color( cnum, vm );
}

void fn_set_alpha( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    PIX_CID alpha_cnum = -1;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    if( pars_num > 1 )
    {
	GET_VAL_FROM_STACK( alpha_cnum, 1, PIX_CID );
    }
    
    pix_vm_set_container_alpha( cnum, alpha_cnum, vm );
}

void fn_get_alpha( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_get_container_alpha( cnum, vm );
}

void fn_print( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    PIX_FLOAT x = 0, y = 0;
    COLOR c = get_color( 255, 255, 255 );
    int align = 0;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    if( pars_num > 1 ) GET_VAL_FROM_STACK( x, 1, PIX_FLOAT );
    if( pars_num > 2 ) GET_VAL_FROM_STACK( y, 2, PIX_FLOAT );
    if( pars_num > 3 ) GET_VAL_FROM_STACK( c, 3, COLOR );
    if( pars_num > 4 ) GET_VAL_FROM_STACK( align, 4, int );
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* cont = vm->c[ cnum ];
	if( cont && cont->data && cont->size )
	{
	    size_t real_size = cont->size * g_pix_container_type_sizes[ cont->type ];
	    pix_vm_gfx_draw_text( (UTF8_CHAR*)cont->data, real_size, x, y, 0, align, c, 0, 0, 0, vm );
	}
    }
}

void fn_get_text_xsize( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    int align = 0;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    if( pars_num > 1 ) GET_VAL_FROM_STACK( align, 1, int );
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* cont = vm->c[ cnum ];
	if( cont && cont->data && cont->size )
	{
	    size_t real_size = cont->size * g_pix_container_type_sizes[ cont->type ];
	    int xsize;
	    pix_vm_gfx_draw_text( (UTF8_CHAR*)cont->data, real_size, 0, 0, 0, align, 1, &xsize, 0, 1, vm );
	    stack[ sp + ( pars_num - 1 ) ].i = xsize;
	}
    }
}

void fn_get_text_ysize( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    PIX_CID cnum;
    int align = 0;
    
    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
    if( pars_num > 1 ) GET_VAL_FROM_STACK( align, 1, int );
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* cont = vm->c[ cnum ];
	if( cont && cont->data && cont->size )
	{
	    size_t real_size = cont->size * g_pix_container_type_sizes[ cont->type ];
	    int ysize;
	    pix_vm_gfx_draw_text( (UTF8_CHAR*)cont->data, real_size, 0, 0, 0, align, 1, 0, &ysize, 1, vm );
	    stack[ sp + ( pars_num - 1 ) ].i = ysize;
	}
    }
}

void fn_set_font( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num < 2 ) return;
    
    UTF32_CHAR first_char;
    PIX_CID cnum;
    int xchars = 0;
    int ychars = 0;
    
    GET_VAL_FROM_STACK( first_char, 0, UTF32_CHAR );
    GET_VAL_FROM_STACK( cnum, 1, PIX_CID );
    if( pars_num >= 3 )
	GET_VAL_FROM_STACK( xchars, 2, int );
    if( pars_num >= 4 )
	GET_VAL_FROM_STACK( ychars, 3, int );
	
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)pix_vm_set_font( first_char, cnum, xchars, ychars, vm );
}

void fn_get_font( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 1 ) return;
    
    UTF32_CHAR char_code;
    GET_VAL_FROM_STACK( char_code, 0, UTF32_CHAR );
    
    pix_vm_font* font = pix_vm_get_font_for_char( char_code, vm );

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    if( font )
	stack[ sp + ( pars_num - 1 ) ].i = font->font;
    else
	stack[ sp + ( pars_num - 1 ) ].i = -1;
}

void fn_effector( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 2 ) return;
    
    int transp = vm->transp;
    if( transp <= 0 ) return;
    
    int type;
    int power;
    COLOR color = get_color( 255, 255, 255 );
    int sx = -vm->screen_xsize / 2;
    int sy = -vm->screen_ysize / 2;
    int xsize = vm->screen_xsize;
    int ysize = vm->screen_ysize;
    int x_step = 1;
    int y_step = 1;
    
    GET_VAL_FROM_STACK( type, 0, int );
    GET_VAL_FROM_STACK( power, 1, int );
    if( pars_num > 2 ) { GET_VAL_FROM_STACK( color, 2, COLOR ); }
    if( pars_num > 3 ) { GET_VAL_FROM_STACK( sx, 3, int ); }
    if( pars_num > 4 ) { GET_VAL_FROM_STACK( sy, 4, int ); }
    if( pars_num > 5 ) { GET_VAL_FROM_STACK( xsize, 5, int ); }
    if( pars_num > 6 ) { GET_VAL_FROM_STACK( ysize, 6, int ); }
    if( pars_num > 7 ) { GET_VAL_FROM_STACK( x_step, 7, int ); }
    if( pars_num > 8 ) { GET_VAL_FROM_STACK( y_step, 8, int ); }

    sx += vm->screen_xsize / 2;
    sy += vm->screen_ysize / 2;
    
    if( sx + xsize < 0 ) return;
    if( sy + ysize < 0 ) return;
    if( sx < 0 ) { xsize -= -sx; sx = 0; }
    if( sy < 0 ) { ysize -= -sy; sy = 0; }
    if( sx + xsize > vm->screen_xsize ) xsize = vm->screen_xsize - sx;
    if( sy + ysize > vm->screen_ysize ) ysize = vm->screen_ysize - sy;
    if( xsize <= 0 ) return;
    if( ysize <= 0 ) return;
    if( x_step <= 0 ) x_step = 1;
    if( y_step <= 0 ) y_step = 1;

    int* cols_r = 0;
    int* cols_g = 0;
    int* cols_b = 0;
    switch( type )
    {
	case PIX_EFFECT_HBLUR:
	case PIX_EFFECT_VBLUR:
	    if( vm->effector_colors_r == 0 ) vm->effector_colors_r = (int*)MEM_NEW2( 512 * sizeof( int ) );
	    if( vm->effector_colors_g == 0 ) vm->effector_colors_g = (int*)MEM_NEW2( 512 * sizeof( int ) );
	    if( vm->effector_colors_b == 0 ) vm->effector_colors_b = (int*)MEM_NEW2( 512 * sizeof( int ) );
	    cols_r = vm->effector_colors_r;
	    cols_g = vm->effector_colors_g;
	    cols_b = vm->effector_colors_b;
	    if( cols_r == 0 ) return;
	    if( cols_g == 0 ) return;
	    if( cols_b == 0 ) return;
	    break;
    }
    
    switch( type )
    {
	case PIX_EFFECT_NOISE:
	    {
		transp *= power;
		transp /= 256;
		if( transp <= 0 ) break;
		for( int cy = 0; cy < ysize; cy += y_step )
		{
		    COLORPTR scr = vm->screen_ptr + ( sy + cy ) * vm->screen_xsize + sx;
		    for( int cx = 0; cx < xsize; cx += x_step )
		    {
			*scr = blend( *scr, color, ( transp * ( pseudo_random() & 255 ) ) / 256 );
			scr += x_step;
		    }
		}
	    }
	    break;
	case PIX_EFFECT_SPREAD_LEFT:
	    {
		power /= 2;
		if( power <= 0 ) break;
		power++;
		for( int cy = 0; cy < ysize; cy += y_step )
		{
		    COLORPTR scr = vm->screen_ptr + ( sy + cy ) * vm->screen_xsize;
		    for( int cx = sx; cx < sx + xsize; cx += x_step )
		    {
		        int cx2;
		        int rnd = ( ( pseudo_random() & 4095 ) * power ) / 4096;
		        cx2 = cx + rnd;
		        if( cx2 >= vm->screen_xsize ) cx2 = vm->screen_xsize - 1;
		        scr[ cx ] = blend( scr[ cx ], scr[ cx2 ], transp );
		    }
		}
	    }
	    break;
	case PIX_EFFECT_SPREAD_RIGHT:
	    {
		power /= 2;
		if( power <= 0 ) break;
		power++;
		for( int cy = 0; cy < ysize; cy += y_step )
		{
		    COLORPTR scr = vm->screen_ptr + ( sy + cy ) * vm->screen_xsize;
		    for( int cx = sx + ( ( xsize - 1 ) / x_step ) * x_step; cx >= sx; cx -= x_step )
		    {
		        int cx2;
		        int rnd = ( ( pseudo_random() & 4095 ) * power ) / 4096;
		        cx2 = cx - rnd;
		        if( cx2 < 0 ) cx2 = 0;
		        scr[ cx ] = blend( scr[ cx ], scr[ cx2 ], transp );
		    }
		}
	    }
	    break;
	case PIX_EFFECT_SPREAD_UP:
	    {
		power /= 2;
		if( power <= 0 ) break;
		power++;
		for( int cx = 0; cx < xsize; cx += x_step )
		{
		    COLORPTR scr = vm->screen_ptr + sx + cx;
		    for( int cy = sy; cy < sy + ysize; cy += y_step )
		    {
		        int cy2;
		        int rnd = ( ( pseudo_random() & 4095 ) * power ) / 4096;
		        cy2 = cy + rnd;
		        if( cy2 >= vm->screen_ysize ) cy2 = vm->screen_ysize - 1;
		        COLORPTR s = &scr[ cy * vm->screen_xsize ];
		        *s = blend( *s, scr[ cy2 * vm->screen_xsize ], transp );
		    }
		}
	    }
	    break;
	case PIX_EFFECT_SPREAD_DOWN:
	    {
		power /= 2;
		if( power <= 0 ) break;
		power++;
		for( int cx = 0; cx < xsize; cx += x_step )
		{
		    COLORPTR scr = vm->screen_ptr + sx + cx;
		    for( int cy = sy + ( ( ysize - 1 ) / y_step ) * y_step; cy >= sy; cy -= y_step )
		    {
		        int cy2;
		        int rnd = ( ( pseudo_random() & 4095 ) * power ) / 4096;
		        cy2 = cy - rnd;
		        if( cy2 < 0 ) cy2 = 0;
		        COLORPTR s = &scr[ cy * vm->screen_xsize ];
		        *s = blend( *s, scr[ cy2 * vm->screen_xsize ], transp );
		    }
		}
	    }
	    break;
	case PIX_EFFECT_HBLUR:
	    {
		power /= 2;
		if( power > 255 ) power = 255;
		if( power <= 0 ) break;
		for( int cy = 0; cy < ysize; cy += y_step )
		{
		    COLORPTR scr = vm->screen_ptr + ( sy + cy ) * vm->screen_xsize + sx;
		    int col_ptr = 0;
		    COLOR pcol = *scr;
		    int start_r = red( pcol );
		    int start_g = green( pcol );
		    int start_b = blue( pcol );
		    cols_r[ 0 ] = 0;
		    cols_g[ 0 ] = 0;
		    cols_b[ 0 ] = 0;
		    for( int i = 1; i < power + 1; i++ ) 
		    { 
			cols_r[ i ] = start_r * i; 
			cols_g[ i ] = start_g * i; 
			cols_b[ i ] = start_b * i; 
		    }
		    col_ptr = power + 1;
		    for( int i = 0; ( i < power ) && ( i < xsize ); i++ ) 
		    {
			pcol = scr[ i ];
			cols_r[ col_ptr ] = cols_r[ col_ptr - 1 ] + red( pcol );
			cols_g[ col_ptr ] = cols_g[ col_ptr - 1 ] + green( pcol );
			cols_b[ col_ptr ] = cols_b[ col_ptr - 1 ] + blue( pcol );
			col_ptr++;
		    }
		    if( power > xsize )
		    {
			pcol = scr[ xsize - 1 ];
			int rr = red( pcol );
			int gg = green( pcol );
			int bb = blue( pcol );
			for( int i = 0; i < power - xsize; i++ ) 
			{
			    cols_r[ col_ptr ] = cols_r[ col_ptr - 1 ] + rr;
			    cols_g[ col_ptr ] = cols_g[ col_ptr - 1 ] + gg;
			    cols_b[ col_ptr ] = cols_b[ col_ptr - 1 ] + bb;
			    col_ptr++;
			}
		    }
		    for( int cx = 0; cx < xsize; cx += x_step )
		    {
			int prev_ptr = ( col_ptr - 1 ) & 511;
			int new_ptr = cx + power; if( new_ptr > xsize - 1 ) new_ptr = xsize - 1;
			pcol = scr[ new_ptr ];
			cols_r[ col_ptr ] = cols_r[ prev_ptr ] + red( pcol );
			cols_g[ col_ptr ] = cols_g[ prev_ptr ] + green( pcol );
			cols_b[ col_ptr ] = cols_b[ prev_ptr ] + blue( pcol );
			COLOR res = get_color(
					      ( cols_r[ col_ptr ] - cols_r[ (col_ptr-(power*2+1)) & 511 ] ) / ( power * 2 + 1 ),
					      ( cols_g[ col_ptr ] - cols_g[ (col_ptr-(power*2+1)) & 511 ] ) / ( power * 2 + 1 ),
					      ( cols_b[ col_ptr ] - cols_b[ (col_ptr-(power*2+1)) & 511 ] ) / ( power * 2 + 1 )
					      );
			scr[ cx ] = fast_blend( scr[ cx ], res, transp );
			col_ptr ++;
			col_ptr &= 511;
		    }
		}
	    }
	    break;
	case PIX_EFFECT_VBLUR:
	    {
		power /= 2;
		if( power > 255 ) power = 255;
		if( power <= 0 ) break;
		for( int cx = 0; cx < xsize; cx += x_step )
		{
		    COLORPTR scr = vm->screen_ptr + sy * vm->screen_xsize + sx + cx;
		    int col_ptr = 0;
		    COLOR pcol = *scr;
		    int start_r = red( pcol );
		    int start_g = green( pcol );
		    int start_b = blue( pcol );
		    cols_r[ 0 ] = 0;
		    cols_g[ 0 ] = 0;
		    cols_b[ 0 ] = 0;
		    int i;
		    for( i = 1; i < power + 1; i++ ) 
		    { 
			cols_r[ i ] = start_r * i; 
			cols_g[ i ] = start_g * i; 
			cols_b[ i ] = start_b * i; 
		    }
		    col_ptr = power + 1;
		    for( int i = 0, i2 = 0; ( i < power ) && ( i < ysize ); i++ ) 
		    {
			pcol = scr[ i2 ];
			cols_r[ col_ptr ] = cols_r[ col_ptr - 1 ] + red( pcol );
			cols_g[ col_ptr ] = cols_g[ col_ptr - 1 ] + green( pcol );
			cols_b[ col_ptr ] = cols_b[ col_ptr - 1 ] + blue( pcol );
			i2 += vm->screen_xsize;
			col_ptr++;
		    }
		    if( power > ysize )
		    {
			pcol = scr[ ( ysize - 1 ) * vm->screen_xsize ];
			int rr = red( pcol );
			int gg = green( pcol );
			int bb = blue( pcol );
			for( int i = 0; i < power - ysize; i++ )
			{ 
			    cols_r[ col_ptr ] = cols_r[ col_ptr - 1 ] + rr;
			    cols_g[ col_ptr ] = cols_g[ col_ptr - 1 ] + gg;
			    cols_b[ col_ptr ] = cols_b[ col_ptr - 1 ] + bb;
			    col_ptr++;
			}
		    }
		    int power2 = power * vm->screen_xsize;
		    int end_ptr = ( ysize - 1 ) * vm->screen_xsize;
		    for( int cy = 0; cy < ysize * vm->screen_xsize; cy += y_step * vm->screen_xsize )
		    {
			int prev_ptr = ( col_ptr - 1 ) & 511;
			int new_ptr = cy + power2; if( new_ptr > end_ptr ) new_ptr = end_ptr;
			pcol = scr[ new_ptr ];
			cols_r[ col_ptr ] = cols_r[ prev_ptr ] + red( pcol );
			cols_g[ col_ptr ] = cols_g[ prev_ptr ] + green( pcol );
			cols_b[ col_ptr ] = cols_b[ prev_ptr ] + blue( pcol );
			COLOR res = get_color(
					      ( cols_r[ col_ptr ] - cols_r[ (col_ptr-(power*2+1)) & 511 ] ) / ( power * 2 + 1 ),
					      ( cols_g[ col_ptr ] - cols_g[ (col_ptr-(power*2+1)) & 511 ] ) / ( power * 2 + 1 ),
					      ( cols_b[ col_ptr ] - cols_b[ (col_ptr-(power*2+1)) & 511 ] ) / ( power * 2 + 1 )
					      );
			scr[ cy ] = fast_blend( scr[ cy ], res, transp );
			col_ptr ++;
			col_ptr &= 511;
		    }
		}
	    }
	    break;
	case PIX_EFFECT_COLOR:
	    {
		transp *= power;
		transp /= 256;
		if( transp <= 0 ) break;
		for( int cy = 0; cy < ysize; cy += y_step )
		{
		    COLORPTR scr = vm->screen_ptr + ( sy + cy ) * vm->screen_xsize + sx;
		    for( int cx = 0; cx < xsize; cx += x_step )
		    {
			*scr = blend( *scr, color, transp );
			scr += x_step;
		    }
		}
	    }
    }
}

void fn_split_rgb( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num < 3 ) return;
    
    int rv = -1;

    int direction; //0 - from image to RGB; 1 - from RGB to image;   
    PIX_CID image;
    PIX_CID channels[ 3 ] = { -1, -1, -1 };
    size_t image_offset = 0;
    size_t channel_offset = 0;
    size_t size = -1;
    
    GET_VAL_FROM_STACK( direction, 0, int );
    GET_VAL_FROM_STACK( image, 1, PIX_CID );
    GET_VAL_FROM_STACK( channels[ 0 ], 2, PIX_CID );
    while( 1 )
    {
	if( pars_num > 3 ) { GET_VAL_FROM_STACK( channels[ 1 ], 3, PIX_CID ); } else break;
	if( pars_num > 4 ) { GET_VAL_FROM_STACK( channels[ 2 ], 4, PIX_CID ); } else break;
	if( pars_num > 5 ) { GET_VAL_FROM_STACK( image_offset, 5, size_t ); } else break;
	if( pars_num > 6 ) { GET_VAL_FROM_STACK( channel_offset, 6, size_t ); } else break;
	if( pars_num > 7 ) { GET_VAL_FROM_STACK( size, 7, size_t ); } else break;
	break;
    }
    
    while( 1 )
    {
	pix_vm_container* image_cont = pix_vm_get_container( image, vm );
	if( image_cont == 0 ) break;
	if( g_pix_container_type_sizes[ image_cont->type ] != COLORLEN )
	{
	    dprint( "split_rgb: image container must be of PIXEL type\n" );
	    break;
	}
	pix_vm_container* channels_cont[ 3 ];
	int channel_type = -1;
	for( int i = 0; i < 3; i++ )
	{
	    channels_cont[ i ] = pix_vm_get_container( channels[ i ], vm );
	    if( channels_cont[ i ] )
	    {
		if( channel_type == -1 ) 
		    channel_type = channels_cont[ i ]->type;
		else
		{
		    if( channels_cont[ i ]->type != channel_type )
		    {
			dprint( "split_rgb: all channels must have the same type\n" );
			channel_type = -1;
			break;
		    }
		}
	    }
	}
	if( channel_type == -1 )
	{
	    dprint( "split_rgb: can't split the image (no channels selected)\n" );
	    break;
	}
	
	if( size == -1 ) size = image_cont->size;
	if( image_offset >= image_cont->size ) break;
	if( image_offset + size > image_cont->size )
	{
	    size = image_cont->size - image_offset;
	}
	
	COLORPTR image_ptr = (COLORPTR)image_cont->data + image_offset;
	void* ch[ 3 ];
	for( int i = 0; i < 3; i++ )
	{
	    if( channels_cont[ i ] && channels_cont[ i ]->data ) 
	    {
		if( channel_offset >= channels_cont[ i ]->size ) size = 0;
		if( channel_offset + size > channels_cont[ i ]->size )
		{
		    size = channels_cont[ i ]->size - channel_offset;
		}
		ch[ i ] = (void*)( (uchar*)channels_cont[ i ]->data + channel_offset * g_pix_container_type_sizes[ channel_type ] );
	    }
	    else ch[ i ] = 0;
	}
	if( direction == 0 )
	{
	    //From image to RGB:
	    switch( channel_type )
	    {
		case PIX_CONTAINER_TYPE_INT8:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    COLOR pixel = image_ptr[ p ];
			    if( ch[ 0 ] ) ((uchar*)ch[ 0 ])[ p ] = red( pixel );
			    if( ch[ 1 ] ) ((uchar*)ch[ 1 ])[ p ] = green( pixel );
			    if( ch[ 2 ] ) ((uchar*)ch[ 2 ])[ p ] = blue( pixel );
			}
			rv = 0;
		    }
		    break;
		case PIX_CONTAINER_TYPE_INT16:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    COLOR pixel = image_ptr[ p ];
			    if( ch[ 0 ] ) ((int16*)ch[ 0 ])[ p ] = red( pixel );
			    if( ch[ 1 ] ) ((int16*)ch[ 1 ])[ p ] = green( pixel );
			    if( ch[ 2 ] ) ((int16*)ch[ 2 ])[ p ] = blue( pixel );
			}
			rv = 0;
		    }
		    break;
		case PIX_CONTAINER_TYPE_INT32:
		    {
			for( size_t p = 0; p < size; p++ )
                        {
                            COLOR pixel = image_ptr[ p ];
			    if( ch[ 0 ] ) ((int*)ch[ 0 ])[ p ] = red( pixel );
			    if( ch[ 1 ] ) ((int*)ch[ 1 ])[ p ] = green( pixel );
			    if( ch[ 2 ] ) ((int*)ch[ 2 ])[ p ] = blue( pixel );
			}
			rv = 0;
		    }
		    break;
#ifdef PIX_INT64_ENABLED
		case PIX_CONTAINER_TYPE_INT64:
		    {
			for( size_t p = 0; p < size; p++ )
                        {
                            COLOR pixel = image_ptr[ p ];
			    if( ch[ 0 ] ) ((int64*)ch[ 0 ])[ p ] = red( pixel );
			    if( ch[ 1 ] ) ((int64*)ch[ 1 ])[ p ] = green( pixel );
			    if( ch[ 2 ] ) ((int64*)ch[ 2 ])[ p ] = blue( pixel );
			}
			rv = 0;
		    }
		    break;
#endif
		case PIX_CONTAINER_TYPE_FLOAT32:
		    {
			for( size_t p = 0; p < size; p++ )
                        {
                            COLOR pixel = image_ptr[ p ];
			    if( ch[ 0 ] ) ((float*)ch[ 0 ])[ p ] = (float)red( pixel ) / 255.0F;
			    if( ch[ 1 ] ) ((float*)ch[ 1 ])[ p ] = (float)green( pixel ) / 255.0F;
			    if( ch[ 2 ] ) ((float*)ch[ 2 ])[ p ] = (float)blue( pixel ) / 255.0F;
			}
			rv = 0;
		    }
		    break;
#ifdef PIX_FLOAT32_ENABLED
		case PIX_CONTAINER_TYPE_FLOAT64:
		    {
			for( size_t p = 0; p < size; p++ )
                        {
                            COLOR pixel = image_ptr[ p ];
			    if( ch[ 0 ] ) ((double*)ch[ 0 ])[ p ] = (float)red( pixel ) / 255.0F;
			    if( ch[ 1 ] ) ((double*)ch[ 1 ])[ p ] = (float)green( pixel ) / 255.0F;
			    if( ch[ 2 ] ) ((double*)ch[ 2 ])[ p ] = (float)blue( pixel ) / 255.0F;
			}
			rv = 0;
		    }
		    break;
#endif
	    }
	}
	else
	{
	    //From RGB to image:
	    switch( channel_type )
	    {
		case PIX_CONTAINER_TYPE_INT8:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    int r, g, b;
			    if( ch[ 0 ] ) r = ((uchar*)ch[ 0 ])[ p ]; else r = 0;
			    if( ch[ 1 ] ) g = ((uchar*)ch[ 1 ])[ p ]; else g = 0;
			    if( ch[ 2 ] ) b = ((uchar*)ch[ 2 ])[ p ]; else b = 0;
			    image_ptr[ p ] = get_color( r, g, b );
			}
			rv = 0;
		    }
		    break;
		case PIX_CONTAINER_TYPE_INT16:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    int r, g, b;
			    if( ch[ 0 ] ) r = ((int16*)ch[ 0 ])[ p ]; else r = 0;
			    if( ch[ 1 ] ) g = ((int16*)ch[ 1 ])[ p ]; else g = 0;
			    if( ch[ 2 ] ) b = ((int16*)ch[ 2 ])[ p ]; else b = 0;
			    if( r < 0 ) r = 0; else if( r > 255 ) r = 255;
			    if( g < 0 ) g = 0; else if( g > 255 ) g = 255;
			    if( b < 0 ) b = 0; else if( b > 255 ) b = 255;
			    image_ptr[ p ] = get_color( r, g, b );
			}
			rv = 0;
		    }
		    break;
		case PIX_CONTAINER_TYPE_INT32:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    int r, g, b;
			    if( ch[ 0 ] ) r = ((int*)ch[ 0 ])[ p ]; else r = 0;
			    if( ch[ 1 ] ) g = ((int*)ch[ 1 ])[ p ]; else g = 0;
			    if( ch[ 2 ] ) b = ((int*)ch[ 2 ])[ p ]; else b = 0;
			    if( r < 0 ) r = 0; else if( r > 255 ) r = 255;
			    if( g < 0 ) g = 0; else if( g > 255 ) g = 255;
			    if( b < 0 ) b = 0; else if( b > 255 ) b = 255;
			    image_ptr[ p ] = get_color( r, g, b );
			}
			rv = 0;
		    }
		    break;
#ifdef PIX_INT64_ENABLED
		case PIX_CONTAINER_TYPE_INT64:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    int64 r, g, b;
			    if( ch[ 0 ] ) r = ((int64*)ch[ 0 ])[ p ]; else r = 0;
			    if( ch[ 1 ] ) g = ((int64*)ch[ 1 ])[ p ]; else g = 0;
			    if( ch[ 2 ] ) b = ((int64*)ch[ 2 ])[ p ]; else b = 0;
			    if( r < 0 ) r = 0; else if( r > 255 ) r = 255;
			    if( g < 0 ) g = 0; else if( g > 255 ) g = 255;
			    if( b < 0 ) b = 0; else if( b > 255 ) b = 255;
			    image_ptr[ p ] = get_color( r, g, b );
			}
			rv = 0;
		    }
		    break;
#endif
		case PIX_CONTAINER_TYPE_FLOAT32:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    int r, g, b;
			    if( ch[ 0 ] ) r = ((float*)ch[ 0 ])[ p ] * 255; else r = 0;
			    if( ch[ 1 ] ) g = ((float*)ch[ 1 ])[ p ] * 255; else g = 0;
			    if( ch[ 2 ] ) b = ((float*)ch[ 2 ])[ p ] * 255; else b = 0;
			    if( r < 0 ) r = 0; else if( r > 255 ) r = 255;
			    if( g < 0 ) g = 0; else if( g > 255 ) g = 255;
			    if( b < 0 ) b = 0; else if( b > 255 ) b = 255;
			    image_ptr[ p ] = get_color( r, g, b );
			}
			rv = 0;
		    }
		    break;
#ifdef PIX_FLOAT64_ENABLED
		case PIX_CONTAINER_TYPE_FLOAT64:
		    {
			for( size_t p = 0; p < size; p++ )
			{
			    int r, g, b;
			    if( ch[ 0 ] ) r = ((double*)ch[ 0 ])[ p ] * 255; else r = 0;
			    if( ch[ 1 ] ) g = ((double*)ch[ 1 ])[ p ] * 255; else g = 0;
			    if( ch[ 2 ] ) b = ((double*)ch[ 2 ])[ p ] * 255; else b = 0;
			    if( r < 0 ) r = 0; else if( r > 255 ) r = 255;
			    if( g < 0 ) g = 0; else if( g > 255 ) g = 255;
			    if( b < 0 ) b = 0; else if( b > 255 ) b = 255;
			    image_ptr[ p ] = get_color( r, g, b );
			}
			rv = 0;
		    }
		    break;
#endif
	    }
	}
	
	break;
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//
// OpenGL graphics
//

void fn_set_gl_callback( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
#ifdef OPENGL
    if( pars_num > 0 )
    {
	gl_lock();
	
	PIX_ADDR callback;
	PIX_VAL userdata;
	userdata.i = 0;
	char userdata_type = 0;
	GET_VAL_FROM_STACK( callback, 0, PIX_ADDR );
	if( callback == -1 || IS_ADDRESS_CORRECT( callback ) )
	{
	    if( callback != -1 )
		callback &= PIX_INT_ADDRESS_MASK;
	    if( pars_num > 1 ) 
	    {
		userdata = stack[ sp + 1 ];
		userdata_type = stack_types[ sp + 1 ];
	    }
	    vm->gl_callback = callback;
	    vm->gl_userdata = userdata;
	    vm->gl_userdata_type = userdata_type;
	    stack[ sp + ( pars_num - 1 ) ].i = 0;
	    stack_types[ sp + ( pars_num - 1 ) ] = 0;
	}
	else
	{
	    stack[ sp + ( pars_num - 1 ) ].i = -1;
	    stack_types[ sp + ( pars_num - 1 ) ] = 0;
	    dprint( "set_gl_callback() error: wrong callback address %d\n", (int)callback );
	}
	
	gl_unlock();
    }
#else
    stack[ sp + ( pars_num - 1 ) ].i = -1;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    dprint( "set_gl_callback() error: no OpenGL. Please use the special Pixilang version with OpenGL support.\n" );
#endif
}

void fn_remove_gl_data( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
#ifdef OPENGL
    if( pars_num > 0 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_remove_container_gl_data( cnum, vm );
    }
#endif
}

//
// Animation
//

void fn_pixi_unpack_frame( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    if( pars_num >= 1 ) 
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	rv = pix_vm_container_hdata_unpack_frame( cnum, vm );
    }
    
    stack[ sp + ( pars_num - 1 ) ].i = rv;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
}

void fn_pixi_pack_frame( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    if( pars_num >= 1 ) 
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	rv = pix_vm_container_hdata_pack_frame( cnum, vm );
    }
    
    stack[ sp + ( pars_num - 1 ) ].i = rv;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
}

void fn_pixi_create_anim( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    while( 1 )
    {
	if( pars_num >= 1 ) 
	{
	    PIX_CID cnum;
	    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	    pix_vm_container* c = pix_vm_get_container( cnum, vm );
	    if( c == 0 ) break;
	    uchar* hdata_ptr = (uchar*)pix_vm_get_container_hdata( cnum, vm );
	    if( hdata_ptr )
	    {
		if( hdata_ptr[ 0 ] != pix_vm_container_hdata_type_anim )
		{
		    pix_vm_remove_container_hdata( cnum, vm );
		    hdata_ptr = 0;
		}
	    }
	    if( hdata_ptr == 0 )
	    {
		pix_vm_create_container_hdata( cnum, pix_vm_container_hdata_type_anim, sizeof( pix_vm_container_hdata_anim ), vm );
		hdata_ptr = (uchar*)pix_vm_get_container_hdata( cnum, vm );
		if( hdata_ptr )
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)hdata_ptr;
		    hdata->frame_count = 1;
		    PIX_VAL prop_val;
		    prop_val.i = 0; pix_vm_set_container_property( cnum, "frame", -1, 0, prop_val, vm );
		    prop_val.i = 10; pix_vm_set_container_property( cnum, "fps", -1, 0, prop_val, vm );
		    prop_val.i = 1; pix_vm_set_container_property( cnum, "frames", -1, 0, prop_val, vm );
		    prop_val.i = -1; pix_vm_set_container_property( cnum, "repeat", -1, 0, prop_val, vm );
		    hdata->frames = (pix_vm_anim_frame*)MEM_NEW2( sizeof( pix_vm_anim_frame ) );
		    mem_zero( hdata->frames );
		    if( hdata->frames )
		    {
			rv = pix_vm_container_hdata_pack_frame_from_buf( cnum, 0, (COLORPTR)c->data, c->type, c->xsize, c->ysize, vm );
		    }
		}
	    }
	}
	break;
    }
    
    stack[ sp + ( pars_num - 1 ) ].i = rv;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
}

void fn_pixi_remove_anim( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    while( 1 )
    {
	if( pars_num >= 1 ) 
	{
	    PIX_CID cnum;
	    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	    pix_vm_container* c = pix_vm_get_container( cnum, vm );
	    if( c == 0 ) break;
	    pix_vm_remove_container_hdata( cnum, vm );
	    rv = 0;
	}
	break;
    }
    
    stack[ sp + ( pars_num - 1 ) ].i = rv;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
}

void fn_pixi_clone_frame( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    while( 1 )
    {
	if( pars_num >= 1 ) 
	{
	    PIX_CID cnum;
	    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	    rv = pix_vm_container_hdata_clone_frame( cnum, vm );
	}
	break;
    }
    
    stack[ sp + ( pars_num - 1 ) ].i = rv;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
}

void fn_pixi_remove_frame( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    while( 1 )
    {
	if( pars_num >= 1 ) 
	{
	    PIX_CID cnum;
	    GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	    rv = pix_vm_container_hdata_remove_frame( cnum, vm );
	}
	break;
    }
    
    stack[ sp + ( pars_num - 1 ) ].i = rv;
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
}

void fn_pixi_play( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 ) 
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c && c->opt_data && c->opt_data->hdata )
	{
	    PIX_VAL prop_val;
	    prop_val.i = 1; pix_vm_set_container_property( cnum, "play", -1, 0, prop_val, vm );
	    prop_val.i = (uint)time_ticks_hires(); pix_vm_set_container_property( cnum, "start_time", -1, 0, prop_val, vm );
	    prop_val.i = pix_vm_get_container_property_i( cnum, "frame", -1, vm ); pix_vm_set_container_property( cnum, "start_frame", -1, 0, prop_val, vm );
	    //Unpack first frame:
	    pix_vm_container_hdata_unpack_frame( cnum, vm );
	}
    }
}

void fn_pixi_stop( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 ) 
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c && c->opt_data && c->opt_data->hdata )
	{
	    PIX_VAL prop_val;
	    prop_val.i = 0; pix_vm_set_container_property( cnum, "play", -1, 0, prop_val, vm );
	}
    }
}

//
// Transformation
//
	    
void fn_t_reset( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    pix_vm_gfx_matrix_reset( vm );
    
#ifdef OPENGL
    pix_vm_gfx_gl_matrix_reset( vm );
#endif
}

void fn_t_rotate( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 4 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	PIX_FLOAT angle, x, y, z;
	GET_VAL_FROM_STACK( angle, 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( x, 1, PIX_FLOAT );
	GET_VAL_FROM_STACK( y, 2, PIX_FLOAT );
	GET_VAL_FROM_STACK( z, 3, PIX_FLOAT );
	angle /= 180;
	angle *= M_PI;
	
	//Normalize vector:
	PIX_FLOAT inv_length = 1.0f / sqrt( x * x + y * y + z * z );
	x *= inv_length;
	y *= inv_length;
	z *= inv_length;
	
	//Create the matrix:
	PIX_FLOAT c = cos( angle );
	PIX_FLOAT s = sin( angle );
	PIX_FLOAT cc = 1 - c;
	PIX_FLOAT r[ 4 * 4 ];
	PIX_FLOAT res[ 4 * 4 ];
	r[ 0 + 0 ] = x * x * cc + c;
	r[ 4 + 0 ] = x * y * cc - z * s;
	r[ 8 + 0 ] = x * z * cc + y * s;
	r[ 12 + 0 ] = 0;
	r[ 0 + 1 ] = y * x * cc + z * s;
	r[ 4 + 1 ] = y * y * cc + c;
	r[ 8 + 1 ] = y * z * cc - x * s;
	r[ 12 + 1 ] = 0;
	r[ 0 + 2 ] = x * z * cc - y * s;
	r[ 4 + 2 ] = y * z * cc + x * s;
	r[ 8 + 2 ] = z * z * cc + c;
	r[ 12 + 2 ] = 0;
	r[ 0 + 3 ] = 0;
	r[ 4 + 3 ] = 0;
	r[ 8 + 3 ] = 0;
	r[ 12 + 3 ] = 1;
	pix_vm_gfx_matrix_mul( res, m, r );
	mem_copy( m, res, sizeof( PIX_FLOAT ) * 4 * 4 );

	vm->t_enabled = 1;

#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    pix_vm_gfx_gl_matrix_reset( vm );
	    pix_vm_gfx_gl_matrix_mul( vm );
	}
#endif
    }
}

void fn_t_translate( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 3 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	PIX_FLOAT x, y, z;
	GET_VAL_FROM_STACK( x, 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( y, 1, PIX_FLOAT );
	GET_VAL_FROM_STACK( z, 2, PIX_FLOAT );
	
	PIX_FLOAT m2[ 4 * 4 ];
	PIX_FLOAT res[ 4 * 4 ];
	mem_set( m2, sizeof( PIX_FLOAT ) * 4 * 4, 0 );
	m2[ 0 ] = 1;
	m2[ 4 + 1 ] = 1;
	m2[ 8 + 2 ] = 1;
	m2[ 12 + 0 ] = x;
	m2[ 12 + 1 ] = y;
	m2[ 12 + 2 ] = z;
	m2[ 12 + 3 ] = 1;

	pix_vm_gfx_matrix_mul( res, m, m2 );
	mem_copy( m, res, sizeof( PIX_FLOAT ) * 4 * 4 );
	
	vm->t_enabled = 1;

#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    pix_vm_gfx_gl_matrix_reset( vm );
	    pix_vm_gfx_gl_matrix_mul( vm );
	}
#endif
    }
}

void fn_t_scale( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 3 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	PIX_FLOAT x, y, z;
	GET_VAL_FROM_STACK( x, 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( y, 1, PIX_FLOAT );
	GET_VAL_FROM_STACK( z, 2, PIX_FLOAT );
	
	PIX_FLOAT m2[ 4 * 4 ];
	PIX_FLOAT res[ 4 * 4 ];
	mem_set( m2, sizeof( PIX_FLOAT ) * 4 * 4, 0 );
	m2[ 0 ] = x;
	m2[ 4 + 1 ] = y;
	m2[ 8 + 2 ] = z;
	m2[ 12 + 3 ] = 1;
	
	pix_vm_gfx_matrix_mul( res, m, m2 );
	mem_copy( m, res, sizeof( PIX_FLOAT ) * 4 * 4 );
	
	vm->t_enabled = 1;

#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    pix_vm_gfx_gl_matrix_reset( vm );
	    pix_vm_gfx_gl_matrix_mul( vm );
	}
#endif
    }
}

void fn_t_push_matrix( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( vm->t_matrix_sp >= PIX_T_MATRIX_STACK_SIZE - 1 )
    {
	dprint( "t_push_matrix(): stack overflow\n" );
    }
    else
    {
	mem_copy( vm->t_matrix + ( vm->t_matrix_sp + 1 ) * 16, vm->t_matrix + vm->t_matrix_sp * 16, 4 * 4 * sizeof( PIX_FLOAT ) );
	vm->t_matrix_sp++;
    }
}

void fn_t_pop_matrix( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( vm->t_matrix_sp == 0 )
    {
	dprint( "t_pop_matrix(): nothing to pop up froms stack\n" );
    }
    else
    {
	vm->t_matrix_sp--;
#ifdef OPENGL
	if( vm->screen == PIX_GL_SCREEN )
	{
	    pix_vm_gfx_gl_matrix_reset( vm );
	    pix_vm_gfx_gl_matrix_mul( vm );
	}
#endif
    }
}

void fn_t_get_matrix( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 1 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	
	PIX_CID m2;
	GET_VAL_FROM_STACK( m2, 0, PIX_CID );
	
	if( (unsigned)m2 < (unsigned)PIX_VM_CONTAINERS && vm->c[ m2 ] )
	{
	    pix_vm_container* c = vm->c[ m2 ];
	    if( sizeof( PIX_FLOAT ) == 32 && c->type != PIX_CONTAINER_TYPE_FLOAT32 ) return;
	    if( sizeof( PIX_FLOAT ) == 64 && c->type != PIX_CONTAINER_TYPE_FLOAT64 ) return;
	    if( c->size < 4 * 4 ) return;
	    mem_copy( c->data, m, sizeof( PIX_FLOAT ) * 4 * 4 );
	}
    }
}

void fn_t_set_matrix( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	
	PIX_CID m2;
	GET_VAL_FROM_STACK( m2, 0, PIX_CID );
	
	if( (unsigned)m2 < (unsigned)PIX_VM_CONTAINERS && vm->c[ m2 ] )
	{
	    pix_vm_container* c = vm->c[ m2 ];
	    if( sizeof( PIX_FLOAT ) == 32 && c->type != PIX_CONTAINER_TYPE_FLOAT32 ) return;
	    if( sizeof( PIX_FLOAT ) == 64 && c->type != PIX_CONTAINER_TYPE_FLOAT64 ) return;
	    if( c->size < 4 * 4 ) return;
	    mem_copy( m, c->data, sizeof( PIX_FLOAT ) * 4 * 4 );
	    
	    vm->t_enabled = 1;

#ifdef OPENGL
	    if( vm->screen == PIX_GL_SCREEN )
	    {
		pix_vm_gfx_gl_matrix_reset( vm );
		pix_vm_gfx_gl_matrix_mul( vm );
	    }
#endif
	}
    }
}

void fn_t_mul_matrix( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	
	PIX_CID m2;
	GET_VAL_FROM_STACK( m2, 0, PIX_CID );
	
	if( (unsigned)m2 < (unsigned)PIX_VM_CONTAINERS && vm->c[ m2 ] )
	{
	    pix_vm_container* c = vm->c[ m2 ];
	    if( sizeof( PIX_FLOAT ) == 32 && c->type != PIX_CONTAINER_TYPE_FLOAT32 ) return;
	    if( sizeof( PIX_FLOAT ) == 64 && c->type != PIX_CONTAINER_TYPE_FLOAT64 ) return;
	    if( c->size < 4 * 4 ) return;
	    PIX_FLOAT res_m[ 4 * 4 ];
	    pix_vm_gfx_matrix_mul( res_m, m, (PIX_FLOAT*)c->data );
	    mem_copy( m, res_m, sizeof( PIX_FLOAT ) * 4 * 4 );
	    
	    vm->t_enabled = 1;

#ifdef OPENGL
	    if( vm->screen == PIX_GL_SCREEN )
	    {
		pix_vm_gfx_gl_matrix_reset( vm );
		pix_vm_gfx_gl_matrix_mul( vm );
	    }
#endif
	}
    }
}

void fn_t_point( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT* m = vm->t_matrix + ( vm->t_matrix_sp * 16 );
	
	PIX_CID p;
	GET_VAL_FROM_STACK( p, 0, PIX_CID );
	
	if( (unsigned)p < (unsigned)PIX_VM_CONTAINERS && vm->c[ p ] )
	{
	    pix_vm_container* c = vm->c[ p ];
	    if( sizeof( PIX_FLOAT ) == 32 && c->type != PIX_CONTAINER_TYPE_FLOAT32 ) return;
	    if( sizeof( PIX_FLOAT ) == 64 && c->type != PIX_CONTAINER_TYPE_FLOAT64 ) return;
	    if( c->size < 3 ) return;
	    while( pars_num >= 2 )
	    {
		PIX_CID mc;
		GET_VAL_FROM_STACK( mc, 1, PIX_CID );
		if( (unsigned)mc < (unsigned)PIX_VM_CONTAINERS && vm->c[ mc ] )
		{
		    pix_vm_container* c2 = vm->c[ mc ];
		    if( sizeof( PIX_FLOAT ) == 32 && c2->type != PIX_CONTAINER_TYPE_FLOAT32 ) break;
		    if( sizeof( PIX_FLOAT ) == 64 && c2->type != PIX_CONTAINER_TYPE_FLOAT64 ) break;
		    if( c2->size < 4 * 4 ) break;
		    m = (PIX_FLOAT*)c2->data;
		}
		break;
	    }
	    pix_vm_gfx_vertex_transform( (PIX_FLOAT*)c->data, m );
	}
    }
}

//
// Audio
//	

void fn_set_audio_callback( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num > 0 )
    {
	PIX_ADDR callback;
	PIX_VAL userdata;
	userdata.i = 0;
	char userdata_type = 0;
	int freq = g_snd.freq;
	int format = PIX_CONTAINER_TYPE_INT16;
	int channels = 1;
	int flags = 0;
	GET_VAL_FROM_STACK( callback, 0, PIX_ADDR );
	if( callback == -1 || IS_ADDRESS_CORRECT( callback ) )
	{
	    if( callback != -1 )
		callback &= PIX_INT_ADDRESS_MASK;
	    if( pars_num > 1 ) 
	    {
		userdata = stack[ sp + 1 ];
		userdata_type = stack_types[ sp + 1 ];
	    }
	    if( pars_num > 2 ) { GET_VAL_FROM_STACK( freq, 2, int ); }
	    if( pars_num > 3 ) { GET_VAL_FROM_STACK( format, 3, int ); }
	    if( pars_num > 4 ) { GET_VAL_FROM_STACK( channels, 4, int ); }
	    if( pars_num > 5 ) { GET_VAL_FROM_STACK( flags, 5, int ); }
	    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_set_audio_callback( callback, userdata, userdata_type, freq, (pix_container_type)format, channels, flags, vm );
	    stack_types[ sp + ( pars_num - 1 ) ] = 0;
	}
	else
	{
	    stack[ sp + ( pars_num - 1 ) ].i = -1;
            stack_types[ sp + ( pars_num - 1 ) ] = 0;
            dprint( "set_audio_callback() error: wrong callback address %d\n", (int)callback );
	}
    }
}

void fn_enable_audio_input( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = 0;

    if( pars_num > 0 )
    {
	int enable = 0;
	GET_VAL_FROM_STACK( enable, 0, int );
	sound_stream_input( enable );
	if( enable ) vm->audio_input_enabled++; else vm->audio_input_enabled--;
    }
}

void fn_get_note_freq( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = 0;

    if( pars_num > 0 )
    {
	int note = 0;
	int fine = 0;
	GET_VAL_FROM_STACK( note, 0, int );
	if( pars_num > 1 ) { GET_VAL_FROM_STACK( fine, 1, int ); }
	int p = 7680 - note * 64 - fine;
	if( p >= 0 )
	    rv = ( g_linear_freq_tab[ p % 768 ] >> ( p / 768 ) );
	else
	    rv = ( g_linear_freq_tab[ (7680*4+p) % 768 ] << -( ( (7680*4+p) / 768 ) - (7680*4)/768 ) ); //if period is negative
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//
// MIDI
//

void fn_midi_open_client( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    PIX_CID rv = -1;
    
    if( pars_num > 0 )
    {
	PIX_CID client_name;
	GET_VAL_FROM_STACK( client_name, 0, PIX_CID );
	bool need_to_free = 0;
	UTF8_CHAR* name = pix_vm_make_cstring_from_container( client_name, &need_to_free, vm );
	if( name == 0 ) name = (UTF8_CHAR*)"Pixilang MIDI Client";
	
	rv = pix_vm_new_container( -1, sizeof( sundog_midi_client ), 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( rv, vm );
	if( c )
	{
	    if( sundog_midi_client_open( c, name ) )
	    {
		//Error:
		pix_vm_remove_container( rv, vm );
		rv = -1;
	    }
	}
	
	if( need_to_free ) mem_free( name );
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_close_client( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = -1;
    
    if( pars_num > 0 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    if( sundog_midi_client_close( c ) == 0 )
	    {
		pix_vm_remove_container( cnum, vm );
		rv = 0;
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_get_device( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    PIX_CID rv = -1;
    
    if( pars_num >= 3 )
    {
	PIX_CID cnum;
	int dev_num;
	int flags;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( dev_num, 1, int );
	GET_VAL_FROM_STACK( flags, 2, int );

	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    UTF8_CHAR** devices = 0;
	    int devs = sundog_midi_client_get_devices( c, &devices, flags );
	    if( devs > 0 && devices )
	    {
		if( dev_num < devs )
		{
		    UTF8_CHAR* name = devices[ dev_num ];
		    if( name )
		    {
			rv = pix_vm_new_container( -1, mem_strlen( name ), 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
			mem_copy( pix_vm_get_container_data( rv, vm ), name, mem_strlen( name ) );
		    }
		}
		for( int i = 0; i < devs; i++ )
		{
		    mem_free( devices[ i ] );
		}
		mem_free( devices );
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}
    
void fn_midi_open_port( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = -1;
    
    if( pars_num >= 4 )
    {
	PIX_CID cnum;
	PIX_CID port_name_cont;
	PIX_CID dev_name_cont;
	int flags;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port_name_cont, 1, PIX_CID );
	GET_VAL_FROM_STACK( dev_name_cont, 2, PIX_CID );
	GET_VAL_FROM_STACK( flags, 3, int );
	bool need_to_free1 = 0;
	bool need_to_free2 = 0;
	UTF8_CHAR* port_name = pix_vm_make_cstring_from_container( port_name_cont, &need_to_free1, vm );
	UTF8_CHAR* dev_name = pix_vm_make_cstring_from_container( dev_name_cont, &need_to_free2, vm );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    rv = sundog_midi_client_open_port( c, (const UTF8_CHAR*)port_name, (const UTF8_CHAR*)dev_name, flags );
	}
	
	if( need_to_free1 ) mem_free( port_name );
	if( need_to_free2 ) mem_free( dev_name );
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_reopen_port( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = -1;
    
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	int port;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port, 1, int );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    rv = sundog_midi_client_reopen_port( c, port );
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_close_port( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = -1;
    
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	int port;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port, 1, int );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    rv = sundog_midi_client_close_port( c, port );
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_get_event( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = 0;
    
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	int port;
	PIX_CID data;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port, 1, int );
	GET_VAL_FROM_STACK( data, 2, PIX_CID );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    sundog_midi_event* evt = sundog_midi_client_get_event( c, port );
	    if( evt && evt->size > 0 && evt->data )
	    {
		pix_vm_container* data_cont = pix_vm_get_container( data, vm );
		if( data_cont )
		{
		    if( data_cont->size * g_pix_container_type_sizes[ data_cont->type ] < evt->size )
		    {
			pix_vm_resize_container( data, evt->size, 1, PIX_CONTAINER_TYPE_INT8, vm );
		    }
		    mem_copy( data_cont->data, evt->data, evt->size );
		}
		rv = evt->size;
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_get_event_time( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    PIX_INT rv = -1;
    
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	int port;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port, 1, int );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    sundog_midi_event* evt = sundog_midi_client_get_event( c, port );
	    if( evt )
	    {
		rv = evt->t;
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_next_event( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = -1;
    
    if( pars_num >= 2 )
    {
	PIX_CID cnum;
	int port;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port, 1, int );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    rv = sundog_midi_client_next_event( c, port );
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_midi_send_event( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = -1;
    
    if( pars_num >= 5 )
    {
	PIX_CID cnum;
	int port;
	PIX_CID data;
	PIX_INT size;
	PIX_INT t;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( port, 1, int );
	GET_VAL_FROM_STACK( data, 2, PIX_CID );
	GET_VAL_FROM_STACK( size, 3, PIX_INT );
	GET_VAL_FROM_STACK( t, 4, PIX_INT );
	
	sundog_midi_client* c = (sundog_midi_client*)pix_vm_get_container_data( cnum, vm );
	if( c )
	{
	    pix_vm_container* data_cont = pix_vm_get_container( data, vm );
	    if( data_cont && data_cont->data && data_cont->size * g_pix_container_type_sizes[ data_cont->type ] >= size )
	    {
		sundog_midi_event evt;
		evt.t = (ticks_hr_t)t;
		evt.size = (size_t)size;
		evt.data = (uchar*)data_cont->data;
	        rv = sundog_midi_client_send_event( c, port, &evt );
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//
// Timers
//
	    
void fn_start_timer( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int tnum = 0;
    if( pars_num >= 1 )
    {
	GET_VAL_FROM_STACK( tnum, 0, int );
    }
    if( (unsigned)tnum < (unsigned)( sizeof( vm->timers ) / sizeof( uint ) ) )
    {
	vm->timers[ tnum ] = time_ticks();
    }
}

void fn_get_timer( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int tnum = 0;
    if( pars_num >= 1 )
    {
	GET_VAL_FROM_STACK( tnum, 0, int );
    }
    if( (unsigned)tnum < (unsigned)( sizeof( vm->timers ) / sizeof( uint ) ) )
    {
	uint t = time_ticks() - vm->timers[ tnum ];
	t *= 1000;
	t /= time_ticks_per_second();
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)t;
    }
}

void fn_get_year( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)time_year();
}

void fn_get_month( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)time_month();
}

void fn_get_day( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)time_day();
}

void fn_get_hours( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)time_hours();
}

void fn_get_minutes( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)time_minutes();
}

void fn_get_seconds( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)time_seconds();
}

void fn_get_ticks( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (uint)time_ticks_hires();
}

void fn_get_tps( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (uint)time_ticks_per_second_hires();
}

void fn_sleep( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int ms = 0;
    if( pars_num >= 1 )
    {
	GET_VAL_FROM_STACK( ms, 0, int );
	time_sleep( ms );
    }
}

//
// Events
//
	    
void fn_get_event( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_get_event( vm );
}

void fn_set_quit_action( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 1 )
    {
	GET_VAL_FROM_STACK( vm->quit_action, 0, char );
    }
}

//
// Threads
//
	    
void fn_thread_create( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_INT rv = -1;
    if( pars_num >= 2 )
    {
	rv = pix_vm_create_active_thread( -1, vm );
	if( rv >= 0 )
	{
	    PIX_ADDR function;
	    GET_VAL_FROM_STACK( function, 0, PIX_ADDR );
	    if( IS_ADDRESS_CORRECT( function ) )
	    {
		function &= PIX_INT_ADDRESS_MASK;
		pix_vm_function fun;
		PIX_VAL pp[ 2 ];
		char pp_types[ 2 ];
		fun.p = pp;
		fun.p_types = pp_types;
    		fun.addr = function;
		fun.p[ 0 ].i = rv;
		fun.p_types[ 0 ] = 0;
		fun.p[ 1 ] = stack[ sp + 1 ];
		fun.p_types[ 1 ] = stack_types[ sp + 1 ];
		fun.p_num = 2;
		pix_vm_run( rv, 1, &fun, PIX_VM_CALL_FUNCTION, vm );
	    }
	    else
	    {
		dprint( "thread_create() error: wrong thread address %d\n", (int)function );
	    }
	}
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_thread_destroy( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    if( pars_num >= 2 )
    {
	int thread_num;
	PIX_INT timeout;
	int time_counter = 0;
	GET_VAL_FROM_STACK( thread_num, 0, int );
	GET_VAL_FROM_STACK( timeout, 1, int );
	if( (unsigned)thread_num < (unsigned)PIX_VM_THREADS )
	{
	    if( timeout == PIX_INT_MAX_POSITIVE ) timeout = 0x7FFFFFFF;
	    pix_vm_thread* t = vm->th[ thread_num ];
	    if( t && t->thread_open )
	    {
		rv = sundog_thread_destroy( &t->th, (int)timeout );
		sundog_mutex_lock( &vm->th_mutex );
		mem_free( t );
		vm->th[ thread_num ] = 0;
		sundog_mutex_unlock( &vm->th_mutex );
	    }
	}
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_mutex_create( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID rv = -1;
    
    rv = pix_vm_new_container( -1, sizeof( sundog_mutex ), 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
    if( rv >= 0 )
    {
	sundog_mutex* m = (sundog_mutex*)pix_vm_get_container_data( rv, vm );
	if( m == 0 || sundog_mutex_init( m, 0 ) )
	{
	    pix_vm_remove_container( rv, vm );
	    rv = -1;
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_mutex_destroy( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    int rv = 0;

    if( pars_num > 0 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c )
	{
	    if( c->size >= sizeof( sundog_mutex ) && c->data )
	    {
		rv = sundog_mutex_destroy( (sundog_mutex*)c->data );
		pix_vm_remove_container( cnum, vm );
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_mutex_lock( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = 0;
    
    if( pars_num > 0 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c )
	{
	    if( c->size >= sizeof( sundog_mutex ) && c->data )
	    {
		rv = sundog_mutex_lock( (sundog_mutex*)c->data );
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_mutex_trylock( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = 0;
    
    if( pars_num > 0 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c )
	{
	    if( c->size >= sizeof( sundog_mutex ) && c->data )
	    {
		rv = sundog_mutex_trylock( (sundog_mutex*)c->data );
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_mutex_unlock( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = 0;
    
    if( pars_num > 0 )
    {
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	pix_vm_container* c = pix_vm_get_container( cnum, vm );
	if( c )
	{
	    if( c->size >= sizeof( sundog_mutex ) && c->data )
	    {
		rv = sundog_mutex_unlock( (sundog_mutex*)c->data );
	    }
	}
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//
// Mathematical functions
//

void fn_acos( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = acos( v );
    }
}

void fn_acosh( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = acosh( v );
    }
}

void fn_asin( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = asin( v );
    }
}

void fn_asinh( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = asinh( v );
    }
}

void fn_atan( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = atan( v );
    }
}

void fn_atanh( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = atanh( v );
    }
}

void fn_ceil( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = ceil( v );
    }
}

void fn_cos( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = cos( v );
    }
}

void fn_cosh( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
	    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = cosh( v );
    }
}

void fn_exp( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = exp( v );
    }
}

void fn_exp2( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = pow( 2.0, v );
    }
}

void fn_expm1( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = expm1( v );
    }
}

void fn_abs( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = abs( v );
    }
}

void fn_floor( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = floor( v );
    }
}

void fn_mod( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 2 )
    {
	PIX_FLOAT v1, v2;
	GET_VAL_FROM_STACK( v1, 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( v2, 1, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = fmod( v1, v2 );
    }
}

void fn_log( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = log( v );
    }
}

void fn_log2( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = log( v ) / log( 2 );
    }
}

void fn_log10( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = log10( v );
    }
}

void fn_pow( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 2 )
    {
	PIX_FLOAT v1, v2;
	GET_VAL_FROM_STACK( v1, 0, PIX_FLOAT );
	GET_VAL_FROM_STACK( v2, 1, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = pow( v1, v2 );
    }
}

void fn_sin( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = sin( v );
    }
}

void fn_sinh( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = sinh( v );
    }
}

void fn_sqrt( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = sqrt( v );
    }
}

void fn_tan( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = tan( v );
    }
}

void fn_tanh( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_FLOAT v;
	GET_VAL_FROM_STACK( v, 0, PIX_FLOAT );
	stack_types[ sp + ( pars_num - 1 ) ] = 1;
	stack[ sp + ( pars_num - 1 ) ].f = tanh( v );
    }
}

void fn_rand( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = pseudo_random_with_seed( &vm->random );
}

void fn_rand_seed( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	PIX_INT v;
	GET_VAL_FROM_STACK( v, 0, PIX_INT );
	vm->random = v;
    }
}

//
// Data processing
//
	
void fn_op_cn( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_VAL retval;
    retval.i = 0;
    char retval_type = 0;
    
    if( pars_num >= 2 )
    {
	int opcode;
	PIX_CID cnum;
	char val_type = 0;
	PIX_VAL val;
	val.i = 0;
	PIX_INT x = 0;
	PIX_INT y = 0;
	PIX_INT xsize = 0;
	PIX_INT ysize = 0;
	
	GET_VAL_FROM_STACK( opcode, 0, int );
	GET_VAL_FROM_STACK( cnum, 1, PIX_CID );
	if( pars_num >= 3 )
	{
	    val_type = stack_types[ sp + 2 ];
	    val = stack[ sp + 2 ];
	}
	
	if( pars_num == 5 )
	{
	    //1D:
	    GET_VAL_FROM_STACK( x, 3, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 4, PIX_INT );
	    if( xsize <= 0 ) return;
	}
	if( pars_num == 7 )
	{
	    //2D:
	    GET_VAL_FROM_STACK( x, 3, PIX_INT );
	    GET_VAL_FROM_STACK( y, 4, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 5, PIX_INT );
	    GET_VAL_FROM_STACK( ysize, 6, PIX_INT );
	    if( xsize <= 0 ) return;
	    if( ysize <= 0 ) return;
	}
	
	pix_vm_op_cn( opcode, cnum, val_type, val, x, y, xsize, ysize, &retval, &retval_type, vm );
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = retval_type;
    stack[ sp + ( pars_num - 1 ) ] = retval;
}

void fn_op_cc( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 3 )
    {
	int opcode;
	PIX_CID cnum1;
	PIX_CID cnum2;
	PIX_INT dest_x = 0;
	PIX_INT dest_y = 0;
	PIX_INT src_x = 0;
	PIX_INT src_y = 0;
	PIX_INT xsize = 0;
	PIX_INT ysize = 0;
	
	GET_VAL_FROM_STACK( opcode, 0, int );
	GET_VAL_FROM_STACK( cnum1, 1, PIX_CID );
	GET_VAL_FROM_STACK( cnum2, 2, PIX_CID );
	
	if( pars_num == 6 )
	{
	    //1D:
	    GET_VAL_FROM_STACK( dest_x, 3, PIX_INT );
	    GET_VAL_FROM_STACK( src_x, 4, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 5, PIX_INT );
	    if( xsize <= 0 ) return;
	}
	if( pars_num == 9 )
	{
	    //2D:
	    GET_VAL_FROM_STACK( dest_x, 3, PIX_INT );
	    GET_VAL_FROM_STACK( dest_y, 4, PIX_INT );
	    GET_VAL_FROM_STACK( src_x, 5, PIX_INT );
	    GET_VAL_FROM_STACK( src_y, 6, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 7, PIX_INT );
	    GET_VAL_FROM_STACK( ysize, 8, PIX_INT );
	    if( xsize <= 0 ) return;
	    if( ysize <= 0 ) return;
	}
	
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = pix_vm_op_cc( opcode, cnum1, cnum2, dest_x, dest_y, src_x, src_y, xsize, ysize, vm );
    }
}

void fn_op_ccn( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 4 )
    {
	int opcode;
	PIX_CID cnum1;
	PIX_CID cnum2;
	char val_type;
	PIX_VAL val;
	PIX_INT dest_x = 0;
	PIX_INT dest_y = 0;
	PIX_INT src_x = 0;
	PIX_INT src_y = 0;
	PIX_INT xsize = 0;
	PIX_INT ysize = 0;
	
	GET_VAL_FROM_STACK( opcode, 0, int );
	GET_VAL_FROM_STACK( cnum1, 1, PIX_CID );
	GET_VAL_FROM_STACK( cnum2, 2, PIX_CID );
	val_type = stack_types[ sp + 3 ];
	val = stack[ sp + 3 ];
	
	if( pars_num == 7 )
	{
	    //1D:
	    GET_VAL_FROM_STACK( dest_x, 4, PIX_INT );
	    GET_VAL_FROM_STACK( src_x, 5, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 6, PIX_INT );
	    if( xsize <= 0 ) return;
	}
	if( pars_num == 10 )
	{
	    //2D:
	    GET_VAL_FROM_STACK( dest_x, 4, PIX_INT );
	    GET_VAL_FROM_STACK( dest_y, 5, PIX_INT );
	    GET_VAL_FROM_STACK( src_x, 6, PIX_INT );
	    GET_VAL_FROM_STACK( src_y, 7, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 8, PIX_INT );
	    GET_VAL_FROM_STACK( ysize, 9, PIX_INT );
	    if( xsize <= 0 ) return;
	    if( ysize <= 0 ) return;
	}
	
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = pix_vm_op_ccn( opcode, cnum1, cnum2, val_type, val, dest_x, dest_y, src_x, src_y, xsize, ysize, vm );
    }
}

void fn_generator( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 2 )
    {
	int opcode;
	PIX_CID cnum;
	PIX_FLOAT fval[ 4 ];
	fval[ 0 ] = 0; //Phase
	fval[ 1 ] = 1; //Amplitude
	fval[ 2 ] = 0; //Delta X
	fval[ 3 ] = 0; //Delta Y
	PIX_INT x = 0;
	PIX_INT y = 0;
	PIX_INT xsize = 0;
	PIX_INT ysize = 0;
	
	GET_VAL_FROM_STACK( opcode, 0, int );
	GET_VAL_FROM_STACK( cnum, 1, PIX_CID );
	if( pars_num >= 3 ) GET_VAL_FROM_STACK( fval[ 0 ], 2, PIX_FLOAT );
	if( pars_num >= 4 ) GET_VAL_FROM_STACK( fval[ 1 ], 3, PIX_FLOAT );
	if( pars_num >= 5 ) GET_VAL_FROM_STACK( fval[ 2 ], 4, PIX_FLOAT );
	if( pars_num >= 6 ) GET_VAL_FROM_STACK( fval[ 3 ], 5, PIX_FLOAT );
	
	if( pars_num == 8 )
	{
	    //1D:
	    GET_VAL_FROM_STACK( x, 6, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 7, PIX_INT );
	    if( xsize <= 0 ) return;
	}
	if( pars_num == 10 )
	{
	    //2D:
	    GET_VAL_FROM_STACK( x, 6, PIX_INT );
	    GET_VAL_FROM_STACK( y, 7, PIX_INT );
	    GET_VAL_FROM_STACK( xsize, 8, PIX_INT );
	    GET_VAL_FROM_STACK( ysize, 9, PIX_INT );
	    if( xsize <= 0 ) return;
	    if( ysize <= 0 ) return;
	}
	
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = pix_vm_generator( opcode, cnum, fval, x, y, xsize, ysize, vm );
    }
}

void fn_wavetable_generator( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 8 )
    {
	PIX_CID dest;
	PIX_INT dest_off;
	PIX_INT dest_len;
	PIX_CID table;
	PIX_CID amp;
	PIX_CID amp_delta;
	PIX_CID pos;
	PIX_CID pos_delta;
	GET_VAL_FROM_STACK( dest, 0, PIX_CID );
	GET_VAL_FROM_STACK( dest_off, 1, PIX_INT );
	GET_VAL_FROM_STACK( dest_len, 2, PIX_INT );
	GET_VAL_FROM_STACK( table, 3, PIX_CID );
	GET_VAL_FROM_STACK( amp, 4, PIX_CID );
	GET_VAL_FROM_STACK( amp_delta, 5, PIX_CID );
	GET_VAL_FROM_STACK( pos, 6, PIX_CID );
	GET_VAL_FROM_STACK( pos_delta, 7, PIX_CID );

	stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = pix_vm_wavetable_generator( dest, dest_off, dest_len, table, amp, amp_delta, pos, pos_delta, vm );
    }
}    

void fn_sampler( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 1 )
    {
	int pars;
	GET_VAL_FROM_STACK( pars, 0, int );
	if( (unsigned)pars >= (unsigned)PIX_VM_CONTAINERS ) return;
	pix_vm_container* pars_cont = vm->c[ pars ];
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = pix_vm_sampler( pars_cont, vm );
    }
}

void fn_envelope2p( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    if( pars_num >= 3 )
    {
	PIX_CID cnum;
	PIX_INT v1;
	PIX_INT v2;
	PIX_INT offset = 0;
	PIX_INT size = -1;
	char dc_off1_type = 0;
	char dc_off2_type = 0; 
	PIX_VAL dc_off1; dc_off1.i = 0;
	PIX_VAL dc_off2; dc_off2.i = 0;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	GET_VAL_FROM_STACK( v1, 1, PIX_INT );
	GET_VAL_FROM_STACK( v2, 2, PIX_INT );
	if( pars_num > 3 ) { GET_VAL_FROM_STACK( offset, 3, PIX_INT ); }
	if( pars_num > 4 ) { GET_VAL_FROM_STACK( size, 4, PIX_INT ); }
	if( pars_num > 5 ) { dc_off1_type = stack_types[ sp + 5 ]; dc_off1 = stack[ sp + 5 ]; }
	if( pars_num > 6 ) { dc_off2_type = stack_types[ sp + 6 ]; dc_off2 = stack[ sp + 6 ]; }
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
        stack[ sp + ( pars_num - 1 ) ].i = pix_vm_envelope2p( cnum, v1, v2, offset, size, dc_off1_type, dc_off1, dc_off2_type, dc_off2, vm );
    }
}

void fn_fft( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 3 )
    {
	int inv;
	PIX_CID cnum1;
	PIX_CID cnum2;
	int size = 0;
	GET_VAL_FROM_STACK( inv, 0, int );
	GET_VAL_FROM_STACK( cnum1, 1, PIX_CID );
	GET_VAL_FROM_STACK( cnum2, 2, PIX_CID );
	if( pars_num >= 4 )
	    GET_VAL_FROM_STACK( size, 3, int );
	if( (unsigned)cnum1 >= (unsigned)PIX_VM_CONTAINERS ) return;
	if( (unsigned)cnum2 >= (unsigned)PIX_VM_CONTAINERS ) return;
	if( size < 0 ) return;
	pix_vm_container* cont1 = vm->c[ cnum1 ];
	pix_vm_container* cont2 = vm->c[ cnum2 ];
	if( cont1 && cont2 && cont1->data && cont2->data )
	{
	    if( cont1->type != cont2->type ) return;
	    if( size == 0 ) size = (int)cont1->size;
	    if( cont1->type == PIX_CONTAINER_TYPE_FLOAT32 )
	    {
		fft_float32( inv, (float*)cont1->data, (float*)cont2->data, size );
	    }
	    if( cont1->type == PIX_CONTAINER_TYPE_FLOAT64 )
	    {
		fft_float64( inv, (double*)cont1->data, (double*)cont2->data, size );
	    }
	}
    }
}

void fn_replace_values( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int rv = -1;
    
    while( pars_num >= 3 )
    {
	PIX_CID dest;
	PIX_CID src;
	PIX_CID values;
	size_t dest_offset = 0;
	size_t src_offset = 0;
	size_t size = -1;
	GET_VAL_FROM_STACK( dest, 0, PIX_CID );
	GET_VAL_FROM_STACK( src, 1, PIX_CID );
	GET_VAL_FROM_STACK( values, 2, PIX_CID );
	while( 1 )
	{
	    if( pars_num >= 4 ) GET_VAL_FROM_STACK( dest_offset, 3, size_t ) else break;
	    if( pars_num >= 5 ) GET_VAL_FROM_STACK( src_offset, 4, size_t ) else break;
	    if( pars_num >= 6 ) GET_VAL_FROM_STACK( size, 5, size_t ) else break;
	    break;
	}
	if( (unsigned)dest >= (unsigned)PIX_VM_CONTAINERS ) break;
	if( (unsigned)src >= (unsigned)PIX_VM_CONTAINERS ) break;
	if( (unsigned)values >= (unsigned)PIX_VM_CONTAINERS ) break;
	pix_vm_container* dest_cont = vm->c[ dest ];
	pix_vm_container* src_cont = vm->c[ src ];
	pix_vm_container* values_cont = vm->c[ values ];
	if( dest_cont == 0 ) break;
	if( src_cont == 0 ) break;
	if( values_cont == 0 ) break;
	size_t values_num = values_cont->size;

	if( size == -1 ) size = dest_cont->size;
	if( dest_offset >= dest_cont->size ) break;
	if( src_offset >= src_cont->size ) break;
	if( dest_offset + size > dest_cont->size )
	{
	    size = dest_cont->size - dest_offset;
	}
	if( src_offset + size > src_cont->size )
	{
	    size = src_cont->size - src_offset;
	}

	if( dest_cont->type == values_cont->type )
	{
	    switch( src_cont->type )
	    {
		case PIX_CONTAINER_TYPE_INT8:
		    {
			uchar* s = (uchar*)src_cont->data + src_offset;
			switch( dest_cont->type )
			{
		    	    case PIX_CONTAINER_TYPE_INT8: { uchar* d = (uchar*)dest_cont->data + dest_offset; uchar* v = (uchar*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT16: { uint16* d = (uint16*)dest_cont->data + dest_offset; uint16* v = (uint16*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break; 
		    	    case PIX_CONTAINER_TYPE_INT32:
		    	    case PIX_CONTAINER_TYPE_FLOAT32:
		    		{ uint* d = (uint*)dest_cont->data + dest_offset; uint* v = (uint*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#if defined(PIX_INT64_ENABLED) || defined(PIX_FLOAT64_ENABLED)
			    case PIX_CONTAINER_TYPE_INT64:
			    case PIX_CONTAINER_TYPE_FLOAT64:
		    		{ uint64* d = (uint64*)dest_cont->data + dest_offset; uint64* v = (uint64*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#endif
			    default:
				break;
			}
		    }
		    break;
		case PIX_CONTAINER_TYPE_INT16:
		    {
			uint16* s = (uint16*)src_cont->data + src_offset;
			switch( dest_cont->type )
			{
		    	    case PIX_CONTAINER_TYPE_INT8: { uchar* d = (uchar*)dest_cont->data + dest_offset; uchar* v = (uchar*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT16: { uint16* d = (uint16*)dest_cont->data + dest_offset; uint16* v = (uint16*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT32:
		    	    case PIX_CONTAINER_TYPE_FLOAT32:
		    		{ uint* d = (uint*)dest_cont->data + dest_offset; uint* v = (uint*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#if defined(PIX_INT64_ENABLED) || defined(PIX_FLOAT64_ENABLED)
			    case PIX_CONTAINER_TYPE_INT64:
			    case PIX_CONTAINER_TYPE_FLOAT64:
		    		{ uint64* d = (uint64*)dest_cont->data + dest_offset; uint64* v = (uint64*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#endif
			    default:
				break;
			}
		    }
		    break;
		case PIX_CONTAINER_TYPE_INT32:
		    {
			uint* s = (uint*)src_cont->data + src_offset;
			switch( dest_cont->type )
			{
		    	    case PIX_CONTAINER_TYPE_INT8: { uchar* d = (uchar*)dest_cont->data + dest_offset; uchar* v = (uchar*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT16: { uint16* d = (uint16*)dest_cont->data + dest_offset; uint16* v = (uint16*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT32:
		    	    case PIX_CONTAINER_TYPE_FLOAT32:
		    		{ uint* d = (uint*)dest_cont->data + dest_offset; uint* v = (uint*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#if defined(PIX_INT64_ENABLED) || defined(PIX_FLOAT64_ENABLED)
			    case PIX_CONTAINER_TYPE_INT64:
			    case PIX_CONTAINER_TYPE_FLOAT64:
		    		{ uint64* d = (uint64*)dest_cont->data + dest_offset; uint64* v = (uint64*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#endif
			    default:
				break;
			}
		    }
		    break;
#ifdef PIX_INT64_ENABLED
		case PIX_CONTAINER_TYPE_INT64:
		    {
			uint64* s = (uint64*)src_cont->data + src_offset;
			switch( dest_cont->type )
			{
		    	    case PIX_CONTAINER_TYPE_INT8: { uchar* d = (uchar*)dest_cont->data + dest_offset; uchar* v = (uchar*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT16: { uint16* d = (uint16*)dest_cont->data + dest_offset; uint16* v = (uint16*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
		    	    case PIX_CONTAINER_TYPE_INT32:
		    	    case PIX_CONTAINER_TYPE_FLOAT32:
		    		{ uint* d = (uint*)dest_cont->data + dest_offset; uint* v = (uint*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#if defined(PIX_INT64_ENABLED) || defined(PIX_FLOAT64_ENABLED)
			    case PIX_CONTAINER_TYPE_INT64:
			    case PIX_CONTAINER_TYPE_FLOAT64:
		    		{ uint64* d = (uint64*)dest_cont->data + dest_offset; uint64* v = (uint64*)values_cont->data; for( size_t i = 0; i < size; i++ ) d[ i ] = v[ s[ i ] ]; rv = 0; } break;
#endif
			    default:
				break;
			}
		    }
		    break;
#endif
		default:
		    break;
	    }
	}
	break;
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//
// Dialogs
//
	    
void fn_file_dialog( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    bool name_need_to_free = 0;
    bool mask_need_to_free = 0;
    bool id_need_to_free = 0;
    bool defname_need_to_free = 0;
    UTF8_CHAR* name = 0;
    UTF8_CHAR* mask = 0;
    UTF8_CHAR* id = 0;
    UTF8_CHAR* defname = 0;
        
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 3 ) { err = 1; break; }
	PIX_CID cnum;
	GET_VAL_FROM_STACK( cnum, 0, PIX_CID );
	name = pix_vm_make_cstring_from_container( cnum, &name_need_to_free, vm );
	if( name == 0 ) { err = 1; break; }
	GET_VAL_FROM_STACK( cnum, 1, PIX_CID );
	mask = pix_vm_make_cstring_from_container( cnum, &mask_need_to_free, vm );
	if( mask == 0 ) { err = 1; break; }
	GET_VAL_FROM_STACK( cnum, 2, PIX_CID );
	id = pix_vm_make_cstring_from_container( cnum, &id_need_to_free, vm );
	if( id == 0 ) { err = 1; break; }
	if( pars_num > 3 )
	{
	    GET_VAL_FROM_STACK( cnum, 3, PIX_CID );
	    defname = pix_vm_make_cstring_from_container( cnum, &defname_need_to_free, vm );
	}
	break;
    }

    //Execute:
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;
    if( err == 0 )
    {
	vm->file_dialog_name = name;
	if( mask[ 0 ] == 0 )
	    vm->file_dialog_mask = 0;
	else
	    vm->file_dialog_mask = mask;
	vm->file_dialog_id = id;
	vm->file_dialog_def_name = defname;
	vm->file_dialog_request = 1;
	while( 1 )
	{
	    if( vm->file_dialog_request == 0 ) break;
	    time_sleep( 100 );
	}
	UTF8_CHAR* filename = vm->file_dialog_result;
	if( filename )
	{
	    size_t size = mem_strlen( filename );
	    void* data = MEM_NEW2( size );
	    mem_copy( data, filename, size );
	    stack[ sp + ( pars_num - 1 ) ].i = pix_vm_new_container( -1, (int)size, 1, PIX_CONTAINER_TYPE_INT8, data, 0, vm );
	}
	
	if( name_need_to_free ) mem_free( name );
	if( mask_need_to_free ) mem_free( mask );
	if( id_need_to_free ) mem_free( id );
	if( defname_need_to_free ) mem_free( defname );
    }
}

void fn_prefs_dialog( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    //Execute:
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = 0;

    vm->prefs_dialog_request = 1;
    while( 1 )
    {
        if( vm->prefs_dialog_request == 0 ) break;
        time_sleep( 100 );
    }
}

//
// Posix compatibility
//

//Open a stream:
void fn_fopen( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name;
    PIX_CID mode;
    pix_vm_container* name_cont;
    pix_vm_container* mode_cont;
    
    bool err = 0;

    //Get parameters:
    while( 1 )
    {
	if( pars_num < 2 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name, 0, PIX_CID );
	GET_VAL_FROM_STACK( mode, 1, PIX_CID );
	if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name ] == 0 ) { err = 1; break; }
	name_cont = vm->c[ name ];
	if( (unsigned)mode >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ mode ] == 0 ) { err = 1; break; }
	mode_cont = vm->c[ mode ];
	if( name_cont->size == 0 ) { err = 1; break; }
	if( mode_cont->size == 0 ) { err = 1; break; }
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free1 = 0;
	bool need_to_free2 = 0;
	UTF8_CHAR* ts1 = pix_vm_make_cstring_from_container( name, &need_to_free1, vm );
	UTF8_CHAR* ts2 = pix_vm_make_cstring_from_container( mode, &need_to_free2, vm );
	
	UTF8_CHAR* full_path = pix_compose_full_path( vm->base_path, ts1 );
	V3_FILE f = v3_open( full_path, ts2 );
	mem_free( full_path );
	
	if( need_to_free1 ) mem_free( ts1 );
	if( need_to_free2 ) mem_free( ts2 );
	
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)f;
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = 0;
    }
}

//Close a stream:
void fn_fclose( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	V3_FILE f;
	GET_VAL_FROM_STACK( f, 0, V3_FILE );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_close( f );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

//Put a byte on a stream:
void fn_fputc( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    int c;
    V3_FILE f;
    if( pars_num >= 2 ) 
    {
	GET_VAL_FROM_STACK( c, 0, int );
	GET_VAL_FROM_STACK( f, 1, V3_FILE );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_putc( c, f );
    }
    else if( pars_num == 1 )
    {
	GET_VAL_FROM_STACK( c, 0, int );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_putc( c, V3_STDOUT );
    }
}

//Put a string on a stream:
void fn_fputs( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_INT rv = -1;
    PIX_CID s = -1;
    V3_FILE f;
    if( pars_num >= 2 ) 
    {
	GET_VAL_FROM_STACK( s, 0, PIX_CID );
	GET_VAL_FROM_STACK( f, 1, V3_FILE );
    }
    else if( pars_num == 1 )
    {
	GET_VAL_FROM_STACK( s, 0, PIX_CID );
	f = V3_STDOUT;
    }
    if( (unsigned)s < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* cont = vm->c[ s ];
	if( cont )
	{
	    size_t str_len;
	    for( str_len = 0; str_len < cont->size; str_len++ )
		if( ((UTF8_CHAR*)cont->data)[ str_len ] == 0 )
		    break;
	    rv = (PIX_INT)v3_write( cont->data, 1, str_len, f );
	}
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

void fn_fgets_OR_fwrite_OR_fread( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    PIX_INT rv = 0;
    PIX_CID s = -1;
    PIX_INT size;
    V3_FILE f;
    if( pars_num >= 3 ) 
    {
	GET_VAL_FROM_STACK( s, 0, PIX_CID );
	GET_VAL_FROM_STACK( size, 1, PIX_INT );
	GET_VAL_FROM_STACK( f, 2, V3_FILE );
    }
    if( (unsigned)s < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* cont = vm->c[ s ];
	if( cont )
	{
	    size_t real_size = cont->size * g_pix_container_type_sizes[ cont->type ];
	    if( real_size < size )
		size = (PIX_INT)real_size;
	    switch( fn_num )
	    {
		case FN_FGETS: 
		    {
			//Get a string from a stream:
			rv = s;
			UTF8_CHAR* str = (UTF8_CHAR*)cont->data;
			size_t p;
			for( p = 0; p < real_size - 1; p++ )
			{
			    int c = v3_getc( f );
			    if( c == -1 ) break;
			    if( c == 0xD ) break;
			    if( c == 0xA ) break;
			    str[ p ] = (UTF8_CHAR)c;
			}
			str[ p ] = 0;
		    }
		    break;
		case FN_FWRITE: 
		    rv = (PIX_INT)v3_write( cont->data, 1, size, f ); 
		    break;
		case FN_FREAD: 
		    rv = (PIX_INT)v3_read( cont->data, 1, size, f ); 
		    break;
	    }
	}
    }
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = rv;
}

//Get a byte from a stream:
void fn_fgetc( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    V3_FILE f;
    if( pars_num >= 1 ) 
    {
	GET_VAL_FROM_STACK( f, 0, V3_FILE );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_getc( f );
    }
    else if( pars_num == 0 )
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_getc( V3_STDIN );
    }
}

void fn_feof_OF_fflush_OR_ftell( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	V3_FILE f;
	GET_VAL_FROM_STACK( f, 0, V3_FILE );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	switch( fn_num )
	{
	    case FN_FEOF: stack[ sp + ( pars_num - 1 ) ].i = v3_eof( f ); break;
	    case FN_FFLUSH: stack[ sp + ( pars_num - 1 ) ].i = v3_flush( f ); break;
	    case FN_FTELL: stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)v3_tell( f ); break;
	}
    }
}

void fn_fseek( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 3 )
    {
	V3_FILE f;
	PIX_INT offset;
	PIX_INT mode;
	GET_VAL_FROM_STACK( f, 0, V3_FILE );
	GET_VAL_FROM_STACK( offset, 1, PIX_INT );
	GET_VAL_FROM_STACK( mode, 2, PIX_INT );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = v3_seek( f, offset, mode );
    }
}

//Appends a copy of the source string to the destination string:
void fn_strcat( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID s1;
    PIX_CID s2;
    pix_vm_container* s1_cont;
    pix_vm_container* s2_cont;
    PIX_INT off1 = 0;
    PIX_INT off2 = 0;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 2 ) { err = 1; break; }
	GET_VAL_FROM_STACK( s1, 0, PIX_CID );
	if( pars_num >= 3 )
	{
	    GET_VAL_FROM_STACK( off1, 1, PIX_INT );
	    GET_VAL_FROM_STACK( s2, 2, PIX_CID );
	    if( pars_num >= 4 )
	    {
		GET_VAL_FROM_STACK( off2, 3, PIX_INT );
	    }
	}
	else
	{
	    GET_VAL_FROM_STACK( s2, 1, PIX_CID );
	}
	if( (unsigned)s1 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	s1_cont = vm->c[ s1 ];
	if( (unsigned)s2 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	s2_cont = vm->c[ s2 ];
	if( s1_cont == 0 ) { err = 1; break; }
	if( s2_cont == 0 ) { err = 1; break; }
	break;
    }
    
    //Execute:
    while( err == 0 )
    {
	size_t s1_size = s1_cont->size * g_pix_container_type_sizes[ s1_cont->type ];
	size_t s2_size = s2_cont->size * g_pix_container_type_sizes[ s2_cont->type ];
	if( off2 >= s2_size ) break;
	s2_size -= off2;
	size_t s1_len;
	size_t s2_len;
	UTF8_CHAR* s1_ptr = (UTF8_CHAR*)s1_cont->data;
	UTF8_CHAR* s2_ptr = (UTF8_CHAR*)s2_cont->data;
	s2_ptr += off2;
	for( s2_len = 0; s2_len < s2_size; s2_len++ )
	    if( s2_ptr[ s2_len ] == 0 ) break;
	if( s2_len == 0 ) break;
	if( off1 >= s1_size )
	{
	    s1_size = off1 + s2_len;
	    if( pix_vm_resize_container( s1, (PIX_INT)( s1_size ), 1, PIX_CONTAINER_TYPE_INT8, vm ) ) break;
	    s1_size -= off1;
	    s1_ptr = (UTF8_CHAR*)s1_cont->data;
	    s1_ptr += off1;
	    mem_copy( s1_ptr, s2_ptr, s2_len );
	}
	else
	{
	    s1_size -= off1;
	    s1_ptr += off1;
	    for( s1_len = 0; s1_len < s1_size; s1_len++ )
		if( s1_ptr[ s1_len ] == 0 ) break;
	    if( s1_len + s2_len > s1_size )
	    {
		if( pix_vm_resize_container( s1, (PIX_INT)( off1 + s1_len + s2_len ), 1, PIX_CONTAINER_TYPE_INT8, vm ) ) break;
		s1_ptr = (UTF8_CHAR*)s1_cont->data;
		s1_ptr += off1;
		s1_size = s1_len + s2_len;
	    }
	    else 
	    {
		if( s1_len + s2_len < s1_size )
		    s1_ptr[ s1_len + s2_len ] = 0;
	    }
	    mem_copy( s1_ptr + s1_len, s2_ptr, s2_len );
	}
	break;
    }
}

//Compare two strings /
//Returns a offset of the first occurrence of str2 in str1, or -1 if str2 is not part of str1:
void fn_strcmp_OR_strstr( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID s1;
    PIX_CID s2;
    pix_vm_container* s1_cont;
    pix_vm_container* s2_cont;
    PIX_INT off1 = 0;
    PIX_INT off2 = 0;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 2 ) { err = 1; break; }
	GET_VAL_FROM_STACK( s1, 0, PIX_CID );
	if( pars_num >= 3 )
	{
	    GET_VAL_FROM_STACK( off1, 1, PIX_INT );
	    GET_VAL_FROM_STACK( s2, 2, PIX_CID );
	    if( pars_num >= 4 )
	    {
		GET_VAL_FROM_STACK( off2, 3, PIX_INT );
	    }
	}
	else
	{
	    GET_VAL_FROM_STACK( s2, 1, PIX_CID );
	}
	if( (unsigned)s1 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	s1_cont = vm->c[ s1 ];
	if( (unsigned)s2 >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	s2_cont = vm->c[ s2 ];
	if( s1_cont == 0 ) { err = 1; break; }
	if( s2_cont == 0 ) { err = 1; break; }
	break;
    }
    
    //Execute:
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    if( err == 0 )
    {
	size_t s1_size = s1_cont->size * g_pix_container_type_sizes[ s1_cont->type ];
	size_t s2_size = s2_cont->size * g_pix_container_type_sizes[ s2_cont->type ];
	if( off1 > s1_size ) { s1_size = 0; off1 = 0; }
	if( off2 > s2_size ) { s2_size = 0; off2 = 0; }
	s1_size -= off1;
	s2_size -= off2;
	size_t s1_len;
	size_t s2_len;
	UTF8_CHAR* s1_ptr = (char*)s1_cont->data;
	UTF8_CHAR* s2_ptr = (char*)s2_cont->data;
	s1_ptr += off1;
	s2_ptr += off2;
	for( s1_len = 0; s1_len < s1_size; s1_len++ )
	    if( s1_ptr[ s1_len ] == 0 ) break;
	for( s2_len = 0; s2_len < s2_size; s2_len++ )
	    if( s2_ptr[ s2_len ] == 0 ) break;
	UTF8_CHAR* s1_str = 0;
	UTF8_CHAR* s2_str = 0;
	if( s1_len == s1_size )
	{
	    s1_str = (UTF8_CHAR*)MEM_NEW2( s1_len + 1 );
	    mem_copy( s1_str, s1_ptr, s1_len );
	    s1_str[ s1_len ] = 0;
	}
	else 
	{
	    s1_str = s1_ptr;
	}
	if( s2_len == s2_size )
	{
	    s2_str = (UTF8_CHAR*)MEM_NEW2( s2_len + 1 );
	    mem_copy( s2_str, s2_ptr, s2_len );
	    s2_str[ s2_len ] = 0;
	}
	else 
	{
	    s2_str = s2_ptr;
	}
	if( fn_num == FN_STRCMP )
	{
	    stack[ sp + ( pars_num - 1 ) ].i = mem_strcmp( s1_str, s2_str );
	}
	else 
	{
	    UTF8_CHAR* substr = strstr( s1_str, s2_str );
	    if( substr == 0 )
		stack[ sp + ( pars_num - 1 ) ].i = -1;
	    else
		stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)( substr - s1_str );
	}
	if( s1_len == s1_size ) mem_free( s1_str );
	if( s2_len == s2_size ) mem_free( s2_str );
    }
    else 
    {
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

//Returns the length of the string:
void fn_strlen( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID str;
    PIX_INT off = 0;

    //Get parameters:
    if( pars_num < 1 ) 
    { 
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = 0;
	return;
    }
    
    GET_VAL_FROM_STACK( str, 0, PIX_CID );
    if( pars_num >= 2 )
    {
	GET_VAL_FROM_STACK( off, 1, PIX_INT );
    }
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)pix_vm_get_container_strlen( str, (size_t)off, vm );
}

//Writes into the array pointed by str a C string consisting on a sequence of data formatted as the format argument specifies:
void fn_sprintf( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;

    bool err = 0;
    int i2;
    
    PIX_CID str = -1;
    PIX_CID format = -1;
    pix_vm_container* str_cont = 0;
    pix_vm_container* format_cont = 0;
    int args_off;
    
    //Get parameters:
    if( fn_num == FN_SPRINTF )
    {
	//sprintf:
	while( 1 )
	{
	    if( pars_num < 2 ) { err = 1; break; }
	    GET_VAL_FROM_STACK( str, 0, PIX_CID );
	    GET_VAL_FROM_STACK( format, 1, PIX_CID );
	    if( (unsigned)str >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	    if( (unsigned)format >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	    str_cont = vm->c[ str ];
	    format_cont = vm->c[ format ];
	    if( str_cont == 0 ) { err = 1; break; }
	    if( format_cont == 0 ) { err = 1; break; }
	    args_off = 2;
	    break;
	}
    }
    else
    {
	//printf:
	while( 1 )
	{
	    if( pars_num < 1 ) { err = 1; break; }
	    GET_VAL_FROM_STACK( format, 0, PIX_CID );
	    if( (unsigned)format >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	    format_cont = vm->c[ format ];
	    if( format_cont == 0 ) { err = 1; break; }
	    args_off = 1;
	    break;
	}
    }
    
    //Execute:
    if( err == 0 ) 
    {
	size_t format_size = format_cont->size * g_pix_container_type_sizes[ format_cont->type ];
	size_t str_size;
	size_t str_ptr;
	UTF8_CHAR* cstr = 0;
	if( fn_num == FN_SPRINTF )
	{
	    //sprintf:
	    str_size = str_cont->size * g_pix_container_type_sizes[ format_cont->type ];
	    str_ptr = 0;
	}
	else
	{
	    //printf:
	    str_size = 256;
	    str_ptr = 0;
	    cstr = (UTF8_CHAR*)MEM_NEW2( str_size );
	    if( cstr == 0 ) goto sprintf_error;
	}
	int arg_num = 0;
	UTF8_CHAR* format_str = (UTF8_CHAR*)format_cont->data;
	for( size_t i = 0; i < format_size; i++ )
	{
	    UTF8_CHAR c = format_str[ i ];
	    bool c_to_output = 0;
	    if( c == '%' )
	    {
		i++;
		if( i >= format_size ) break;
		c = format_str[ i ];
		
		//Parse format:
		    
		UTF8_CHAR flags[ 2 ];
		flags[ 0 ] = 0;
		flags[ 1 ] = 0;
		switch( c )
		{
		    case '-':
		    case '+':
		    case ' ':
		    case '#':
		    case '0':
			flags[ 0 ] = c;
			break;
		}
		if( flags[ 0 ] )
		{
		    i++;
		    if( i >= format_size ) break;
		    c = format_str[ i ];
		}
		
		UTF8_CHAR width[ 4 ];
		i2 = 0;
		while( 1 )
		{
		    if( ( c >= '0' && c <= '9' ) || c == '*' )
		    {
			if( i2 < 3 )
			    width[ i2 ] = c;
			i2++;
			i++;
			if( i >= format_size ) break;
			c = format_str[ i ];
		    }
		    else break;
		}
		if( i >= format_size ) break;
		if( i2 < 4 ) width[ i2 ] = 0;
		
		UTF8_CHAR prec[ 5 ];
		i2 = 0;
		while( 1 )
		{
		    if( ( c >= '0' && c <= '9' ) || c == '*' || c == '.' )
		    {
			if( i2 < 4 )
			    prec[ i2 ] = c;
			i2++;
			i++;
			if( i >= format_size ) break;
			c = format_str[ i ];
		    }
		    else break;
		}
		if( i >= format_size ) break;
		if( i2 < 5 ) prec[ i2 ] = 0;
		
		UTF8_CHAR len[ 2 ];
		len[ 0 ] = 0;
		len[ 1 ] = 0;
		switch( c )
		{
		    case 'h':
		    case 'l':
		    case 'L':
			len[ 0 ] = c;
			break;
		}
		if( len[ 0 ] )
		{
		    i++;
		    if( i >= format_size ) break;
		    c = format_str[ i ];
		}
		
		UTF8_CHAR specifier[ 2 ];
		specifier[ 0 ] = c;
		specifier[ 1 ] = 0;
		
		if( specifier[ 0 ] == '%' )
		{
		    c_to_output = 1;
		}
		else 
		{
		    UTF8_CHAR arg_format[ 16 ];
		    arg_format[ 0 ] = 0;
		    strncat( arg_format, "%", 16 );
		    strncat( arg_format, flags, 16 );
		    strncat( arg_format, width, 16 );
		    strncat( arg_format, prec, 16 );
		    strncat( arg_format, len, 16 );
		    strncat( arg_format, specifier, 16 );
		    switch( specifier[ 0 ] )
		    {
			case 's': //String of characters:
			    {
				PIX_CID arg_str;
				GET_VAL_FROM_STACK( arg_str, args_off + arg_num, PIX_CID );
				arg_num++;
				if( (unsigned)arg_str >= (unsigned)PIX_VM_CONTAINERS ) break;
				pix_vm_container* arg_str_cont = vm->c[ arg_str ];
				if( arg_str_cont == 0 ) break;
				size_t arg_str_size = pix_vm_get_container_strlen( arg_str, 0, vm );
				if( str_ptr + arg_str_size > str_size )
				{
				    if( fn_num == FN_SPRINTF )
				    {
					//sprintf:
					if( pix_vm_resize_container( str, (PIX_INT)( str_ptr + arg_str_size + 8 ), 1, PIX_CONTAINER_TYPE_INT8, vm ) ) 
					    break;
				    }
				    else
				    {
					//printf:
					cstr = (UTF8_CHAR*)mem_resize( cstr, str_ptr + arg_str_size + 8 );
					if( cstr == 0 )
					    break;
				    }
				    str_size = str_ptr + arg_str_size + 8;
				}
				for( i2 = 0; i2 < arg_str_size; i2++ )
				{
				    if( fn_num == FN_SPRINTF )
				    {
					//sprintf:
					((UTF8_CHAR*)str_cont->data)[ str_ptr ] = ((UTF8_CHAR*)arg_str_cont->data)[ i2 ];
				    }
				    else
				    {
					//printf:
					cstr[ str_ptr ] = ((UTF8_CHAR*)arg_str_cont->data)[ i2 ];
				    }
				    str_ptr++;
				}
			    }
			    break;
			case 'c': //Character:
			case 'd': //Signed decimal integer:
			case 'i': //Signed decimal integer:
			case 'o': //Signed octal:
			case 'u': //Unsigned decimal integer:
			case 'x': //Unsigned hexadecimal integer:
			case 'X': //Unsigned hexadecimal integer (capital letters):
			    {
				PIX_INT arg_int;
				GET_VAL_FROM_STACK( arg_int, args_off + arg_num, PIX_INT );
				arg_num++;
				UTF8_CHAR ts[ 32 ];
				snprintf( ts, 32, arg_format, (int)arg_int );
				for( i2 = 0; i2 < 32; i2++ )
				{
				    if( ts[ i2 ] == 0 ) break;
				    if( str_ptr >= str_size )
				    {
					if( fn_num == FN_SPRINTF )
					{
					    //sprintf:
					    if( pix_vm_resize_container( str, (PIX_INT)str_ptr + 8, 1, PIX_CONTAINER_TYPE_INT8, vm ) ) 
						break;
					}
					else
					{
					    //printf:
					    cstr = (UTF8_CHAR*)mem_resize( cstr, str_ptr + 8 );
					    if( cstr == 0 ) break;
					}
					str_size = str_ptr + 8;
				    }
				    if( fn_num == FN_SPRINTF )
				    {
					//sprintf:
					((UTF8_CHAR*)str_cont->data)[ str_ptr ] = ts[ i2 ];
				    }
				    else
				    {
					//printf:
					cstr[ str_ptr ] = ts[ i2 ];
				    }
				    str_ptr++;
				}
			    }
			    break;
			case 'e': //Scientific notation (mantise/exponent) using e character:
			case 'E': //Scientific notation (mantise/exponent) using E character:
			case 'f': //Decimal floating point:
			case 'g': //Use the shorter of %e or %f:
			case 'G': //Use the shorter of %E or %f:
			    {
				PIX_FLOAT arg_float;
				GET_VAL_FROM_STACK( arg_float, args_off + arg_num, PIX_FLOAT );
				arg_num++;
				UTF8_CHAR ts[ 32 ];
				snprintf( ts, 32, arg_format, (float)arg_float );
				for( i2 = 0; i2 < 32; i2++ )
				{
				    if( ts[ i2 ] == 0 ) break;
				    if( str_ptr >= str_size )
				    {
					if( fn_num == FN_SPRINTF )
					{
					    //sprintf:
					    if( pix_vm_resize_container( str, (PIX_INT)str_ptr + 8, 1, PIX_CONTAINER_TYPE_INT8, vm ) ) 
						break;
					}
					else
					{
					    //printf:
					    cstr = (UTF8_CHAR*)mem_resize( cstr, str_ptr + 8 );
					    if( cstr == 0 ) break;
					}
					str_size = str_ptr + 8;
				    }
				    if( fn_num == FN_SPRINTF )
				    {
					//sprintf:
					((UTF8_CHAR*)str_cont->data)[ str_ptr ] = ts[ i2 ];
				    }
				    else
				    {
					//printf:
					cstr[ str_ptr ] = ts[ i2 ];
				    }
				    str_ptr++;
				}
			    }
			    break;
		    }
		}
	    }
	    else
	    {
		c_to_output = 1;
	    }
	    if( c_to_output )
	    {
		if( str_ptr >= str_size )
		{
		    if( fn_num == FN_SPRINTF )
		    {
			//sprintf:
			if( pix_vm_resize_container( str, (PIX_INT)str_ptr + 8, 1, PIX_CONTAINER_TYPE_INT8, vm ) ) 
			    break;
		    }
		    else
		    {
			//printf:
			cstr = (UTF8_CHAR*)mem_resize( cstr, str_ptr + 8 );
			if( cstr == 0 ) break;
		    }
		    str_size = str_ptr + 8;
		}
		if( fn_num == FN_SPRINTF )
		{
		    //sprintf:
		    ((UTF8_CHAR*)str_cont->data)[ str_ptr ] = c;
		}
		else
		{
		    //printf:
		    cstr[ str_ptr ] = c;
		}
		str_ptr++;
	    }
	}
	if( fn_num == FN_SPRINTF )
	{
	    //sprintf:
	    if( str_ptr < str_size )
	    {
		((UTF8_CHAR*)str_cont->data)[ str_ptr ] = 0;
	    }
	}
	else
	{
	    //printf:
	    if( cstr )
	    {
		if( str_ptr + 1 >= str_size )
		{
		    cstr = (UTF8_CHAR*)mem_resize( cstr, str_ptr + 1 );
		}
		if( cstr )
		{
		    cstr[ str_ptr ] = 0;
		    printf( "%s", cstr );
		}
	    }
	}
	mem_free( cstr );
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)str_ptr;
    }
    else 
    {
	//Some error occured:
sprintf_error:
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

//Issue a command:
void fn_system( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
#ifndef WINCE
    PIX_CID name;
    pix_vm_container* name_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 1 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name, 0, PIX_CID );
	if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name ] == 0 ) { err = 1; break; }
	name_cont = vm->c[ name ];
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free = 0;
	UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );

	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)system( ts );
	
	if( need_to_free ) mem_free( ts );
    }
    else 
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)system( 0 );
    }
#else
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = 0;
#endif
}

void fn_argc( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = (PIX_INT)g_wm->argc;
}

void fn_argv( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    bool err = 1;
    if( pars_num >= 1 )
    {
	PIX_INT arg_num;
	GET_VAL_FROM_STACK( arg_num, 0, PIX_INT );
	if( (unsigned)arg_num < g_wm->argc && g_wm->argv && g_wm->argv[ arg_num ] )
	{
	    int arg_len = (int)mem_strlen( g_wm->argv[ arg_num ] );
	    PIX_CID arg = pix_vm_new_container( -1, arg_len, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
	    if( arg >= 0 )
	    {
		pix_vm_container* arg_cont = vm->c[ arg ];
		mem_copy( arg_cont->data, g_wm->argv[ arg_num ], arg_len );
		stack_types[ sp + ( pars_num - 1 ) ] = 0;
		stack[ sp + ( pars_num - 1 ) ].i = arg;
		err = 0;
	    }
	}
    }
    if( err )
    {
	stack_types[ sp + ( pars_num - 1 ) ] = 0;
	stack[ sp + ( pars_num - 1 ) ].i = -1;
    }
}

void fn_exit( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    if( pars_num >= 1 )
    {
	GET_VAL_FROM_STACK( g_wm->exit_code, 0, int );
    }
    th->active = 0;
    g_wm->exit_request = 1;
}

//
// Private API
//
	
void fn_system_copy_OR_open_url( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    PIX_CID name;
    pix_vm_container* name_cont;
    
    bool err = 0;
    
    //Get parameters:
    while( 1 )
    {
	if( pars_num < 1 ) { err = 1; break; }
	GET_VAL_FROM_STACK( name, 0, PIX_CID );
	if( (unsigned)name >= (unsigned)PIX_VM_CONTAINERS ) { err = 1; break; }
	if( vm->c[ name ] == 0 ) { err = 1; break; }
	name_cont = vm->c[ name ];
	break;
    }
    
    //Execute:
    if( err == 0 )
    {
	bool need_to_free = 0;
	UTF8_CHAR* ts = pix_vm_make_cstring_from_container( name, &need_to_free, vm );
    
	if( fn_num == FN_SYSTEM_COPY )
	    system_copy( ts, 0 );
	if( fn_num == FN_OPEN_URL )
	    open_url( ts );
	
	if( need_to_free ) mem_free( ts );
    }
}

void fn_system_paste( PIX_BUILTIN_FN_PARAMETERS )
{
    FN_HEADER;
    
    stack_types[ sp + ( pars_num - 1 ) ] = 0;
    stack[ sp + ( pars_num - 1 ) ].i = -1;
    
    UTF8_CHAR* fname = system_paste( 0 );
    if( fname )
    {
        PIX_CID name = pix_vm_new_container( -1, mem_strlen( fname ), 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
        if( name >= 0 )
        {
            pix_vm_container* name_cont = vm->c[ name ];
            mem_copy( name_cont->data, fname, mem_strlen( fname ) );
            stack_types[ sp + ( pars_num - 1 ) ] = 0;
            stack[ sp + ( pars_num - 1 ) ].i = name;
        }
        mem_free( fname );
    }
}

void fn_system_webserver( PIX_BUILTIN_FN_PARAMETERS )
{
    vm->webserver_request = 1;
    while( vm->webserver_request != 0 )
    {
        time_sleep( 100 );
    }
}
